Require Import Arith.
Require Import Omega.
Require Import Bool.
Require Import List.
Require Import Program.Equality.

Require Import step_defs.
Require Import step_db_basics.
Require Import step_reduction_basics.

(* This file proves the diamond property for the parallel reduction
   relation, but nothing about the standard cbv relation for now,
   because I didn't need it in the main theorem. *)
Theorem open_preserves_pstep : forall a1 a2,
     pstep a1 a2
  -> forall v1 v2, pstep v1 v2 -> val v1
      -> forall n, pstep ({n ~> v1}a1) ({n ~> v2}a2).
Proof.
  induction 1; intros; simpl in *; try constructor; eauto;
  try rewrite open_open_swap with (n2 := 0) by auto;
  try rewrite open_open_swap with (n2 := 0) by auto;
  try solve [constructor; autorewrite with step; eauto];
  try match goal with
  | [ |- pstep (exp_app (exp_fun _) _) _ ] =>
    extract_open_lift;
    rewrite open_lift_commute_1 by auto; rewrite lift_0_is_id; simpl;
    rewrite lift_lift_collapse by auto; simpl;
    constructor; autorewrite with step; eauto
  | [ |- pstep (exp_pcase _ _) _ ] => 
    rewrite open_open_swap with (n2 := 0) (a2 := exp_join) by auto;
    simpl; rewrite open_lift_commute_1 by auto;
    constructor; autorewrite with step; eauto
  | [ |- pstep (exp_app (exp_nrec ({S (S ?n) ~> _} _)) _) _ ] =>
    simpl; repeat (rewrite lift_lift_collapse by auto); simpl;
    replace (S (S (S (S (S n))))) with (3 + (S (S n))) by omega;
    replace 5 with (3 + 2) by omega;
    rewrite open_lift_commute by omega; eauto 7
  end.
Qed.
Hint Resolve open_preserves_pstep.

Lemma open_preserves_mpstep_l : forall a b1 b2,
     mpstep b1 b2
  -> forall n, mpstep ({n ~> b1}a) ({n ~> b2}a).
Proof.
  induction 1; intros; eauto.
Qed.
Hint Immediate open_preserves_mpstep_l.
 

Theorem diamond_pstep : forall a b1,
     pstep a b1 -> forall b2, pstep a b2
  -> exists c, pstep b1 c /\ pstep b2 c.
Proof.
  (* This tactic does 2 things:
       - Find the IHs and the corresponding (par a b2) derivations and
         apply the IHs appropriately
       - Destruct the resulting existential with some fresh names
         involving "c" *)
  Ltac inst_pstep := repeat (match goal with
  | [ IHPstep : forall b, pstep (?P ?a) b -> exists _, pstep (?P ?b1) _ /\ _,
      HPstep : pstep ?a ?b |- _ ] =>
    require_neq b1 b;
    assert (pstep (P a) (P b)) by auto; clear HPstep
  | [ IHPstep : forall b, pstep (?P ?a1 ?a2) b -> exists _, pstep (?P ?b1' ?b2') _ /\ _,
      HPstep1 : pstep ?a1 ?b1, HPstep2 : pstep ?a2 ?b2 |- _ ] =>
    require_neq b1 b1'; require_neq b2 b2';
    assert (pstep (P a1 a2) (P b1 b2)) by auto; clear HPstep1 HPstep2
  | [ IHPstep : forall b, pstep ?a b -> exists _, pstep ?b1 _ /\ _,
      HPstep : pstep ?a ?b |- _ ] =>
    require_neq b1 b; let IH' := fresh "IH" in
    assert (IH' := IHPstep _ HPstep); clear IHPstep;
    let c := fresh "c" in
    let HcL := fresh "HcL" in
    let HcR := fresh "HcR" in
    destruct IH' as [c [HcL HcR]]
  | [ H : pstep exp_zero _ |- _ ] => inversion H; clear H; subst
  | [ H : pstep (exp_succ _) _ |- _ ] => inversion H; clear H; subst
  | [ H : pstep (exp_lam _) _ |- _ ] => inversion H; clear H; subst
  | [ H : pstep (exp_fun _) _ |- _ ] => inversion H; clear H; subst
  | [ H : pstep (exp_nrec _) _ |- _ ] => inversion H; clear H; subst
  | [ H : pstep (exp_roll _) _ |- _ ] => inversion H; clear H; subst
  | [ H : pstep (exp_succ _) _ |- _ ] => inversion H; clear H; subst
  | [ H : pstep (exp_inl _) _ |- _ ] => inversion H; clear H; subst
  | [ H : pstep (exp_inr _) _ |- _ ] => inversion H; clear H; subst
  | [ H : pstep (exp_pair _ _) _ |- _ ] => inversion H; clear H; subst
  end).

  induction 1; intros; simpl in *; eauto;
  try match goal with
    (* We want to invert the second derivation.  I'm going to find all
       your secrets. *)
    | [ H : pstep _ ?b2 |- exists _, _ /\ pstep ?b2 _ ] =>
      inversion H; subst; simpl in *; clear H
  end; inst_pstep; eauto 15.
Qed.

(* Now we extend the diamond property to mpstep, using a "stripping" lemma *)

Theorem stripping_mpstep : forall a b1,
     mpstep a b1 -> forall b2, pstep a b2
  -> exists c, pstep b1 c /\ mpstep b2 c.
Proof.
  induction 1; intros; subst; eauto;
  match goal with
  | [ HP1 : pstep ?A ?B1, HMP1 : mpstep ?B1 ?B1',
      HP2 : pstep ?A ?B2, IH : forall _, pstep ?B1 _ 
                    -> exists _:_, _ /\ _
     |- exists _:exp, pstep ?B1' _ /\ mpstep ?B2 _ ] =>
    let B := fresh "B" in
    let C := fresh "C" in
    destruct (diamond_pstep _ _ HP1 _ HP2) as [B [HB1 HB2]]; eauto;
    destruct (IH _ HB1) as [C [HC1 HC2]]; exists C; split; eauto
  end.
Qed.

Theorem diamond_mpstep : forall a b1,
     mpstep a b1 -> forall b2, mpstep a b2
  -> exists c, mpstep b1 c /\ mpstep b2 c.
Proof.
  induction 1; intros; eauto;
  match goal with
  | [ HP1 : pstep ?A ?A1, HMP1 : mpstep ?A1 ?B1, HMP2 : mpstep ?A ?B2,
      IH : forall _, _ -> exists _:_, _ /\ _ 
     |- exists _:_, mpstep ?B1 _ /\ mpstep ?B2 _ ] =>
    let B2' := fresh B2 "'" in
    destruct (stripping_mpstep _ _ HMP2 _ HP1) as [B2' [HB21 HB22]];
    let C := fresh "C" in
    destruct (IH _ HB22) as [C [? ?]]; eauto
  end.
Qed.

(* Finally, just for good measure, show the diamond property for mstep too *)
Lemma MPTrans : forall a b c,
  mpstep a b -> mpstep b c -> mpstep a c.
Proof.
  induction 1; intros; eauto.
Qed.

Theorem mstep_mpstep : forall n a b,
  mstep n a b -> mpstep a b.
Proof.
  induction 1; eauto.
Qed.
Hint Extern 2 (mpstep ?a ?b) =>
  eapply mstep_mpstep; eassumption.


(* we need to show pstep -> mstep first, and I'm too lazy *)
(*Theorem diamond_mstep : forall n m a b1 b2,
     mstep n a b1 -> mstep m a b2 
  -> exists c, exists n', exists m', mstep n' b1 c /\ mstep m' b2 c.
Proof.
  intros a n m b1 b2 HMS1 HMS2;
  destruct (diamond_mpstep _ _ (mstep_mpstep _ _ _ HMS1) 
                             _ (mstep_mpstep _ _ _ HMS2)) as [? [? ?]];
  eauto.
Qed.
*)

Theorem open_preserves_mpstep : forall a1 a2,
     mpstep a1 a2 
  -> forall v1 v2, mpstep v1 v2 -> val v1 
      -> forall n, mpstep ({n ~> v1}a1) ({n ~> v2}a2).
Proof.
  induction 1; intros; eauto.
Qed.
Hint Resolve open_preserves_mpstep.
