Require Import Arith.
Require Import Bool.
Require Import List.

Require Import Wf_nat.

Require Import step_defs.
Require Import step_db_basics.

Require Import Program.Equality.
Require Import Omega.


(******************************************)
(******** Term environments (rhos) ********)

Notation env := (list exp).

(* the term interp is really just a multi-substitution *)
Fixpoint close (rho : env) (a : exp) (m : nat) {struct rho} : exp :=
  match rho with
  | nil       => a
  | b :: rho' => close rho' ({m ~> b} a) m
  end.

Fixpoint close_c (rho : env) (G : ctx) (m : nat) {struct rho} : ctx :=
  match rho with
  | nil       => G
  | A :: rho' => close_c rho' (open_c m A G) m
  end.

(***********************************************************)
(******** The ordering for the interpretation **************)
Inductive CV :=
  | Comp : CV
  | Val  : CV.

Fixpoint typ_size (A : exp) : nat :=
  match A with
  | exp_nat         => 0
  | exp_arr B1 B2   => 1 + typ_size B1 + typ_size B2
  | exp_at  A' _    => 1 + typ_size A'
  | exp_var _       => 0
  | exp_mu A'       => 1 + typ_size A'
  | exp_sum B1 B2   => 1 + typ_size B1 + typ_size B2
  | exp_sigma B1 B2 => 1 + typ_size B1 + typ_size B2
  | exp_eq _ _      => 0
  | exp_lam A'      => 1 + typ_size A'
  | exp_fun A'      => 1 + typ_size A'
  | exp_app A' _    => 1 + typ_size A'
  | _      => 0
  end.

(* Some lemmas about type size to help automation *)
Lemma ts_arr_sub1 : forall A B,
  typ_size A < typ_size (exp_arr A B).
Proof. intros; simpl; omega. Defined.
Hint Resolve ts_arr_sub1.

Lemma lift_preserves_ts : forall A n m,
  typ_size (lift A n m) = typ_size A.
Proof.
  induction A; intros; simpl in *; eauto.
Defined.
Hint Rewrite lift_preserves_ts : step.
Hint Extern 1 =>
  match goal with
  | [ |- context[ typ_size (lift _ _ _) ] ] => rewrite lift_preserves_ts
  end.

Lemma ts_arr_sub2 : forall A B,
  typ_size B < typ_size (exp_arr A B).
Proof.
  intros; simpl; omega.
Defined.
Hint Resolve ts_arr_sub2.

Lemma ts_at_sub : forall A t,
  typ_size A < typ_size (exp_at A t).
Proof. intros; simpl; omega. Defined.

Lemma ts_sum_sub_l : forall A B,
  typ_size A < typ_size (exp_sum A B).
Proof. intros; simpl; omega. Defined.

Lemma ts_sum_sub_r : forall A B,
  typ_size B < typ_size (exp_sum A B).
Proof. intros; simpl; omega. Defined.

Lemma ts_sigma_sub_l : forall A B,
  typ_size A < typ_size (exp_sigma A B).
Proof. intros; simpl; omega. Defined.

Lemma ts_sigma_sub_r : forall A B,
  typ_size B < typ_size (exp_sigma A B).
Proof. intros; simpl; omega. Defined.

Lemma ts_app_sub : forall A A' B,
     typ_size A < typ_size A'
  -> typ_size A < typ_size (exp_app A' B).
Proof.  intros; simpl; omega.  Defined.

Implicit Arguments ts_arr_sub1 [A B].
Implicit Arguments ts_arr_sub2 [A B].
Implicit Arguments ts_at_sub [A t].
Implicit Arguments ts_sum_sub_l [A B].
Implicit Arguments ts_sum_sub_r [A B].
Implicit Arguments ts_sigma_sub_l [A B].
Implicit Arguments ts_sigma_sub_r [A B].
Implicit Arguments ts_app_sub [A A' B].

Notation ntc := (nat * exp * CV) % type.

Inductive IOrder : ntc -> ntc -> Prop :=
  | IONat : forall n1 n2 A1 A2 cv1 cv2,
     n1 < n2 -> IOrder (n1,A1,cv1) (n2,A2,cv2)
  | IOTyp : forall n A1 A2 cv1 cv2,
     typ_size A1 < typ_size A2 -> IOrder (n,A1,cv1) (n,A2,cv2)
  | IOCV  : forall n A,
      IOrder (n,A,Val) (n,A,Comp)
.
Hint Constructors IOrder.
Implicit Arguments IONat [n1 n2 A1 A2 cv1 cv2].
Implicit Arguments IOTyp [n A1 A2 cv1 cv2].

Theorem wfIO : well_founded IOrder.
Proof.
  intros p1; destruct p1 as [[n A] cv]. revert A cv.

  apply (lt_wf_rec n (fun n => forall A cv, Acc IOrder (n, A, cv)));
  clear n; intros n Hwfnat A.

  apply (induction_ltof1 _ typ_size 
                        (fun A => forall cv, Acc IOrder (n,A,cv)));
  clear A; intros A Hwftyp cv.

  apply Acc_intro; intros p2 HOrd; inversion HOrd; subst; eauto;
  apply Acc_intro; intros p3 HOrd'; inversion HOrd'; subst; eauto.
Defined.
Hint Resolve wfIO.

(* Some particular cases to help with the function definition *)
Theorem IOrder_minus : forall k j A,
  IOrder (k-j, A, Val) (k, A, Comp).
Proof.
  intros; destruct j;
  [ rewrite <- minus_n_O; apply IOCV |
    destruct k; simpl; eauto; apply IONat; omega ].
Defined.
Hint Resolve IOrder_minus.

Theorem IO_le_sub : forall k j A B cv cv',
     j <= k 
  -> typ_size A < typ_size B
  -> IOrder (j, A, cv) (k, B, cv').
Proof.
  intros; match goal with
  | [ Hleq : _ <= _ |- _] => destruct (le_lt_or_eq _ _ Hleq)
  end; subst; eauto.
Defined. 
Implicit Arguments IO_le_sub [k j A B cv cv'].

Fixpoint looks_like_nat (a : exp) : Prop :=
  match a with
  | exp_zero => True
  | exp_succ a' => looks_like_nat a'
  | _ => False
  end.

(* This is the definition of the interp except that we've made
   the recursive call a parameter *)
Definition interp_innards p1 :
     (forall p2, IOrder p2 p1 -> th -> env -> exp -> Prop)
  -> th -> env -> exp -> Prop :=
  match p1 as p1
    return    (forall p2, IOrder p2 p1 -> th -> env -> exp -> Prop) 
           -> th -> env -> exp -> Prop
    with
  | (k,A,cv) =>
    match cv as cv
      return   (forall y, IOrder y (k, A, cv) -> th -> env -> exp -> Prop)
            -> th -> env -> exp -> Prop
    with 
    | Comp => fun f t rho a => 
          typing nil t a (close rho A 0)
       /\ match t with
          | th_prog => forall j v, j <= k -> mstep j a v -> val v
                         -> f (k - j, A, Val) (IOrder_minus _ _ _)
                              th_prog rho v
          | th_log => exists v, exists n, mstep n a v
                         /\ f (k,A,Val) (IOCV _ _) th_log rho v
          end
    | Val =>
      match A as A
        return (forall y, IOrder y (k, A, Val) -> th -> env -> exp -> Prop)
            -> th -> env -> exp -> Prop
      with
      | exp_star   => fun _ t _ a =>
        val a /\ typing nil t a exp_star
      | exp_var _    => fun _ _ _   _ => False
      | exp_nat      => fun _ _ _   a => looks_like_nat a
      | exp_at A' t' => fun f t rho a =>
            typing nil t' (close rho A' 0) exp_star
         /\ f (k,A',Val) (IOTyp ts_at_sub) t' rho a
      | exp_arr B1 B2 => fun f t rho a =>
            typing nil t a (close rho (exp_arr B1 B2) 0)
         /\ match t with
            | th_log =>
              match a with
              | exp_lam b => 
                  forall j v (Hle : j <= k),
                              f (j,B1,Val) (IO_le_sub Hle ts_arr_sub1)
                                t rho v
                           -> f (j,B2,Comp) (IO_le_sub Hle ts_arr_sub2)
                                th_log (v :: rho) ({0 ~> v}b) 
             | exp_nrec b =>
                  forall j v (Hle : j <= k),
                              f (j,B1,Val) (IO_le_sub Hle ts_arr_sub1)
                                t rho v
                           -> f (j,B2,Comp) (IO_le_sub Hle ts_arr_sub2)
                                th_log (v :: rho)
                                ({0 ~> v}
                                   ({0 ~> exp_lam (exp_lam (exp_app 
                                            (exp_nrec (lift b 3 2)) (exp_var 1)))}
                                    b))

             | _ => False
              end
            | th_prog =>
              match a with
              | exp_lam b =>
                  forall j v (Hlt : j < k),
                              f (j,B1,Val) (IONat Hlt) t rho v
                           -> f (j,B2,Comp) (IONat Hlt) 
                                th_prog (v :: rho) ({0 ~> v}b)
              | exp_fun b =>
                  forall j v (Hlt : j < k),
                              f (j,B1,Val) (IONat Hlt) t rho v
                           -> f (j,B2,Comp) (IONat Hlt) th_prog (v :: rho)
                                  ({0 ~> v} ({0 ~> exp_fun b} b))
             | exp_nrec b =>
                  forall j v (Hlt : j < k),
                              f (j,B1,Val) (IONat Hlt) t rho v
                           -> f (j,B2,Comp) (IONat Hlt)
                                th_prog (v :: rho)
                                ({0 ~> v}
                                   ({0 ~> exp_lam (exp_lam (exp_app 
                                            (exp_nrec (lift b 3 2)) (exp_var 1)))}
                                    b))
              | _ => False
              end 
            end
      | exp_sum A B => fun f t rho a =>
        match a with
        | exp_inl a' =>   val a' 
                       /\ typing nil t (close rho (exp_sum A B) 0) exp_star
                       /\ f (k,A,Val) (IOTyp ts_sum_sub_l) t rho a'
        | exp_inr a' =>   val a'
                       /\ typing nil t (close rho (exp_sum A B) 0) exp_star
                       /\ f (k,B,Val) (IOTyp ts_sum_sub_r) t rho a'
        | _ => False
        end
      | exp_sigma B1 B2 => fun f t rho a =>
           typing nil t a (close rho (exp_sigma B1 B2) 0)
        /\ match a with
           | exp_pair b1 b2 => val b1 /\ val b2 
                            /\ f (k,B1,Val) (IOTyp ts_sigma_sub_l) t rho b1
                            /\ f (k,B2,Val) (IOTyp ts_sigma_sub_r) t (b1 :: rho) b2
           | _ => False
           end
      | exp_mu B => fun f t rho a => 
        match a with
        | exp_roll a' =>
             val a' /\ typing nil t (exp_roll a') (close rho (exp_mu B) 0)
          /\ forall j (Hlt : j < k),
                f (j,{0 ~> exp_mu B}B,Val) (IONat Hlt) t
                  rho a'
        | _ => False
        end 
      | exp_eq b1 b2 => fun f t rho a => 
            typing nil t (close rho (exp_eq b1 b2) 0) exp_star
         /\ match a with
            | exp_join => exists c, mpstep (close rho b1 0) c 
                                 /\ mpstep (close rho b2 0) c
            | _ => False
            end
      | _ => fun _ _ _ _ => False
      end
    end
  end.

Definition interp : ntc -> th -> env -> exp -> Prop :=
  Fix wfIO (fun _ => th -> env -> exp -> Prop) interp_innards.

(* We need a nice induction principle for reasoning about this
   definition.  Sadly, built in coq magic like Functional Scheme
   doesn't work.  So, we're going to have to define and prove our
   own principle.  There is a little support in the coq libraries
   for this kind of reasoning, but first we must demonstrate that
   interp is "extensional" in the sense that it doesn't care what
   proof of well-ordering it is passed.

   The problem is that the proof of extensionality that the
   coq library demands uses the propositional equality.  This makes
   sense, since it doesn't know my function actually defines
   an (exp -> Prop) for which a more extensional equality would
   work, but it's annoying.  As a result, I have to assume
   functional extensionality.  *)
Axiom extensionality : forall (A : Type) (B : A -> Type)
                              (f g : forall a : A, B a),
     (forall a : A, f a = g a) -> f = g.

Theorem interp_extensional : 
  forall (p1 : ntc) (f g : forall p2 : ntc, IOrder p2 p1 
                                         -> th -> env -> exp -> Prop),
        (forall (p2 : ntc) (io : IOrder p2 p1), f p2 io = g p2 io) 
     -> interp_innards p1 f = interp_innards p1 g.
Proof.                       
  intros;
  assert (f = g) by (eauto using extensionality); subst; eauto.
Qed.
Hint Resolve interp_extensional.


(* OK, now we prove a bunch of equations showing how the interp unfolds.
   Rewriting by these will be much faster than actually unfolding the
   definition. *)
Tactic Notation "prim_unfold_interp" "in" hyp(H) := 
  unfold interp in H; rewrite Fix_eq in H;
  [ simpl in H; fold interp in H | apply interp_extensional ].

Ltac prim_unfold_interp :=
  unfold interp; rewrite Fix_eq;
  [ simpl; fold interp | apply interp_extensional ].

Theorem ui_comp : forall k A t rho a,
    interp (k,A,Comp) t rho a
  = (   typing nil t a (close rho A 0)
     /\ match t with
        | th_prog => 
            forall j v, j <= k -> mstep j a v -> val v
                     -> interp (k - j, A, Val) th_prog rho v
        | th_log =>
            exists v, exists n, mstep n a v
                            /\ interp (k,A,Val) th_log rho v
        end).
Proof. intros; prim_unfold_interp; auto. Qed.

Theorem ui_val_var : forall k n t rho a,
    interp (k,exp_var n,Val) t rho a = False.
Proof. intros; prim_unfold_interp; auto. Qed.

Theorem ui_val_nat : forall k t rho a,
    interp (k,exp_nat,Val) t rho a
  = looks_like_nat a.
Proof. intros; prim_unfold_interp; auto. Qed.

Theorem ui_val_at : forall k A t' t rho a,
    interp (k,exp_at A t',Val) t rho a
  = (typing nil t' (close rho A 0) exp_star /\ interp (k,A,Val) t' rho a).
Proof. intros; prim_unfold_interp; auto. Qed.

Theorem ui_val_arr : forall k B1 B2 t rho a,
    interp (k,exp_arr B1 B2,Val) t rho a
  = (   typing nil t a (close rho (exp_arr B1 B2) 0)
     /\ match t with
        | th_log =>
          match a with
          | exp_lam b => 
              forall j v, j <= k
               -> interp (j,B1,Val) t rho v
               -> interp (j,B2,Comp) th_log (v :: rho) ({0 ~> v}b)
          | exp_nrec b =>
               forall j v, j <= k
                -> interp (j,B1,Val) t rho v 
                -> interp (j,B2,Comp) th_log (v :: rho)
                     ({0 ~> v}
                        ({0 ~> exp_lam (exp_lam (exp_app 
                                 (exp_nrec (lift b 3 2)) (exp_var 1)))}
                          b))
          | _ => False
          end
        | th_prog =>
          match a with
          | exp_lam b =>
              forall j v, j < k
               -> interp (j,B1,Val) t rho v
               -> interp (j,B2,Comp) th_prog (v :: rho) ({0 ~> v}b)
          | exp_fun b =>
              forall j v, j < k
               -> interp (j,B1,Val) t rho v
               -> interp (j,B2,Comp) th_prog (v :: rho)
                         ({0 ~> v} ({0 ~> exp_fun b} b))
          | exp_nrec b =>
               forall j v, j < k
                 -> interp (j,B1,Val) t rho v
                 -> interp (j,B2,Comp) th_prog (v :: rho)
                      ({0 ~> v}
                         ({0 ~> exp_lam (exp_lam (exp_app 
                                  (exp_nrec (lift b 3 2)) (exp_var 1)))}
                           b))
          | _ => False
          end 
        end).
Proof. intros; prim_unfold_interp; auto. Qed.

Theorem ui_val_sum : forall k A B t rho a,
    interp (k,exp_sum A B,Val) t rho a
  = match a with
    | exp_inl a' =>   val a'
                   /\ typing nil t (close rho (exp_sum A B) 0) exp_star
                   /\ interp (k,A,Val) t rho a'
    | exp_inr a' =>   val a'
                   /\ typing nil t (close rho (exp_sum A B) 0) exp_star
                   /\ interp (k,B,Val) t rho a'
    | _ => False
    end.
Proof. intros; prim_unfold_interp; auto. Qed.

Theorem ui_val_sigma : forall k B1 B2 t rho a,
    interp (k,exp_sigma B1 B2,Val) t rho a
  = (   typing nil t a (close rho (exp_sigma B1 B2) 0)
     /\ match a with
        | exp_pair b1 b2 => val b1 /\ val b2
                         /\ interp (k,B1,Val) t rho b1
                         /\ interp (k,B2,Val) t (b1 :: rho) b2
        | _ => False
        end).
Proof. intros; prim_unfold_interp; auto. Qed.


Theorem ui_val_mu : forall k A t rho a,
    interp (k,exp_mu A,Val) t rho a
  = match a with
    | exp_roll a' =>
         val a' /\ typing nil t (exp_roll a') (close rho (exp_mu A) 0)
      /\ forall j (Hlt : j < k),
            interp (j,{0 ~> exp_mu A}A,Val) t rho a'
    | _ => False
    end.
Proof. intros; prim_unfold_interp; auto. Qed.

Theorem ui_val_eq : forall k b1 b2 t rho a,
    interp (k,exp_eq b1 b2,Val) t rho a
  = (   typing nil t (close rho (exp_eq b1 b2) 0) exp_star
     /\ match a with
        | exp_join => exists c, mpstep (close rho b1 0) c 
                             /\ mpstep (close rho b2 0) c
        | _ => False
        end).
Proof. intros; prim_unfold_interp; auto. Qed.

Theorem ui_val_star : forall k t rho a,
    interp (k,exp_star,Val) t rho a
  = (val a /\ typing nil t a exp_star).
Proof. intros; prim_unfold_interp; auto. Qed.

Theorem ui_val_fun : forall k b t rho a,
    interp (k,exp_fun b,Val) t rho a = False. 
Proof. intros; prim_unfold_interp; auto. Qed.

Theorem ui_val_lam : forall k b t rho a,
    interp (k,exp_lam b,Val) t rho a = False. 
Proof. intros; prim_unfold_interp; auto. Qed.

Theorem ui_val_nrec : forall k b t rho a,
    interp (k,exp_nrec b,Val) t rho a = False. 
Proof. intros; prim_unfold_interp; auto. Qed.

Theorem ui_val_zero : forall k t rho a,
    interp (k,exp_zero,Val) t rho a = False. 
Proof. intros; prim_unfold_interp; auto. Qed.

Theorem ui_val_succ : forall k b t rho a,
    interp (k,exp_succ b,Val) t rho a = False. 
Proof. intros; prim_unfold_interp; auto. Qed.

Theorem ui_val_ncase : forall k b b1 b2 t rho a,
    interp (k,exp_ncase b b1 b2,Val) t rho a = False. 
Proof. intros; prim_unfold_interp; auto. Qed.

Theorem ui_val_inl : forall k b t rho a,
    interp (k,exp_inl b,Val) t rho a = False. 
Proof. intros; prim_unfold_interp; auto. Qed.

Theorem ui_val_inr : forall k b t rho a,
    interp (k,exp_inr b,Val) t rho a = False. 
Proof. intros; prim_unfold_interp; auto. Qed.

Theorem ui_val_pair : forall k b1 b2 t rho a,
    interp (k,exp_pair b1 b2,Val) t rho a = False. 
Proof. intros; prim_unfold_interp; auto. Qed.

Theorem ui_val_pcase : forall k b1 b2 t rho a,
    interp (k,exp_pcase b1 b2,Val) t rho a = False. 
Proof. intros; prim_unfold_interp; auto. Qed.

Theorem ui_val_scase : forall k b b1 b2 t rho a,
    interp (k,exp_scase b b1 b2,Val) t rho a = False. 
Proof. intros; prim_unfold_interp; auto. Qed.

Theorem ui_val_roll : forall k b t rho a,
    interp (k,exp_roll b,Val) t rho a = False. 
Proof. intros; prim_unfold_interp; auto. Qed.

Theorem ui_val_unroll : forall k b t rho a,
    interp (k,exp_unroll b,Val) t rho a = False. 
Proof. intros; prim_unfold_interp; auto. Qed.

Theorem ui_val_join : forall k t rho a,
    interp (k,exp_join,Val) t rho a = False. 
Proof. intros; prim_unfold_interp; auto. Qed.

Theorem ui_val_app : forall k b1 b2 t rho a,
    interp (k,exp_app b1 b2,Val) t rho a = False.
Proof. intros; prim_unfold_interp; auto. Qed.

(* Now a few tactics for unfolding the interpretation and cleaning up the
   resulting mess *)
Ltac unfold_cleanup :=
  repeat (match goal with
  | [ H : False |- _ ] => inversion H
  | [ H : val _ /\ _ |- _ ] =>
    let HVal := fresh "HVal" in destruct H as [HVal H]
  | [ H : typing _ _ _ _ /\ _ |- _ ] =>
    let HTyp := fresh "HTyp" in destruct H as [HTyp H]
  | [ H : interp _ _ _ _ /\ _ |- _ ] =>
    let H' := fresh "H" in destruct H as [H H']
  | [ H : match ?t with | th_prog => _ | th_log => _ end |- _ ] =>
    destruct t
  | [ H : exists v, exists n, mstep _ _ _ /\ _ |- _] =>
    let v := fresh "v" in let n := fresh "n" in let Hastep := fresh "Hastep" in
    destruct H as [v [n [Hastep H]]]
  | [ H : exists c, mpstep _ c /\ mpstep _ c |- _ ] =>
    destruct H as [? [? ?]]
  | [ H : match ?v with | exp_var _ => _ | exp_star => _ | exp_arr _ _ => _ 
            | exp_fun _ => _ | exp_lam _ => _ | exp_nrec _ => _ | exp_app _ _ => _
            | exp_at _ _ => _ | exp_nat => _  | exp_zero => _ | exp_succ _ => _
            | exp_ncase _ _ _ => _ | exp_sum _ _ => _
            | exp_inl _ => _ | exp_inr _ => _ | exp_scase _ _ _ => _
            | exp_sigma _ _ => _ | exp_pair _ _ => _ | exp_pcase _ _ => _
            | exp_mu _ => _ | exp_roll _ => _ | exp_unroll _ => _ 
            | exp_eq _ _ => _ | exp_join => _
          end |- _ ] =>
    destruct v; try solve [inversion H]
  end).

Tactic Notation "unfold_interp" "in" hyp(H) := 
  first [rewrite ui_comp in H | rewrite ui_val_var in H |
         rewrite ui_val_nat in H | rewrite ui_val_at in H |
         rewrite ui_val_arr in H | rewrite ui_val_sum in H |
         rewrite ui_val_star in H | rewrite ui_val_fun in H |
         rewrite ui_val_lam in H | rewrite ui_val_nrec in H |
         rewrite ui_val_zero in H |
         rewrite ui_val_succ in H | rewrite ui_val_ncase in H |
         rewrite ui_val_inl in H | rewrite ui_val_inr in H |
         rewrite ui_val_sigma in H | rewrite ui_val_pair in H |
         rewrite ui_val_pcase in H |
         rewrite ui_val_scase in H | rewrite ui_val_join in H |
         rewrite ui_val_app in H |
         rewrite ui_val_roll in H | rewrite ui_val_unroll in H |
         rewrite ui_val_mu in H | rewrite ui_val_eq in H ];
  simpl in H;
  unfold_cleanup.

Ltac unfold_interp_hyps :=
  repeat (match goal with
  | [ H : interp (_,_,Comp) _ _ _ |- _ ] => unfold_interp in H
  | [ H : interp (_,?A,Val) _ _ _ |- _ ] =>
    first [is_var A | is_evar A | unfold_interp in H]
          end).

Ltac unfold_interp :=
  first [rewrite ui_comp | rewrite ui_val_var |
         rewrite ui_val_nat | rewrite ui_val_at |
         rewrite ui_val_arr | rewrite ui_val_sum |
         rewrite ui_val_star | rewrite ui_val_fun |
         rewrite ui_val_lam | rewrite ui_val_nrec |
         rewrite ui_val_zero |
         rewrite ui_val_succ | rewrite ui_val_ncase |
         rewrite ui_val_inl | rewrite ui_val_inr |
         rewrite ui_val_scase | rewrite ui_val_join |
         rewrite ui_val_sigma | rewrite ui_val_pair |
         rewrite ui_val_pcase |
         rewrite ui_val_app |
         rewrite ui_val_mu | rewrite ui_val_eq ];
  repeat match goal with
  | [ |- typing _ _ _ _ /\ _ ] => split; [ eauto | ]
  | [ |- val _ /\ _ ] => split; [ eauto | ]
  | [ |- interp _ _ _ _ /\ _ ] => split
  | [ |- match ?t with | th_log => _ | th_prog => _ end ] =>
    destruct t
  end; simpl.

Global Opaque interp_innards.
Global Opaque interp.

(* Induction Scheme doesn't seem to work for interp.  So!  We define our
   own induction principle by hand ... *)
Theorem interp_induction :
     forall (P : nat -> exp -> CV -> th -> env -> exp -> Prop),
     (forall k A rho a, typing nil th_prog a (close rho A 0)
        -> (forall j v, j <= k -> mstep j a v -> val v
                     -> P (k - j) A Val th_prog rho v)
        -> (forall j v, j <= k -> mstep j a v -> val v
                     -> interp ((k - j),A,Val) th_prog rho v)
        -> P k A Comp th_prog rho a)
  -> (forall k A rho a v n, typing nil th_log a (close rho A 0)
        -> mstep n a v
        -> interp (k,A,Val) th_log rho v
        -> P k A Val th_log rho v
        -> P k A Comp th_log rho a)
  -> (forall k t rho a, val a -> typing nil t a exp_star
                              -> P k exp_star Val t rho a)
  -> (forall k t rho a, looks_like_nat a -> P k exp_nat Val t rho a)
  -> (forall k A t' t rho a,
           interp (k,A,Val) t' rho a
        -> typing nil t' (close rho A 0) exp_star
        -> P k A Val t' rho a
        -> P k (exp_at A t') Val t rho a)
  -> (forall k A B rho a, typing nil th_log (exp_lam a)
                                    (close rho (exp_arr A B) 0)
        -> (forall j v, j <= k
                -> interp (j,A,Val) th_log rho v
                -> interp (j,B,Comp) th_log (v :: rho) ({0 ~> v}a))
        -> (forall j v, j <= k
                -> interp (j,A,Val) th_log rho v
                -> P j A Val th_log rho v 
                -> P j B Comp th_log (v :: rho) ({0 ~> v}a))
        -> (forall j v, j <= k
                -> interp (j,A,Val) th_log rho v
                -> P j A Val th_log rho v)
        -> P k (exp_arr A B) Val th_log rho (exp_lam a))
  -> (forall k A B rho a, typing nil th_prog (exp_lam a)
                                 (close rho (exp_arr A B) 0)
        -> (forall j v, j < k
                -> interp (j,A,Val) th_prog rho v
                -> interp (j,B,Comp) th_prog (v :: rho) ({0 ~> v}a))
        -> (forall j v, j < k
                -> interp (j,A,Val) th_prog rho v
                -> P j A Val th_prog rho v
                -> P j B Comp th_prog (v :: rho) ({0 ~> v}a))
        -> (forall j v, j < k 
              -> interp (j,A,Val) th_prog rho v -> P j A Val th_prog rho v)
        -> P k (exp_arr A B) Val th_prog rho (exp_lam a))
  -> (forall k A B rho a, typing nil th_log (exp_nrec a)
                                    (close rho (exp_arr A B) 0)
        -> (forall j v, j <= k
                -> interp (j,A,Val) th_log rho v
                -> interp (j,B,Comp) th_log (v :: rho)
                          ({0 ~> v}
                            ({0 ~> exp_lam (exp_lam (exp_app 
                                   (exp_nrec (lift a 3 2)) (exp_var 1)))}
                             a)))
        -> (forall j v, j <= k
                -> interp (j,A,Val) th_log rho v
                -> P j A Val th_log rho v 
                -> P j B Comp th_log (v :: rho)
                     ({0 ~> v}
                       ({0 ~> exp_lam (exp_lam (exp_app 
                              (exp_nrec (lift a 3 2)) (exp_var 1)))}
                        a)))
        -> (forall j v, j <= k
                -> interp (j,A,Val) th_log rho v
                -> P j A Val th_log rho v)
        -> P k (exp_arr A B) Val th_log rho (exp_nrec a))
  -> (forall k A B rho a, typing nil th_prog (exp_nrec a)
                                    (close rho (exp_arr A B) 0)
        -> (forall j v, j < k
                -> interp (j,A,Val) th_prog rho v
                -> interp (j,B,Comp) th_prog (v :: rho)
                          ({0 ~> v}
                            ({0 ~> exp_lam (exp_lam (exp_app 
                                   (exp_nrec (lift a 3 2)) (exp_var 1)))}
                             a)))
        -> (forall j v, j < k
                -> interp (j,A,Val) th_prog rho v
                -> P j A Val th_prog rho v 
                -> P j B Comp th_prog (v :: rho)
                     ({0 ~> v}
                       ({0 ~> exp_lam (exp_lam (exp_app 
                              (exp_nrec (lift a 3 2)) (exp_var 1)))}
                        a)))
        -> (forall j v, j < k
                -> interp (j,A,Val) th_prog rho v
                -> P j A Val th_prog rho v)
        -> P k (exp_arr A B) Val th_prog rho (exp_nrec a))
  -> (forall k A B rho a, typing nil th_prog (exp_fun a)
                                    (close rho (exp_arr A B) 0)
        -> (forall j v, j < k
                -> interp (j,A,Val) th_prog rho v
                -> interp (j,B,Comp) th_prog (v :: rho)
                          ({0 ~> v}({0 ~> exp_fun a} a)))
        -> (forall j v, j < k
                -> interp (j,A,Val) th_prog rho v 
                -> P j A Val th_prog rho v
                -> P j B Comp th_prog (v :: rho) 
                     ({0 ~> v}({0 ~> exp_fun a} a)))
        -> (forall j v, j < k 
              -> interp (j,A,Val) th_prog rho v -> P j A Val th_prog rho v)
        -> P k (exp_arr A B) Val th_prog rho (exp_fun a))
  -> (forall k A B t rho a,
           interp (k,A,Val) t rho a
        -> typing nil t (close rho (exp_sum A B) 0) exp_star
        -> P k A Val t rho a
        -> P k (exp_sum A B) Val t rho (exp_inl a))
  -> (forall k A B t rho a,
           interp (k,B,Val) t rho a
        -> typing nil t (close rho (exp_sum A B) 0) exp_star
        -> P k B Val t rho a
        -> P k (exp_sum A B) Val t rho (exp_inr a))
  -> (forall k B1 B2 t rho b1 b2,
           interp (k,B1,Val) t rho b1
        -> interp (k,B2,Val) t (b1 :: rho) b2
        -> typing nil t (exp_pair b1 b2) (close rho (exp_sigma B1 B2) 0)
        -> P k B1 Val t rho b1
        -> P k B2 Val t (b1 :: rho) b2
        -> P k (exp_sigma B1 B2) Val t rho (exp_pair b1 b2))
  -> (forall k A t rho a,
           val a
        -> typing nil t (exp_roll a) (close rho (exp_mu A) 0)
        -> (forall j, j < k ->
              interp (j,{0 ~> exp_mu A}A,Val) t rho a)
        -> (forall j, j < k -> 
              P j ({0 ~> exp_mu A}A) Val t rho a)
        -> P k (exp_mu A) Val t rho (exp_roll a))
  -> (forall k b1 b2 t rho c,
           typing nil t (close rho (exp_eq b1 b2) 0) exp_star
        -> mpstep (close rho b1 0) c -> mpstep (close rho b2 0) c
        -> P k (exp_eq b1 b2) Val t rho exp_join)
  -> (forall k A cv t rho a, interp (k,A,cv) t rho a -> P k A cv t rho a).
Proof.
  intros P HCProg HCLog HVSort HVNat HVAt HVLamL HVLamP HVNRecL HVNRecP
         HVFun HSumL HSumR HSigma HVMu HVEq;
  induction k as [k IHWfk] using lt_wf_ind; induction A; destruct cv; intros;
  unfold_interp_hyps; try match goal with
  | [ |- P _ exp_nat Comp th_log _ _ ] => eauto
  | [ HRed : forall _ _, _ <= _ -> mstep _ ?a _ -> val _ -> interp _ _ _ _
     |- P _ _ Comp th_prog _ ?a ] =>
    apply HCProg; eauto; intros ? ? HLe HMs HVal;
    let HInterp := fresh "HInterp" in
    assert (HInterp := HRed _ _ HLe HMs HVal); clear HRed;
    unfold_interp_hyps; eauto
  | [ Hstep : mstep ?n' ?a ?v'
     |- P _ _ Comp th_log _ ?a ] =>
    apply HCLog with (v := v') (n := n'); auto;
    try unfold_interp; eauto
  end;
  try match goal with
  | [ |- P _ (exp_eq _ _) Val _ _ _ ] => eauto
  | [ |- P _ exp_star Val _ _ _ ] => eauto 6
  | [ |- P _ (exp_arr _ _) Val th_log _ (exp_lam _) ] =>
    apply HVLamL; eauto; intros;
    match goal with
    | [ Hle : _ <= _ |- _ ] => 
      apply le_lt_or_eq in Hle; destruct Hle; subst; eauto
    end
  | [ |- P _ (exp_arr _ _) Val th_log _ (exp_nrec _) ] =>
    apply HVNRecL; eauto; intros;
    match goal with
    | [ Hle : _ <= _ |- _ ] => 
      apply le_lt_or_eq in Hle; destruct Hle; subst; eauto
    end
  | [ |- P _ (exp_arr _ _) Val th_prog _ _ ] => eauto 6
  | [ |- P (?k - ?j) (?Q _ _) Val _ _ _ ] => 
    (* this does both At and Sum *)
    assert (k - j <= k) as Hleminus by omega; apply le_lt_or_eq in Hleminus;
    destruct Hleminus as [ Hle' | Hle' ]; eauto; try rewrite Hle' in *; eauto
  | [ |- P _ (exp_at _ _) Val _ _ _ ] => eauto 6
  | [ |- P _ (exp_sum _ _) Val _ _ _ ] => eauto 6
  | [ |- P _ (exp_sigma _ _) Val _ _ _ ] => eauto 6
  | [ |- P _ (exp_mu _) Val _ _ _ ] => eauto 6
  | [ |- P _ exp_nat Val _ _ _ ] => eauto
  end.
Qed.
