Require Import Arith.
Require Import Omega.
Require Import Bool.
Require Import List.

Require Import step_defs.
Require Import step_db_basics.
Require Import step_typing_basics.
Require Import step_substitution.
Require Import step_preservation.  (* just for invert_typ_wf, I think *)
Require Import step_interp.
Require Import step_reduction_basics.
Require Import step_confluence.

Require Import Program.Equality.

Hint Rewrite <- minus_n_O : step.
Hint Resolve le_O_n.

Theorem looks_like_nat_zero : looks_like_nat exp_zero.
Proof. simpl; eauto.  Qed.
Hint Resolve looks_like_nat_zero.

Theorem looks_like_nat_val : forall a,
  looks_like_nat a -> val a.
Proof.
  induction a; intros Hlook; try inversion Hlook; eauto.
Qed.
Hint Resolve looks_like_nat_val.

Theorem looks_like_nat_typing : forall G t a,
     looks_like_nat a
  -> ctx_wf G
  -> typing G t a exp_nat.
Proof.
  intros G t a; induction a; intros Hlook Hwf; try (inversion Hlook); eauto.
Qed.
Hint Immediate looks_like_nat_typing.

(* This one is nice just because it doesn't create a new existential
   for th or G *)
Theorem looks_like_nat_typing_inst : forall a,
     looks_like_nat a
  -> typing nil th_log a exp_nat.
Proof.
  eauto.
Qed.
Hint Immediate looks_like_nat_typing_inst.

Theorem looks_like_nat_step : forall a b,
  looks_like_nat a -> step a b -> False.
Proof.
  induction a; intros b Hlook HStep; inversion HStep; eauto.
Qed.
Hint Resolve looks_like_nat_step.

Theorem looks_like_nat_mstep : forall a b j,
     looks_like_nat a
  -> mstep j a b
  -> looks_like_nat b.
Proof.
  inversion 2; subst; eauto;
  edestruct looks_like_nat_step; eauto.
Qed.
Hint Resolve looks_like_nat_mstep.

Lemma looks_like_nat_closed : forall v,
  looks_like_nat v -> fvs v = 0.
Proof.
  induction v; intros H; simpl in H; eauto; inversion H.
Qed.
Hint Immediate looks_like_nat_closed.

(******************** Congruences for close *************)

(***
 * OK, there is a small issue with env_wf (defined below).  Often we
 * want to include it as a premise, just to show that all the terms in
 * the rho are closed.  But, lemmas that use it in this way won't
 * mention k and G in the conclusion, so one has eapply/erewrite with
 * them, messing up autorewrite in particular.  So, we define this
 * other datatype which only encodes the closedness bit, and use it as
 * a premise.  Then we prove env_wf implies closed, and now
 * autorewrite is happy.
 ***)

Inductive closed_rho : env -> Prop :=
  | CRNil  : closed_rho nil
  | CRCons : forall a rho, 
      fvs a = 0 -> val a -> closed_rho rho -> closed_rho (a :: rho).
Hint Constructors closed_rho.

Theorem close_open_commute : forall rho,
     closed_rho rho
  -> forall n m a b,
        n <= m
     -> close rho ({n ~> a}b) m = {n ~> close rho a m}(close rho b (S m)).
Proof.
  induction 1; intros; simpl; eauto; autorewrite with step;
  rewrite open_open_swap, lift_closed_id by auto;
  match goal with
  | [ IH : forall _ _ _ _, _ -> _ = _ |- _ ] =>
    rewrite <- IH by auto; auto
  end.
Qed.

Theorem close_c_open_e_commute : forall rho,
     closed_rho rho
  -> forall a G,
          close_c rho (open_c 0 a G) 0
        = open_c 0 (close rho a 0) (close_c rho G 1).
Proof.
  induction 1; intros; simpl; eauto; match goal with
  | [ IH : forall _ _, _ = _ |- _] => rewrite <- IH; eauto
  end; rewrite open_open_swap_c; eauto.
Qed.
Hint Rewrite close_c_open_e_commute using (eauto 2; fail) : step.

Theorem close_c_app : forall rho G1 G2 m,
    close_c rho (G1 ++ G2) m
  = (close_c rho G1 (m + length G2)) ++ close_c rho G2 m.
Proof.
  induction rho; intros; simpl; eauto;
  autorewrite with step; rewrite_ih ;autorewrite with step; eauto.
Qed.

Theorem close_c_nil : forall rho m,
  close_c rho nil m = nil.
Proof.
  induction rho; simpl; eauto.
Qed.

Theorem close_c_cons : forall rho,
    closed_rho rho
 -> forall t A G m ,
       close_c rho ((t,A) :: G) m
     = (t,close rho A (m + length G)) :: close_c rho G m.
Proof.
  induction 1; intros; simpl; eauto;
  rewrite_ih; autorewrite with step; eauto.
Qed.
Hint Rewrite close_c_app close_c_nil close_c_cons
       using (eauto 3; fail) : step.

Theorem close_c_length : forall rho G k,
  length (close_c rho G k) = length G.
Proof.
  induction rho; intros; simpl in *; 
  rewrite_ih; autorewrite with step; eauto.
Qed.
Hint Rewrite close_c_length : step.

Theorem close_cong_nat : forall rho m,
  close rho exp_nat m = exp_nat.
Proof.
  induction rho; intros; simpl in *; eauto.
Qed.

Theorem close_cong_at : forall rho A t m,
  close rho (exp_at A t) m = exp_at (close rho A m) t.
Proof.
  induction rho; intros; simpl in *; eauto.
Qed.

Theorem close_cong_eq : forall rho a b m,
  close rho (exp_eq a b) m = exp_eq (close rho a m) (close rho b m).
Proof.
  induction rho; intros; simpl in *; eauto.
Qed.

Theorem close_cong_sum : forall rho A B m,
  close rho (exp_sum A B) m = exp_sum (close rho A m) (close rho B m).
Proof.
  induction rho; intros; simpl in *; eauto.
Qed.

Theorem close_cong_star : forall rho m,
  close rho exp_star m = exp_star.
Proof.
  induction rho; intros; simpl in *; eauto.
Qed.

Theorem close_cong_join : forall rho m,
  close rho exp_join m = exp_join.
Proof.
  induction rho; intros; simpl in *; eauto.
Qed.

Theorem close_cong_arr : forall rho,
     closed_rho rho
  -> forall A B m,
       close rho (exp_arr A B) m
     = exp_arr (close rho A m) (close rho B (S m)).
Proof.
  induction 1; intros; simpl in *; rewrite_ih; autorewrite with step; eauto.
Qed.

Theorem close_cong_sigma : forall rho,
     closed_rho rho 
  -> forall A B m,
     close rho (exp_sigma A B) m = exp_sigma (close rho A m) (close rho B (S m)).
Proof.
  induction 1; intros; simpl in *; rewrite_ih; autorewrite with step; eauto.
Qed.

Theorem close_cong_mu : forall rho, closed_rho rho
  -> forall A m,
       close rho (exp_mu A) m
     = exp_mu (close rho A (S m)).
Proof.
  induction 1; intros; simpl in *; rewrite_ih; autorewrite with step; eauto.
Qed.

Hint Rewrite close_cong_mu using (eauto 3; fail) : step.
Hint Rewrite close_cong_arr using (eauto 3; fail) : step.
Hint Rewrite close_cong_sigma using (eauto 3; fail) : step.
Hint Rewrite close_cong_nat close_cong_at close_cong_eq
             close_cong_sum close_cong_star close_cong_join : step.

Theorem close_cong_lam : forall rho,
     closed_rho rho
  -> forall b n,
        close rho (exp_lam b) n
      = exp_lam (close rho b (S n)).
Proof.
  induction 1; intros; simpl; auto; rewrite_ih; autorewrite with step; eauto.
Qed.

Theorem close_cong_nrec : forall rho,
     closed_rho rho
  -> forall b n,
        close rho (exp_nrec b) n
      = exp_nrec (close rho b (S (S n))).
Proof.
  induction 1; intros; simpl; auto; rewrite_ih; autorewrite with step; eauto.
Qed.

Theorem close_cong_app : forall rho a b n,
    close rho (exp_app a b) n
  = exp_app (close rho a n) (close rho b n).
Proof. induction rho; intros; simpl; eauto. Qed.

Theorem close_cong_fun : forall rho,
     closed_rho rho
  -> forall a n,
       close rho (exp_fun a) n
     = exp_fun (close rho a (S (S n))).
Proof.
  induction 1; intros; simpl; eauto; rewrite_ih;
  autorewrite with step; eauto.
Qed.

Theorem close_cong_zero : forall rho n,
    close rho exp_zero n = exp_zero.
Proof. induction rho; intros; simpl; eauto. Qed.

Theorem close_cong_succ : forall rho a n,
    close rho (exp_succ a) n = exp_succ (close rho a n).
Proof. induction rho; intros; simpl; eauto. Qed.

Theorem close_cong_ncase : forall rho,
     closed_rho rho
  -> forall a b1 b2 n,
       close rho (exp_ncase a b1 b2) n
     = exp_ncase (close rho a n)
                 (close rho b1 (S n)) (close rho b2 (S (S n))).
Proof.
  induction 1; intros; simpl; rewrite_ih; autorewrite with step; eauto.
Qed.

Theorem close_cong_inl : forall rho a n,
    close rho (exp_inl a) n = exp_inl (close rho a n).
Proof. induction rho; intros; simpl; eauto. Qed.

Theorem close_cong_inr : forall rho a n,
    close rho (exp_inr a) n = exp_inr (close rho a n).
Proof. induction rho; intros; simpl; eauto. Qed.

Theorem close_cong_pair : forall rho a b n,
    close rho (exp_pair a b) n = exp_pair (close rho a n) (close rho b n).
Proof. induction rho; intros; simpl; eauto. Qed.

Theorem close_cong_scase : forall rho,
     closed_rho rho
  -> forall a b1 b2 n,
       close rho (exp_scase a b1 b2) n
     = exp_scase (close rho a n)
                 (close rho b1 (S (S n))) (close rho b2 (S (S n))).
Proof.
  induction 1; intros; simpl; rewrite_ih; autorewrite with step; eauto.
Qed.

Theorem close_cong_pcase : forall rho,
     closed_rho rho
  -> forall a b n,
       close rho (exp_pcase a b) n
     = exp_pcase (close rho a n) (close rho b (S (S (S n)))).
Proof.
  induction 1; intros; simpl; rewrite_ih; autorewrite with step; eauto.
Qed.


Theorem close_cong_roll : forall rho a n,
   close rho (exp_roll a) n = exp_roll (close rho a n).
Proof.
  induction rho; intros; simpl; eauto.
Qed.

Theorem close_cong_unroll : forall rho a n,
   close rho (exp_unroll a) n = exp_unroll (close rho a n).
Proof.
  induction rho; intros; simpl; eauto.
Qed.

Hint Rewrite close_cong_lam close_cong_nrec close_cong_fun close_cong_app 
             close_cong_zero close_cong_pair
             close_cong_succ close_cong_ncase
             close_cong_inl close_cong_inr close_cong_scase
             close_cong_pcase close_cong_roll close_cong_unroll
       using (eauto 3; fail) : step.


(***************** Other facts about close ***********)
(*
Theorem close_t_weakening : forall rho1,
     closed_rho rho1
  -> forall bd rho2 A,
       close_t (rho1 ++ rho2) A 0
     = close_t (rho1 ++ bd :: rho2) (lift_t A 1 (length rho1)) 0.
Proof.
  induction 1; intros bd ? ?; simpl in *; eauto;
  destruct bd; autorewrite with step; eauto;
  match goal with
  | [ IH : forall _ _ _, _ = _ |- _ ] => 
    erewrite IH; autorewrite with step; eauto
  end.
Qed.
Hint Rewrite <- close_t_weakening using (eauto 3; fail) : step.
*)
Theorem closed_rho_app_1 : forall r1 r2,
  closed_rho (r1 ++ r2) -> closed_rho r1.
Proof.
  induction r1; intros r2 H; inversion H; subst; eauto.
Qed.
Hint Immediate closed_rho_app_1.

Theorem closed_rho_mid : forall r1 v r2,
  closed_rho (r1 ++ v :: r2) -> fvs v = 0.
Proof.
  induction r1; intros v r2 H; inversion H; subst; eauto.
Qed.
Hint Immediate closed_rho_mid.

Theorem closed_rho_app_2 : forall r1 r2,
  closed_rho (r1 ++ r2) -> closed_rho r2.
Proof.
  induction r1; intros r2 H; inversion H; subst; eauto.
Qed.
Hint Immediate closed_rho_app_2.

Theorem closed_rho_app_cons : forall r1 v r2,
  closed_rho (r1 ++ v :: r2) -> closed_rho r2.
Proof.
  intros r1 v r2 H; apply closed_rho_app_2 in H; inversion H; eauto.
Qed.
Hint Immediate closed_rho_app_cons.

Theorem closed_rho_strengthen : forall r1 v r2,
  closed_rho (r1 ++ v :: r2) -> closed_rho (r1 ++ r2).
Proof.
  induction r1; intros v r2 H;
  inversion H; subst; simpl; eauto.
Qed.
Hint Immediate closed_rho_strengthen.

Theorem close_over_fvs_id : forall rho a m,
  fvs a <= m -> close rho a m = a.
Proof.
  induction rho; intros; simpl; eauto;
  autorewrite with step; eauto.
Qed.
Hint Rewrite close_over_fvs_id using (simpl; eauto 2; fail) : step.

Theorem close_closed_id : forall rho a m,
  fvs a = 0 -> close rho a m = a.
Proof.
  induction rho; intros; simpl; eauto;
  rewrite open_closed_id by eauto; autorewrite with step in *; eauto.
Qed.
Hint Rewrite close_closed_id using (eauto 2; fail) : step.

Hint Extern 1 =>
  match goal with
  | [ HTyp : typing nil _ ?a _ |- context [close _ ?a _] ] =>
    rewrite (close_closed_id _ _ _ (typing_fvs_closed_exp _ _ _ HTyp))
  | [ HEq : fvs ?a  = 0 |- context [close _ ?a _] ] =>
    rewrite (close_closed_id _ _ _ HEq)
  end.

Theorem close_open_commute_closed : forall rho,
     closed_rho rho
  -> forall n m a b,
        fvs a = 0
     -> n <= m
     -> close rho ({n ~> a}b) m = {n ~> a}(close rho b (S m)).
Proof.
  intros rho ? n m a b ? ?;
  replace ({n ~> a}b) with ({n ~> close rho a m}b) by auto;
  rewrite close_open_commute; auto.
Qed.
Hint Rewrite close_open_commute_closed using (eauto 2; fail) : step.

Theorem close_app : forall r1 r2 a,
     closed_rho (r1 ++ r2)
  -> close (r1 ++ r2) a 0 = close r1 (close r2 a (length r1)) 0.
Proof.
  induction r1; intros; simpl in *; eauto;
  match goal with
  | [ H : closed_rho (_ :: _) |- _ ] =>
    inversion H; subst; rewrite_ih; autorewrite with step; eauto
  end.
Qed.
Hint Rewrite close_app using (eauto 3; fail) : step.

Theorem close_mid_trm_extract : forall r1 b r2 a,
     closed_rho (r1 ++ close r2 b 0 :: r2)
  ->   close (r1 ++ close r2 b 0 :: r2) a 0
     = close (r1 ++ r2) ({length r1 ~> lift b (length r1) 0} a) 0.
Proof.
  induction r1; intros b r2 A HC;
  match goal with
  | [ H : closed_rho _ |- _ ] => inversion HC; subst; simpl in *
  end; try rewrite_ih;
  [ autorewrite with step; rewrite close_open_commute
  | rewrite open_open_swap, (open_closed_id a),
            lift_lift_collapse by auto
  ]; eauto.
Qed.

(********** facts about the interpretation ***********)
Theorem interp_val : forall k A t rho a,
     interp (k,A,Val) t rho a 
  -> val a.
Proof.
  intros k A t rho a HInterp;
  apply interp_induction with (P := fun k A cv t rho a => cv = Val -> val a)
     in HInterp;
  intros; try match goal with
  | [ H : Comp = Val |- _ ] => inversion H
  | [ H : Val = Comp |- _ ] => inversion H
  end; eauto.
Qed.
Hint Immediate interp_val.

Theorem interp_well_typed : forall k A cv t rho a,
     interp (k,A,cv) t rho a 
  -> typing nil t a (close rho A 0).
Proof.
  apply interp_induction; intros; autorewrite with step in *;
  try match goal with
  | [ HI : interp (_,_,Val) ?t' _ ?a |- typing _ ?t ?a (exp_at _ _) ] =>
    apply interp_val in HI; destruct t; destruct t'
  end; go_go_regularity; mycrush.
Qed.
Hint Immediate interp_well_typed.

Theorem interp_closed : forall k A cv t rho a,
     interp (k, A, cv) t rho a
  -> fvs a = 0.
Proof.
  intros; eauto using typing_fvs_closed_exp.
Qed.
Hint Immediate interp_closed.

(* With interp_closed, we can give eauto a little more help *)
Hint Extern 1 =>
  match goal with
  | [ H : interp _ _ _ ?a |- context [{?n ~> ?b}?a] ] =>
    rewrite (open_closed_id _ _ _ (interp_closed _ _ _ _ _ _ H) n b) 
  | [ H : interp _ _ _ ?a |- context [lift ?a ?n ?m] ] =>
    rewrite (lift_closed_id _ _ _ (interp_closed _ _ _ _ _ _ H) n m)
  | [ HInt : interp (?k,?A,?cv) ?t _ ?a |- context [close _ ?a _] ] =>
    rewrite (close_closed_id _ _ _ (interp_closed _ _ _ _ _ _ HInt))
  end.

(* Downward closure, and some instances that are good for auto *)
Theorem downward_closure : forall k A cv t rho a, 
     interp (k,A,cv) t rho a 
  -> forall j, j < k -> interp (j,A,cv) t rho a.
Proof.
  intros k A cv t rho a HI;
  apply interp_induction 
            with (P := fun k A cv t rho a => forall j, j < k 
                                               -> interp (j,A,cv) t rho a)
        in HI; intros; eauto; clear HI k A cv t a;
  unfold_interp; eauto 6; intros.
  repeat (match goal with
  | [ H : forall j _, j <= _ -> _ |- interp (?j' - ?j0, _,_) _ _ _ ] =>
    apply H with (j := j0); eauto; omega
  | [ H : forall j _, j < _ -> _ |- _ ] => 
    try solve [apply H; eauto; omega]; clear H
  | [ H : forall j _, j <= _ -> _ |- _ ] => 
    try solve [apply H; eauto; omega]; clear H
  end).
Qed.

Hint Extern 4 =>
  match goal with
  | [ H : interp (?k, ?A, ?cv) ?t ?a |- interp (?j, ?A, ?cv) ?t ?a ] =>
    assert (j < k) by omega; apply (downward_closure k)
  end.

Theorem downward_closure_leq : forall k j A cv t rho a,
     interp (k,A,cv) t rho a
  -> j <= k
  -> interp (j,A,cv) t rho a.
Proof.
  intros; match goal with
  | [ Hle : _ <= _ |- _ ] =>
    apply le_lt_or_eq in Hle; destruct Hle;
    subst; eauto using downward_closure
  end.
Qed.
Hint Immediate downward_closure downward_closure_leq.

Theorem downward_closure_minus : forall k j A cv t rho a,
     interp (k,A,cv) t rho a
  -> interp (k - j,A,cv) t rho a.
Proof.
  intros; assert (k - j <= k) by omega; eauto using downward_closure_leq.
Qed.
Hint Immediate downward_closure_minus.

(* V [A] t k  \subseteq  C [A] t k *)
Theorem comp_contains_values' : forall k A cv t rho a,
     interp (k,A,cv) t rho a 
  -> cv = Val
  -> interp (k,A,Comp) t rho a.
Proof.
  intros k A cv t a rho HI;
  apply interp_induction
   with (P := fun k A cv t rho a => cv = Val -> interp (k,A,Comp) t rho a)
     in HI; intros; eauto; clear HI k A cv t a;
  try match goal with
  | [ H : Comp = Val |- _ ] => inversion H
  end; unfold_interp; autorewrite with step in *;
  try match goal with
  | [ |- exists v, exists n, mstep n ?a v /\ interp _ _ _ v ] =>
    exists a; exists 0; split; [ eauto | unfold_interp; eauto ]
  | [ |- forall _ _, _ <= _ -> mstep _ _ _ -> val _ -> interp _ _ _ _ ] =>
    intros; step_val_elim; unfold_interp; autorewrite with step; eauto
  | [ |- typing _ ?t1 _ (exp_at _ ?t2) ] =>
    destruct t1; destruct t2
  end; go_go_regularity; mycrush.
Qed.

Theorem comp_contains_values : forall k A t rho a,
     interp (k,A,Val) t rho a 
  -> interp (k,A,Comp) t rho a.
Proof.
  intros; eauto using comp_contains_values'.
Qed.
Hint Immediate comp_contains_values.

Theorem FO_type_mobile' : forall G t A B,
     typing G t A B -> FO A -> B = exp_star
  -> typing G th_log A exp_star.
Proof.
  induction 1; intros HFO HEq; subst;
  try match goal with
  | [ H : typing _ _ exp_star _ |- _ ] => 
    apply star_has_no_type in H; inversion H
  | [ H : {_ ~> _}_ = exp_star |- _ ] =>
    apply eq_star in H; destruct_things; subst; eauto;
    star_isnt_equal
  | [ H : _ = exp_star |- _ ] => solve [inversion H; subst]
  end; eauto; try solve [inversion HFO; subst; eauto; destruct t; eauto 6].
Qed.

Theorem FO_type_mobile : forall G A,
     typing G th_prog A exp_star 
  -> FO A
  -> typing G th_log A exp_star.
Proof.
  intros G T H HFO; eapply FO_type_mobile' in H; eauto.
Qed.
Hint Immediate FO_type_mobile.

Theorem close_preserves_FO : forall A,
  FO A -> forall rho n, closed_rho rho -> FO (close rho A n).
Proof.
  induction 1; intros; autorewrite with step in *; eauto.
Qed.
Hint Immediate close_preserves_FO.

Hint Extern 2 =>
  match goal with
  | [ HFO : FO ?A, HP : typing ?G th_prog (close ?rho ?A ?n) exp_star
     |- typing ?G th_log (close ?rho ?A ?n) exp_star ] =>
  apply FO_type_mobile
  end.

Theorem log_val_fo_contains_prog : forall k A rho a,
     interp (k,A,Val) th_prog rho a
  -> closed_rho rho
  -> FO A
  -> interp (k,A,Val) th_log rho a.
Proof.
  intros k A a rho HI; 
  remember th_prog as t; revert Heqt;
  remember Val as cv; revert Heqcv;
  apply interp_induction
   with (P := fun k A cv t rho a =>  
                 cv = Val -> t = th_prog -> closed_rho rho -> FO A 
              -> interp (k,A,cv) th_log rho a)
     in HI; intros; subst; try match goal with
  | [ H : FO exp_star |- _ ] => inversion H
  end; eauto; clear HI k A cv t rho a; subst;
  try match goal with
  | [ H : th_log = th_prog |- _ ] => inversion H
  | [ H : Comp = Val |- _ ] => inversion H
  | [ H : FO _ |- _ ] => inversion H; subst; clear H
  end; unfold_interp; eauto;
  try match goal with
  | [ HI : interp _ _ (?b1 :: _) ?b2
     |- typing _ _ (exp_pair ?b1 ?b2) _ ] =>
    assert (HWT := interp_well_typed _ _ _ _ _ _ HI); mycrush
  end; autorewrite with step in *; go_go_regularity; eauto 7.
Qed.
Hint Extern 2 =>
  match goal with
  | [ H : FO ?A |- interp (_,?A,Val) th_log _ _ ] =>
    apply log_val_fo_contains_prog
  end.

Hint Extern 0 => 
match goal with
| [ |- FO (close ?rho ({0 ~> ?b}?B) 0) ] =>
  replace (close rho ({0 ~> b}B) 0) with ({0 ~> b} (close rho B 1))
  by (autorewrite with step; auto)
end.

Theorem log_val_fo_contains_prog_closed : forall k A rho a,
     interp (k,A,Val) th_prog rho a
  -> closed_rho rho
  -> FO (close rho A 0)
  -> interp (k,A,Val) th_log rho a.
Proof.
  intros k A a rho HI; 
  remember th_prog as t; revert Heqt;
  remember Val as cv; revert Heqcv;
  apply interp_induction
   with (P := fun k A cv t rho a =>  
                 cv = Val -> t = th_prog -> closed_rho rho -> FO (close rho A 0) 
              -> interp (k,A,cv) th_log rho a)
     in HI; intros; subst; eauto; clear HI k A cv t rho a; subst;
  try match goal with
  | [ H : th_log = th_prog |- _ ] => inversion H
  | [ H : Comp = Val |- _ ] => inversion H
  end; unfold_interp;
  match goal with
  | [ H : FO _ |- _ ] => 
    autorewrite with step in *; inversion H; subst;
    go_go_regularity; eauto 7
  end.
Qed.
Hint Immediate log_val_fo_contains_prog_closed.

(* [A] L k  \subseteq  [A] P k *)
Theorem prog_contains_log : forall k A cv rho a,
     interp (k,A,cv) th_log rho a
  -> closed_rho rho
  -> interp (k,A,cv) th_prog rho a.
Proof.
  intros k A cv a rho HI; remember th_log as t; revert Heqt;
  apply interp_induction
   with (P := fun k A cv t rho a => 
              t = th_log -> closed_rho rho
              -> interp (k,A,cv) th_prog rho a)
     in HI; intros; eauto; clear HI k A cv t rho a; subst;
  repeat (match goal with
  | [ H : th_prog = th_log |- _ ] => inversion H
  | [ HC : closed_rho ?rho,
      H : th_log = th_log -> closed_rho ?rho -> _ |- _ ] =>
    let H' := fresh H' in
    assert (H' := H eq_refl HC); clear H; rename H' into H
  end); unfold_interp; eauto; intros; match goal with
  | [ HAS : mstep _ ?a ?v, HMS : mstep _ ?a ?v' |- _ ] =>
    assert (v = v') 
       by (subst; eauto using mstep_val_deterministic_exp);
    subst; eauto
  | [ H : typing _ _ (exp_lam _) (close _ (exp_arr _ _) _) |- _ ] =>
    autorewrite with step in H; go_go_regularity; eauto
  | [ H : typing _ _ (exp_nrec _) (close _ (exp_arr _ _) _) |- _ ] =>
    autorewrite with step in H; go_go_regularity; eauto
  end.
Qed.
Hint Immediate prog_contains_log.
Hint Extern 2 =>
  match goal with
  | [ HI : interp ?p th_log ?rho ?a |- interp ?p ?t ?rho ?a ] =>
    destruct t
  end.


(**************** Well formed rhos ******************)
Inductive env_wf (k : nat) : ctx -> env -> Prop :=
  | EWFNil  : env_wf k nil nil
  | EWFCons : forall G rho t A a,
       interp (k,A,Val) t rho a
    -> env_wf k G rho
    -> typing G t A exp_star
    -> env_wf k ((t,A) :: G) (a :: rho).
Hint Constructors env_wf.

Theorem env_wf_closed : forall k G rho,
  env_wf k G rho -> closed_rho rho.
Proof.
  induction 1; eauto.
Qed.
Hint Resolve env_wf_closed.

Lemma env_wf_app_2 : forall G1 G2 r1 r2 k,
     env_wf k (G1 ++ G2) (r1 ++ r2)
  -> length G1 = length r1
  -> env_wf k G2 r2.
Proof.
  induction G1; intros G2 r1 r2 k HE HL; inversion HE;
  destruct r1; inversion HL; subst; eauto;
  match goal with
  | [ H : _ :: _ = (_ :: _) ++ _ |- _ ] =>
    inversion H; subst; eauto
  end.
Qed.

Theorem env_substitution : forall rho G2 k,
     env_wf k G2 rho
  -> forall G1 t a A, typing (G1 ++ G2) t a A
          -> typing (close_c rho G1 0) t (close rho a (length G1))
                    (close rho A (length G1)).
Proof.
  induction 1; intros; simpl; eauto;
  try match goal with
  | [ HTyp : typing (?G ++ nil) _ _ _ |- typing ?G _ _ _ ] =>
    rewrite app_nil_r in HTyp; auto
  end; match goal with
  | [ HTyp : typing (?G1 ++ ?b :: ?G2) _ _ _,
      IH : forall _ _ _ _, _ -> _ |- _ ] =>
    replace (G1 ++ b :: G2) with ((G1 ++ (b :: nil)) ++ G2)
         in HTyp by (rewrite <- app_assoc; simpl; auto); apply IH in HTyp;
    match goal with
    | [ H : interp _ _ _ ?a |- _ ] =>
      erewrite close_c_app, close_c_cons, close_c_nil in HTyp by eauto;
      simpl in HTyp;
      apply typing_substitution_end_closed with (v := a) in HTyp; eauto;
      erewrite close_c_open_e_commute, close_open_commute by eauto
    end;
    autorewrite with step in HTyp; simpl in HTyp; 
    autorewrite with step; eauto
  end.
Qed.
Hint Resolve env_substitution.

Theorem env_substitution_cons : forall k G rho t' A' t a A,
     env_wf k G rho
  -> typing ((t',A') :: G) t a A
  -> typing ((t',close rho A' 0) :: nil) t 
            (close rho a 1) (close rho A 1).
Proof.
  intros; match goal with
  | [ |- typing ((?t',close ?rho ?A' 0) :: nil) _ _ _ ] =>
    replace ((t',close rho A' 0) :: nil) 
       with (close_c rho ((t',A') :: nil) 0) 
         by (erewrite close_c_cons, close_c_nil by eauto; simpl; auto)
  end; eapply env_substitution; eauto.
Qed.
Hint Resolve env_substitution_cons.

Theorem env_substitution_nil : forall k G rho t a A,
     env_wf k G rho
  -> typing G t a A
  -> typing nil t (close rho a 0) (close rho A 0).
Proof.
  intros; erewrite <- close_c_nil; eapply env_substitution; eauto.
Qed.
Hint Resolve env_substitution_nil.

Theorem env_substitution_typ : forall rho G2 k,
     env_wf k G2 rho
  -> forall G1 t A, typing (G1 ++ G2) t A exp_star
        -> typing (close_c rho G1 0) t (close rho A (length G1)) exp_star.
Proof.
  intros;
  replace exp_star with (close rho exp_star (length G1))
       by (autorewrite with step; auto);
  eapply env_substitution; eauto.
Qed.
Hint Resolve env_substitution_typ.

Theorem env_substitution_typ_all : forall rho G k t A,
    env_wf k G rho -> typing G t A exp_star
 -> typing nil t (close rho A 0) exp_star.
Proof.  
  intros rho G k t A HE HT;
  replace G with (nil ++ G) in HT by auto;
  eapply env_substitution_typ in HT; autorewrite with step in *; eauto.
Qed.
Hint Resolve env_substitution_typ_all.

Theorem env_substitution_all_arr : forall rho G k t A1 A2,
      env_wf k G rho -> typing G t (exp_arr A1 A2) exp_star
   -> typing nil t (exp_arr (close rho A1 0) (close rho A2 1)) exp_star.
Proof.  
  intros rho G k t A1 A2 HE HT;
  eapply env_substitution_typ_all in HT; eauto;
  autorewrite with step in *; eauto.
Qed.
Hint Resolve env_substitution_all_arr.


Theorem env_wf_downward_closure : forall k G rho,
  env_wf k G rho -> forall j, j < k -> env_wf j G rho.
Proof.
  induction 1; intros; eauto.
Qed.
Hint Immediate env_wf_downward_closure.

Theorem env_wf_downward_closure_leq : forall k j G rho,
     env_wf k G rho
  -> j <= k
  -> env_wf j G rho.
Proof.
  intros; match goal with
  | [ Hle : _ <= _ |- _ ] =>
    apply le_lt_or_eq in Hle; destruct Hle;
    subst; eauto
  end.
Qed.
Hint Immediate env_wf_downward_closure_leq.

Theorem env_wf_downward_closure_minus : forall k j G rho,
     env_wf k G rho
  -> env_wf (k - j) G rho.
Proof.
  intros; assert (k - j <= k) by omega; eauto.
Qed.
Hint Immediate env_wf_downward_closure_minus.

Theorem close_preserves_val : forall rho v m,
     closed_rho rho
  -> val v
  -> val (close rho v m).
Proof.
  induction rho; intros; simpl; eauto;
  match goal with
  | [ HTI : closed_rho (_ :: _) |- _ ] =>
    inversion HTI; subst; eauto
  end.
Qed.
Hint Resolve close_preserves_val.

Hint Extern 3 =>
  match goal with
  | [ IH : forall _ _, _ <= ?k -> mstep _ ?a _ -> _ 
                               -> interp (?k - _,_,_) _ _ _
     |- interp (?k,_,_) _ _ ?a ] =>
  replace k with (k - 0) by omega
  end.


Theorem comp_interp_of_val : forall m A t a rho,
     val a
  -> interp (m, A, Comp) t rho a
  -> interp (m, A, Val) t rho a.
Proof.
  intros; unfold_interp_hyps; try step_val_elim; eauto.
Qed.
Hint Extern 4 => 
  match goal with
  | [ H : val ?a |- interp (_,_,Val) _ _ ?a ] =>
    apply (comp_interp_of_val _ _ _ _ _ H)
  | [ |- interp (_,_,Val) _ _ (exp_fun ?a) ] =>
    let HVal := fresh "HVal" in
    assert (val (exp_fun a)) as HVal by eauto;
    apply (comp_interp_of_val _ _ _ _ _ HVal)
  end.

(*** Some tactics for instantiating the interpretation *)
Ltac apply_close_congs :=
  repeat (rewrite close_cong_nat in *); repeat (rewrite close_cong_at in *);
  repeat (rewrite close_cong_eq in *); repeat (rewrite close_cong_app in *);
  repeat (rewrite close_cong_zero in *);
  repeat (rewrite close_cong_star in *); repeat (rewrite close_cong_join in *);
  repeat (rewrite close_cong_succ in *); repeat (rewrite close_cong_roll in *);
  repeat (rewrite close_cong_inl in *); repeat (rewrite close_cong_inr in *);
  repeat (rewrite close_cong_pair in *);
  repeat (rewrite close_cong_unroll in *); repeat (rewrite close_cong_sum in *);
  repeat (match goal with
  | [ H : env_wf _ _ ?rho |- _ ] =>
    notProvenYet (closed_rho rho); assert (closed_rho rho) by eauto
  end);
  repeat (rewrite close_cong_fun in * by assumption);
  repeat (rewrite close_cong_ncase in * by assumption);
  repeat (rewrite close_cong_scase in * by assumption);
  repeat (rewrite close_cong_pcase in * by assumption);
  repeat (rewrite close_cong_lam in * by assumption);
  repeat (rewrite close_cong_nrec in * by assumption);
  repeat (rewrite close_cong_arr in * by assumption);
  repeat (rewrite close_cong_sigma in * by assumption);
  repeat (rewrite close_cong_mu in * by assumption);
  repeat (rewrite close_app in * by eauto 2);
  repeat (rewrite close_app in * by eauto 2).

Ltac interp_rewrite :=
  apply_close_congs;
  try match goal with
  | [ H : context [ exp_fun (lift ?a _ 2) ] |- _ ] =>
    rewrite lift_closed_id_gen in H 
      by (assert (fvs (exp_fun a) = 0) as HFOO by eauto;
          simpl in HFOO; omega)
  | [ H : context [ exp_nrec (lift ?a _ 2) ] |- _ ] =>
    rewrite lift_closed_id_gen in H 
      by (assert (fvs (exp_nrec a) = 0) as HFOO by eauto;
          simpl in HFOO; omega)
  end;
  repeat (rewrite close_open_commute in * by auto 2);
  apply_close_congs.

Ltac inst_interp :=
  interp_rewrite; match goal with
  | [ H1 : interp (?j, ?A, Val) ?t ?rho ?v,
      H2 : forall j _, j <= ?k -> interp (_, ?A, Val) ?t ?rho _ -> _ |- _ ] =>
    let Hle := fresh "Hle" in let HI := fresh "HI" in
    assert (j <= k) as Hle by eauto;
    assert (HI := H2 _ _ Hle H1); clear Hle H2; unfold_interp_hyps
  | [ H1 : mstep ?j ?a ?b,
      H2 : forall j _, j <= ?k -> mstep j ?a _ -> val _ -> interp _ _ _ _
     |- _ ] =>
    let Hle := fresh "Hle" in let HI := fresh "HI" in
    let HV := fresh "HV" in
    assert (j <= k) as Hle by eauto; assert (val b) as HV by eauto;
    assert (HI := H2 _ _ Hle H1 HV); clear Hle H2 HV; unfold_interp_hyps
  end.

(**** I can't prove weakening with the nicer derived induction principle
  from before.  So, I use this less-convenient one instead *)
Theorem index_size_combined_ind : forall (P : nat -> exp -> Prop),
     (forall n A, (forall n' A', (n' < n \/ (n' <= n /\ typ_size A' < typ_size A)) 
                                 -> P n' A')
               -> P n A)
  -> forall n A, P n A.
Proof.
  intros; revert A;
  induction n as [k IHWfk] using lt_wf_ind;
  intros A; remember (typ_size A) as sA; revert A HeqsA;
  induction sA as [sA IHWfsA] using lt_wf_ind; intros A Hsize; subst;
  apply H; intros n' A' HS; destruct_things; eauto;
  match goal with
  | [ Hle : _ <= _ |- _ ] => 
    apply le_lt_or_eq in Hle; destruct Hle; subst; eauto
  end.
Qed.

(* Computation at the type level is carefully controlled in this
   system (see for example restrictions on sorts in the Contra and
   Conversion rules).  As a result, we can prove the canonical forms
   lemma for Star early.  We need this to reason about our
   interpretation, since we didn't use the more complicated techniques
   needed to handle the extra lambdas from the (Star,Box) and
   (Box,Box) rules.  The other canonical forms lemmas can't be proved
   until after the normalization result. *)
Theorem canonical_forms_star' : forall G t A C,
     typing G t A C -> C = exp_star
  -> (   (exists G1, exists G2, exists t',
              G = (G1 ++ (t',exp_star) :: G2)
           /\ A = exp_var (length G1))
      \/ (A = exp_nat)
      \/ (exists B1, exists B2, A = exp_arr B1 B2)
      \/ (exists B, exists t, A = exp_at B t)
      \/ (exists B1, exists B2, A = exp_sum B1 B2)
      \/ (exists B1, exists B2, A = exp_sigma B1 B2)
      \/ (exists B, A = exp_mu B)
      \/ (exists b1, exists b2, A = exp_eq b1 b2)).
Proof.
  induction 1; intros; subst; eauto;
  try solve [match goal with
  | [ H : ?A = exp_star, HT : typing _ _ ?A _ |- _ ] =>
    rewrite H in HT; auto
  | [ H : lift ?A _ _ = exp_star |- _ ] =>
    destruct A; simpl in H; inversion H; eauto 6
  | [ H : _ = exp_star |- _ ] => inversion H
  end]; eauto 10.
Qed.

Theorem canonical_forms_star : forall G t A,
     typing G t A exp_star
  -> (   (exists G1, exists G2, exists t',
              G = (G1 ++ (t',exp_star) :: G2)
           /\ A = exp_var (length G1))
      \/ (A = exp_nat)
      \/ (exists B1, exists B2, A = exp_arr B1 B2)
      \/ (exists B, exists t, A = exp_at B t)
      \/ (exists B1, exists B2, A = exp_sum B1 B2)
      \/ (exists B1, exists B2, A = exp_sigma B1 B2)
      \/ (exists B, A = exp_mu B)
      \/ (exists b1, exists b2, A = exp_eq b1 b2)).
Proof.  eauto using canonical_forms_star'.  Qed.

Theorem no_star_in_env : forall k G1 t G2 rho,
     env_wf k (G1 ++ (t,exp_star) :: G2) rho
  -> False.
Proof.
  intros k G1; induction G1; intros t G2 rho H; inversion H; subst; eauto.
Qed.

Hint Extern 0 =>
  match goal with
  | [ H : env_wf _ (_ ++ (_,exp_star) :: _) _ |- _ ] => apply no_star_in_env in H; inversion H
  | [ H : env_wf _ ?G _, HEq : ?G = _ ++ (_,exp_star) :: _ |- _ ] => 
    rewrite HEq in H; apply no_star_in_env in H; inversion H
  end.

Theorem close_preserves_mpstep : forall rho,
     closed_rho rho
  -> forall n a b,
         mpstep a b
      -> mpstep (close rho a n) (close rho b n).
Proof.
  induction 1; intros; simpl; eauto.
Qed.
Hint Immediate close_preserves_mpstep.

Section InterpShift.
  Hint Extern 2 =>
    match goal with
    | [ H : typing ?G _ ?a (close (_ ++ _ :: ?r) ?A 0)
       |- typing ?G _ ?a (close (_ ++ ?r) ({_ ~> _} ?A) 0) ] =>
      rewrite close_mid_trm_extract in H by eauto
    | [ H : typing ?G _ ?a (close (_ ++ ?r) ({_ ~> _} ?A) 0)
       |- typing ?G _ ?a (close (_ ++ _ :: ?r) ?A 0) ] =>
      rewrite close_mid_trm_extract by eauto
    | [ H : typing ?G _ (close (_ ++ _ :: ?r) ?A 0) ?B
       |- typing ?G _ (close (_ ++ ?r) ({_ ~> _} ?A) 0) ?B ] =>
      rewrite close_mid_trm_extract in H by eauto
    | [ H : typing ?G _ (close (_ ++ ?r) ({_ ~> _} ?A) 0) ?B
       |- typing ?G _ (close (_ ++ _ :: ?r) ?A 0) ?B ] =>
      rewrite close_mid_trm_extract by eauto
    | [ H : typing ?G _ (close (_ ++ _ :: ?r) (?P _ _) 0) ?B
       |- typing ?G _ (close (_ ++ ?r) (?P _ _) 0) ?B ] =>
      rewrite close_mid_trm_extract in H by eauto
    | [ H : typing ?G _ (close (_ ++ ?r) (?P _ _) 0) ?B
       |- typing ?G _ (close (_ ++ _ :: ?r) (?P _ _) 0) ?B ] =>
      rewrite close_mid_trm_extract by eauto
    | [ H : typing ?G _ ?a (close (_ ++ _ :: ?r) (?P _ _) 0)
       |- typing ?G _ ?a (close (_ ++ ?r) (?P _ _) 0) ] =>
      rewrite close_mid_trm_extract in H by eauto
    | [ H : typing ?G _ ?a (close (_ ++ ?r) (?P _ _) 0)
       |- typing ?G _ ?a (close (_ ++ _ :: ?r) (?P _ _) 0) ] =>
       rewrite close_mid_trm_extract by eauto
    end.

  Hint Extern 2 (typing _ _ _ _) => simpl;
    try (rewrite close_cong_nat in * );
    try extract_open_lift.

  Theorem TConv_full : forall G p t a A B,
      typing G th_log p (exp_eq A B)
   -> typing G t a A
   -> typing G t B exp_star
   -> typing G t a B.
  Proof. intros; eauto. Qed.
  Hint Immediate TConv_full.
  
  Theorem TConv_full_joinable : forall G t a A1 A2 B,
      typing G t A1 exp_star
   -> typing G t A2 exp_star
   -> mpstep A1 B -> mpstep A2 B
   -> typing G t a A1
   -> typing G t a A2.
  Proof.
    intros;
    assert (typing G th_log exp_join (exp_eq A1 A2)) by eauto; eauto.
  Qed.
  Hint Immediate TConv_full_joinable.

  Theorem interp_rho_shift_trm : forall k A, 
        (forall cv t r1 b r2 a G1 t' B G2,
              interp (k, A, cv) t (r1 ++ (close r2 b 0) :: r2) a
           -> length r1 = length G1
           -> env_wf k (G1 ++ (t',B) :: G2) (r1 ++ (close r2 b 0) :: r2)
           -> typing (G1 ++ (t',B) :: G2) t A exp_star
           -> interp (k, {length r1 ~> lift b (length r1) 0}A, cv) t (r1 ++ r2) a)
    /\  (forall cv t r1 b r2 a G1 t' B G2,
              interp (k, {length r1 ~> lift b (length r1) 0}A, cv) t (r1 ++ r2) a
           -> length r1 = length G1
           -> env_wf k (G1 ++ (t',B) :: G2) (r1 ++ (close r2 b 0) :: r2)
           -> typing (G1 ++ (t',B) :: G2) t A exp_star
           -> interp (k, A, cv) t (r1 ++ (close r2 b 0) :: r2) a).
  Proof.
    apply index_size_combined_ind; intros k A IH;
    split; intros cv t r1 b r2 a G1 t' B G2 HI HL HE HT; subst; destruct cv;
    try match goal with
    | [ |- interp (_, _, Comp) _ _ _ ] =>
      unfold_interp_hyps; unfold_interp; intros; try inst_interp;
      try match goal with
      | [ H : mstep ?n ?a ?v |- exists v, exists n, mstep n ?a v /\ _ ] =>
        exists v; exists n; split; [ auto | ]
      end
    end; destruct A; simpl in *; try match goal with
    | [ H : typing _ _ exp_star _ |- _ ] => apply star_has_no_type in H; inversion H
    end; unfold_interp_hyps; unfold_interp; intros;
    try match goal with
    | [ HT : typing (_ ++ _ :: _) _ (exp_var _) exp_star |- _ ] =>
      (* var case 2 *)
      apply canonical_forms_star in HT; destruct_things; match goal with 
      | [ H : exp_var _ = _ |- _ ] => inversion H; subst; auto
      end
    | [ |- looks_like_nat _ ] => (* nat case *) eauto
    | [ H : interp (_, {_ ~> _}?A1, Val) ?t _ ?v,
        HW : typing (?G1' ++ _ :: _) ?t' (exp_arr ?A1 ?A2) exp_star
       |- interp (?j, {_ ~> _}?A2, Comp) _ 
                 (?v :: ?r1 ++ _) ({_ ~> ?v}_) ] =>
      (* arr case 1 *)
      let H1 := fresh "H1" in let H2 := fresh "H2" in
      destruct (IH j A2) as [H2 _]; auto; destruct (IH j A1) as [_ H1]; auto;
      clear IH; autorewrite with step;
      go_go_regularity; rewrite app_comm_cons in *;
      replace (S (length r1)) with (length (v :: r1)) by auto;
      eapply H2 with (G1 := (t',A1) :: G1'); eauto; simpl;
      eauto using env_wf_downward_closure
    | [ H : interp (_, ?A1, Val) ?t (_ ++ _ :: _) ?v,
        HW : typing (?G1' ++ _ :: _) ?t' (exp_arr ?A1 ?A2) exp_star,
        HI : forall j _, ?R j _ -> interp _ _ _ _ -> interp _ _ _ _
       |- interp (?j, ?A2, Comp) _ 
                 (?v :: ?r1 ++ _ :: _) ({_ ~> ?v}_) ] =>
      (* arr case 2 *)
      let H1 := fresh "H1" in let H2 := fresh "H2" in
      destruct (IH j A2) as [_ H2]; auto; destruct (IH j A1) as [H1 _]; auto;
      clear IH; go_go_regularity; rewrite app_comm_cons in *; autorewrite with step in HI;
      eapply H2 with (G1 := (t',A1) :: G1'); simpl;
      eauto using env_wf_downward_closure 
    | [ HTv : typing _ ?t (exp_pair ?v1 ?v2) _,
        HTS : typing (?G1' ++ _) ?t (exp_sigma ?A1 ?A2) exp_star
       |- interp (?j, {S (length ?r1) ~> _} ?A2, _) _ (?v1 :: ?r1 ++ _) ?v2 ] =>
      (* second pair component, case 1 *)
      autorewrite with step; let H2 := fresh "H2" in
      replace (S (length r1)) with (length (v1 :: r1)) by auto;
      destruct (IH j A2) as [H2 _]; eauto; rewrite app_comm_cons;
      eapply H2 with (G1 := (t,A1)::G1'); simpl; eauto
    | [ HTv : typing _ ?t (exp_pair ?v1 ?v2) _,
        HTS : typing (?G1' ++ _) ?t (exp_sigma ?A1 ?A2) exp_star
       |- interp (?j, ?A2, _) _ (?v1 :: ?r1 ++ ?rrest) ?v2 ] =>
      (* second pair component, case 2 *)
      assert (interp (j, A1, Val) t (r1 ++ rrest) v1)
        by (go_go_regularity; destruct (IH j A1); eauto);
      autorewrite with step in *; let H2 := fresh "H2" in
      destruct (IH j A2) as [_ H2]; [auto | ]; rewrite app_comm_cons;
      eapply H2 with (G1 := (t,A1)::G1'); simpl; eauto
    | [ H : interp (?k, ?A, Val) _ _ ?v,
        HT : typing _ _ (exp_at ?A _) exp_star
       |- interp (?k, {_ ~> _}?A, Val) _ _ ?v ] =>
      (* At case 1 *)
      destruct (IH k A) as [? _]; clear IH; go_go_regularity; eauto
    | [ H : interp (?k, {_ ~> _}?A, Val) _ _ ?v,
        HT : typing _ _ (exp_at ?A _) exp_star
       |- interp (?k, ?A, Val) _ _ ?v ] =>
      (* At case 2 *)
      destruct (IH k A) as [_ ?]; clear IH; go_go_regularity; eauto
    | [ H : typing _ _ (exp_mu ?A) exp_star, Hlt : ?j < _
       |- interp (?j,_,_) _ _ ?v ] =>
      (* both mu cases *)
      let IH1 := fresh IH in let IH2 := fresh IH in
      extract_open_lift; rewrite <- open_open_swap in * by auto;
      destruct (IH j ({0 ~> exp_mu A} A)) as [IH1 IH2]; clear IH; auto;
      go_go_regularity; match goal with
      | [ |- interp _ _ (_ ++ _ :: _) _ ] => eapply IH2; eauto using env_wf_downward_closure
      | [ |- interp _ _ (_ ++ _) _ ] => eapply IH1; eauto using env_wf_downward_closure
      end
    | [ |- exists _, mpstep _ _ /\ mpstep _ _ ] =>
      (* both Eq cases *)
      repeat (rewrite close_mid_trm_extract in * by eauto); eauto
    | [ |- interp (?j, {_ ~> _}?A, _) _ _ _ ] =>
      (* sum case 1 *)
      go_go_regularity; destruct (IH j A) as [? _]; eauto
    | [ |- interp (?j, ?A, _) _ (_ ++ _ :: _) _ ] =>
      (* sum case 2 *)
      go_go_regularity; destruct (IH j A) as [_ ?]; eauto
    end. 
  Qed.

  Theorem interp_rho_shift_in_cons : forall k A cv t b G t' B rho a,
       env_wf k ((t',B) :: G) ((close rho b 0) :: rho)
    -> typing ((t',B) :: G) t A exp_star
    -> interp (k, A, cv) t ((close rho b 0) :: rho) a
    -> interp (k, {0 ~> b}A, cv) t rho a.
  Proof.
    intros k A cv t b G t' B rho a HE HT HI;
    edestruct (interp_rho_shift_trm k A) as [? _];
    replace rho with (nil ++ rho) by auto;
    replace 0 with (length (@nil exp)) by auto;
    replace b with (lift b (length (@nil exp)) 0)
         by (autorewrite with step; auto);
    eapply H with (G1 := nil); simpl; eauto.
  Qed.

  Theorem interp_rho_shift_out_cons : forall k A cv t b G t' B rho a,
       env_wf k ((t',B) :: G) ((close rho b 0) :: rho)
    -> typing ((t',B) :: G) t A exp_star
    -> interp (k, {0 ~> b}A, cv) t rho a
    -> interp (k, A, cv) t ((close rho b 0) :: rho) a.
  Proof.
    intros k A cv t b G t' B rho a HE HT HI;
    edestruct (interp_rho_shift_trm k A) as [_ HShift];
    replace rho with (nil ++ rho) in HI by auto;
    replace 0 with (length (@nil exp)) in HI by auto;
    replace b with (lift b (length (@nil exp)) 0) in HI
         by (autorewrite with step; auto);
    eapply HShift with (G1 := nil) (t' := t') (G2 := G) in HI; eauto.
  Qed.
End InterpShift.


Section ConvSwap.
  Hint Extern 2 (typing _ _ _ _) => try extract_open_lift.


  Hint Extern 0 =>
    match goal with
    | [ HT : typing ?G _ ?a (close ?rho ?B1' ?n),
        HS1 : mpstep (close ?rho ?B1' ?n) ?C,
        HS2 : mpstep (close ?rho ?B2' ?n) ?C
       |- typing ?G _ ?a (close ?rho ?B2' ?n) ] =>
      eapply TConv_full_joinable with (A1 := close rho B1' n) (B := C)
    end.
  (***************************************************************************)
  (***************************************************************************)

  (***************************************************************************)
  (***************************************************************************)

  Theorem typing_fvs_closed_typ : forall t a A,
    typing nil t a A -> fvs A = 0.
  Proof.
    intros t a A HT; apply regularity in HT; destruct_things; subst; eauto.
  Qed.
  Hint Immediate typing_fvs_closed_typ.

  Theorem mpstep_preservation : forall G t a a' A,
    mpstep a a' -> typing G t a A -> typing G t a' A.
  Proof.
    induction 1; intros; eauto using pstep_preservation.
  Qed.
  Hint Immediate mpstep_preservation.


  Theorem typing_weakening_closed_type : forall G t a,
       typing nil t a exp_star
    -> ctx_wf G
    -> typing G t a exp_star.
  Proof.
    intros G t a HT HC;
    replace G with (lift_c nil (length G) 0 ++ G ++ nil)
                by (simpl; rewrite app_nil_r; auto);
    replace a with (lift a (length G) 0) by (autorewrite with step; eauto);
    replace exp_star with (lift exp_star (length G) 0) by eauto;
    apply typing_weakening; try rewrite app_nil_r; eauto.
  Qed.

  Theorem typing_weakening_mostly_closed_type : forall t' A G t a,
       typing ((t',A)::nil) t a exp_star
    -> fvs A = 0
    -> ctx_wf G
    -> typing ((t',A)::G) t a exp_star.
  Proof.
    intros t' A G t a HT HFV HCWF;
    replace ((t',A)::G) with ((lift_c ((t',A)::nil) (length G)) 0 ++ G ++ nil)
                by (simpl; rewrite app_nil_r; autorewrite with step; auto);
    replace a with (lift a (length G) (length ((t',A)::nil)))
                by (apply typing_fvs in HT; simpl in *;
                    rewrite lift_closed_id_gen; auto);
    replace exp_star with (lift exp_star (length G) (length ((t',A)::nil)))
                       by eauto;
    apply typing_weakening; try rewrite app_nil_r; simpl; eauto.
  Qed.


  Hint Extern 0 =>
  match goal with
  | [ HT : typing nil ?t ?a exp_star |- typing ?G ?t ?a exp_star ] =>
    assert (ctx_wf G) by eauto;
    apply typing_weakening_closed_type
  | [ HT : typing ((_,?A)::nil) ?t ?a exp_star
     |- typing ((_,?A)::?G) ?t ?a exp_star ] =>
    assert (ctx_wf G) by eauto; assert (fvs A = 0) by eauto;
    apply typing_weakening_mostly_closed_type
  end.

  Theorem ts_sub' : forall G t B C,
       typing G t B C
    -> forall G1 t' A G2 v,
            C = exp_star -> G = G1 ++ (t',A) :: G2
         -> typing G2 t' A exp_star
         -> typ_size ({length G1 ~> v}B) = typ_size B.
  Proof.
    induction 1; intros; subst; eauto;
    try match goal with
    | [ H : _ = exp_star |- _ ] => inversion H
    end; simpl; eauto;
    try match goal with
    | [ |- context[nat_compare _ _] ] =>
      db_var_case; invert_list_equalities; match goal with
      | [ HL : _ :: _ = _ :: _ |- _ ] => inversion HL; subst;
        match goal with | [ H : lift ?A _ _ = exp_star |- _ ] => 
          destruct A; simpl in H; inversion H; subst; auto
        end
      end
    | [ H : {_ ~> _}_ = exp_star |- _ ] => 
      go_go_regularity; apply eq_star in H; destruct_things; subst; eauto
    | [ H : typing ((?t,?A) :: ?G ++ _ :: _) _ _ _
       |- context[{S (length ?G) ~> _}_] ] =>
      replace (S (length G)) with (length ((t,A) :: G)) by auto;
      rewrite app_comm_cons in *; eauto
    end.
  Qed.

  Theorem ts_sub : forall t t' A G B v,
       typing ((t,A)::G) t' B exp_star
    -> typing G t A exp_star
    -> typ_size ({0 ~> v}B) = typ_size B.
  Proof.
    intros t t' A G B v HT HA.
    replace 0 with (length (@nil exp)) by auto;
    eapply (ts_sub' _ _ _ _ HT nil t A G); eauto.
  Qed.

  Theorem wf_star_does_not_step_to_var : forall G1 G2 t B r1 r2 n k,
       typing (G1 ++ G2) t B exp_star
    -> mpstep (close r2 B (length r1)) (exp_var n)
    -> env_wf k (G1 ++ G2) (r1 ++ r2)
    -> False.
  Proof.
    intros G1 G2 t B r1 r2 n k HT HMS HWF;
    assert (closed_rho (r1 ++ r2)) by eauto; assert (closed_rho r2) by eauto;
    assert (HCF := canonical_forms_star _ _ _ HT);
    destruct_things; subst;
    try solve [autorewrite with step in HMS; mpstep_inversion];
    auto.
  Qed.

  Theorem wf_star_does_not_step_to_star : forall G1 G2 t B r1 r2 k,
       typing (G1 ++ G2) t B exp_star
    -> mpstep (close r2 B (length r1)) exp_star
    -> env_wf k (G1 ++ G2) (r1 ++ r2)
    -> False.
  Proof.
    intros G1 G2 t B r1 r2 k HT HMS HWF;
    assert (closed_rho (r1 ++ r2)) by eauto; assert (closed_rho r2) by eauto;
    assert (HCF := canonical_forms_star _ _ _ HT);
    destruct_things; subst;
    try solve [autorewrite with step in HMS; mpstep_inversion];
    auto.
  Qed.

  Theorem wf_star_steps_to_arr : forall G1 G2 t B r1 r2 A1 A2 k,
       typing (G1 ++ G2) t B exp_star
    -> mpstep (close r2 B (length r1)) (exp_arr A1 A2)
    -> env_wf k (G1 ++ G2) (r1 ++ r2)
    -> exists B1, exists B2,
            B = exp_arr B1 B2 /\ mpstep (close r2 B1 (length r1)) A1
         /\ mpstep (close r2 B2 (S (length r1))) A2.
  Proof.
    intros G1 G2 t B r1 r2 A1 A2 k HT HMS HWF;
    assert (closed_rho (r1 ++ r2)) by eauto; assert (closed_rho r2) by eauto;
    assert (HCF := canonical_forms_star _ _ _ HT);
    destruct_things; subst;
    autorewrite with step in HMS; mpstep_inversion; eauto 8.
  Qed.

  Theorem wf_star_does_not_step_to_fun : forall G1 G2 t B r1 r2 b k,
       typing (G1 ++ G2) t B exp_star
    -> mpstep (close r2 B (length r1)) (exp_fun b)
    -> env_wf k (G1 ++ G2) (r1 ++ r2)
    -> False.
  Proof.
    intros G1 G2 t B r1 r2 b k HT HMS HWF;
    assert (closed_rho (r1 ++ r2)) by eauto; assert (closed_rho r2) by eauto;
    assert (HCF := canonical_forms_star _ _ _ HT);
    destruct_things; subst;
    try solve [autorewrite with step in HMS; mpstep_inversion];
    auto.
  Qed.

  Theorem wf_star_does_not_step_to_lam : forall G1 G2 t B r1 r2 b k,
       typing (G1 ++ G2) t B exp_star
    -> mpstep (close r2 B (length r1)) (exp_lam b)
    -> env_wf k (G1 ++ G2) (r1 ++ r2)
    -> False.
  Proof.
    intros G1 G2 t B r1 r2 b k HT HMS HWF;
    assert (closed_rho (r1 ++ r2)) by eauto; assert (closed_rho r2) by eauto;
    assert (HCF := canonical_forms_star _ _ _ HT);
    destruct_things; subst;
    try solve [autorewrite with step in HMS; mpstep_inversion];
    auto.
  Qed.

  Theorem wf_star_does_not_step_to_nrec : forall G1 G2 t B r1 r2 b k,
       typing (G1 ++ G2) t B exp_star
    -> mpstep (close r2 B (length r1)) (exp_nrec b)
    -> env_wf k (G1 ++ G2) (r1 ++ r2)
    -> False.
  Proof.
    intros G1 G2 t B r1 r2 b k HT HMS HWF;
    assert (closed_rho (r1 ++ r2)) by eauto; assert (closed_rho r2) by eauto;
    assert (HCF := canonical_forms_star _ _ _ HT);
    destruct_things; subst;
    try solve [autorewrite with step in HMS; mpstep_inversion];
    auto.
  Qed.


  Theorem wf_star_does_not_step_to_app : forall G1 G2 t B r1 r2 b1 b2 k,
       typing (G1 ++ G2) t B exp_star
    -> mpstep (close r2 B (length r1)) (exp_app b1 b2)
    -> env_wf k (G1 ++ G2) (r1 ++ r2)
    -> False.
  Proof.
    intros G1 G2 t B r1 r2 b1 b2 k HT HMS HWF;
    assert (closed_rho (r1 ++ r2)) by eauto; assert (closed_rho r2) by eauto;
    assert (HCF := canonical_forms_star _ _ _ HT);
    destruct_things; subst;
    try solve [autorewrite with step in HMS; mpstep_inversion];
    auto.
  Qed.

  Theorem wf_star_steps_to_at : forall G1 G2 t B r1 r2 A t' k,
       typing (G1 ++ G2) t B exp_star
    -> mpstep (close r2 B (length r1)) (exp_at A t')
    -> env_wf k (G1 ++ G2) (r1 ++ r2)
    -> exists B',
            B = exp_at B' t' /\ mpstep (close r2 B' (length r1)) A.
  Proof.
    intros G1 G2 t B r1 r2 A t' k HT HMS HWF;
    assert (closed_rho (r1 ++ r2)) by eauto; assert (closed_rho r2) by eauto;
    assert (HCF := canonical_forms_star _ _ _ HT);
    destruct_things; subst;
    autorewrite with step in HMS; mpstep_inversion; eauto 8.
  Qed.

  Theorem wf_star_steps_to_nat : forall G1 G2 t B r1 r2 k,
       typing (G1 ++ G2) t B exp_star
    -> mpstep (close r2 B (length r1)) exp_nat
    -> env_wf k (G1 ++ G2) (r1 ++ r2)
    -> B = exp_nat.
  Proof.
    intros G1 G2 t B r1 r2 k HT HMS HWF;
    assert (closed_rho (r1 ++ r2)) by eauto; assert (closed_rho r2) by eauto;
    assert (HCF := canonical_forms_star _ _ _ HT);
    destruct_things; subst;
    autorewrite with step in HMS; mpstep_inversion; eauto 8.
  Qed.

  Theorem wf_star_does_not_step_to_zero : forall G1 G2 t B r1 r2 k,
       typing (G1 ++ G2) t B exp_star
    -> mpstep (close r2 B (length r1)) exp_zero
    -> env_wf k (G1 ++ G2) (r1 ++ r2)
    -> False.
  Proof.
    intros G1 G2 t B r1 r2 k HT HMS HWF;
    assert (closed_rho (r1 ++ r2)) by eauto; assert (closed_rho r2) by eauto;
    assert (HCF := canonical_forms_star _ _ _ HT);
    destruct_things; subst;
    try solve [autorewrite with step in HMS; mpstep_inversion];
    auto.
  Qed.

  Theorem wf_star_does_not_step_to_succ : forall G1 G2 t B r1 r2 a k,
       typing (G1 ++ G2) t B exp_star
    -> mpstep (close r2 B (length r1)) (exp_succ a)
    -> env_wf k (G1 ++ G2) (r1 ++ r2)
    -> False.
  Proof.
    intros G1 G2 t B r1 r2 a k HT HMS HWF;
    assert (closed_rho (r1 ++ r2)) by eauto; assert (closed_rho r2) by eauto;
    assert (HCF := canonical_forms_star _ _ _ HT);
    destruct_things; subst;
    try solve [autorewrite with step in HMS; mpstep_inversion];
    auto.
  Qed.

  Theorem wf_star_does_not_step_to_ncase : forall G1 G2 t B r1 r2 a b1 b2 k,
       typing (G1 ++ G2) t B exp_star
    -> mpstep (close r2 B (length r1)) (exp_ncase a b1 b2)
    -> env_wf k (G1 ++ G2) (r1 ++ r2)
    -> False.
  Proof.
    intros G1 G2 t B r1 r2 a b1 b2 k HT HMS HWF;
    assert (closed_rho (r1 ++ r2)) by eauto; assert (closed_rho r2) by eauto;
    assert (HCF := canonical_forms_star _ _ _ HT);
    destruct_things; subst;
    try solve [autorewrite with step in HMS; mpstep_inversion];
    auto.
  Qed.

  Theorem wf_star_steps_to_sum : forall G1 G2 t B r1 r2 A1 A2 k,
       typing (G1 ++ G2) t B exp_star
    -> mpstep (close r2 B (length r1)) (exp_sum A1 A2)
    -> env_wf k (G1 ++ G2) (r1 ++ r2)
    -> exists B1, exists B2, B = exp_sum B1 B2
         /\ mpstep (close r2 B1 (length r1)) A1
         /\ mpstep (close r2 B2 (length r1)) A2.
  Proof.
    intros G1 G2 t B r1 r2 A1 A2 k HT HMS HWF;
    assert (closed_rho (r1 ++ r2)) by eauto; assert (closed_rho r2) by eauto;
    assert (HCF := canonical_forms_star _ _ _ HT);
    destruct_things; subst;
    autorewrite with step in HMS; mpstep_inversion; eauto 8.
  Qed.

  Theorem wf_star_steps_to_sigma : forall G1 G2 t B r1 r2 A1 A2 k,
       typing (G1 ++ G2) t B exp_star
    -> mpstep (close r2 B (length r1)) (exp_sigma A1 A2)
    -> env_wf k (G1 ++ G2) (r1 ++ r2)
    -> exists B1, exists B2, B = exp_sigma B1 B2
         /\ mpstep (close r2 B1 (length r1)) A1
         /\ mpstep (close r2 B2 (S (length r1))) A2.
  Proof.
    intros G1 G2 t B r1 r2 A1 A2 k HT HMS HWF;
    assert (closed_rho (r1 ++ r2)) by eauto; assert (closed_rho r2) by eauto;
    assert (HCF := canonical_forms_star _ _ _ HT);
    destruct_things; subst;
    autorewrite with step in HMS; mpstep_inversion; eauto 8.
  Qed.

  Theorem wf_star_does_not_step_to_inl : forall G1 G2 t B r1 r2 a k,
       typing (G1 ++ G2) t B exp_star
    -> mpstep (close r2 B (length r1)) (exp_inl a)
    -> env_wf k (G1 ++ G2) (r1 ++ r2)
    -> False.
  Proof.
    intros G1 G2 t B r1 r2 a k HT HMS HWF;
    assert (closed_rho (r1 ++ r2)) by eauto; assert (closed_rho r2) by eauto;
    assert (HCF := canonical_forms_star _ _ _ HT);
    destruct_things; subst;
    try solve [autorewrite with step in HMS; mpstep_inversion];
    auto.
  Qed.

  Theorem wf_star_does_not_step_to_pair : forall G1 G2 t B r1 r2 a b k,
       typing (G1 ++ G2) t B exp_star
    -> mpstep (close r2 B (length r1)) (exp_pair a b)
    -> env_wf k (G1 ++ G2) (r1 ++ r2)
    -> False.
  Proof.
    intros G1 G2 t B r1 r2 a b k HT HMS HWF;
    assert (closed_rho (r1 ++ r2)) by eauto; assert (closed_rho r2) by eauto;
    assert (HCF := canonical_forms_star _ _ _ HT);
    destruct_things; subst;
    try solve [autorewrite with step in HMS; mpstep_inversion];
    auto.
  Qed.

  Theorem wf_star_does_not_step_to_inr : forall G1 G2 t B r1 r2 a k,
       typing (G1 ++ G2) t B exp_star
    -> mpstep (close r2 B (length r1)) (exp_inr a)
    -> env_wf k (G1 ++ G2) (r1 ++ r2)
    -> False.
  Proof.
    intros G1 G2 t B r1 r2 a k HT HMS HWF;
    assert (closed_rho (r1 ++ r2)) by eauto; assert (closed_rho r2) by eauto;
    assert (HCF := canonical_forms_star _ _ _ HT);
    destruct_things; subst;
    try solve [autorewrite with step in HMS; mpstep_inversion];
    auto.
  Qed.

  Theorem wf_star_does_not_step_to_scase : forall G1 G2 t B r1 r2 a b1 b2 k,
       typing (G1 ++ G2) t B exp_star
    -> mpstep (close r2 B (length r1)) (exp_scase a b1 b2)
    -> env_wf k (G1 ++ G2) (r1 ++ r2)
    -> False.
  Proof.
    intros G1 G2 t B r1 r2 a b1 b2 k HT HMS HWF;
    assert (closed_rho (r1 ++ r2)) by eauto; assert (closed_rho r2) by eauto;
    assert (HCF := canonical_forms_star _ _ _ HT);
    destruct_things; subst;
    try solve [autorewrite with step in HMS; mpstep_inversion];
    auto.
  Qed.

  Theorem wf_star_does_not_step_to_pcase : forall G1 G2 t B r1 r2 a b k,
       typing (G1 ++ G2) t B exp_star
    -> mpstep (close r2 B (length r1)) (exp_pcase a b)
    -> env_wf k (G1 ++ G2) (r1 ++ r2)
    -> False.
  Proof.
    intros G1 G2 t B r1 r2 a b k HT HMS HWF;
    assert (closed_rho (r1 ++ r2)) by eauto; assert (closed_rho r2) by eauto;
    assert (HCF := canonical_forms_star _ _ _ HT);
    destruct_things; subst;
    try solve [autorewrite with step in HMS; mpstep_inversion];
    auto.
  Qed.

  Theorem wf_star_steps_to_mu : forall G1 G2 t B r1 r2 A k,
       typing (G1 ++ G2) t B exp_star
    -> mpstep (close r2 B (length r1)) (exp_mu A)
    -> env_wf k (G1 ++ G2) (r1 ++ r2)
    -> exists B',
            B = exp_mu B' /\ mpstep (close r2 B' (S (length r1))) A.
  Proof.
    intros G1 G2 t B r1 r2 A k HT HMS HWF;
    assert (closed_rho (r1 ++ r2)) by eauto; assert (closed_rho r2) by eauto;
    assert (HCF := canonical_forms_star _ _ _ HT);
    destruct_things; subst;
    autorewrite with step in HMS; mpstep_inversion; eauto 8.
  Qed.

  Theorem wf_star_does_not_step_to_roll : forall G1 G2 t B r1 r2 a k,
       typing (G1 ++ G2) t B exp_star
    -> mpstep (close r2 B (length r1)) (exp_roll a)
    -> env_wf k (G1 ++ G2) (r1 ++ r2)
    -> False.
  Proof.
    intros G1 G2 t B r1 r2 a k HT HMS HWF;
    assert (closed_rho (r1 ++ r2)) by eauto; assert (closed_rho r2) by eauto;
    assert (HCF := canonical_forms_star _ _ _ HT);
    destruct_things; subst;
    try solve [autorewrite with step in HMS; mpstep_inversion];
    auto.
  Qed.

  Theorem wf_star_does_not_step_to_unroll : forall G1 G2 t B r1 r2 a k,
       typing (G1 ++ G2) t B exp_star
    -> mpstep (close r2 B (length r1)) (exp_unroll a)
    -> env_wf k (G1 ++ G2) (r1 ++ r2)
    -> False.
  Proof.
    intros G1 G2 t B r1 r2 a k HT HMS HWF;
    assert (closed_rho (r1 ++ r2)) by eauto; assert (closed_rho r2) by eauto;
    assert (HCF := canonical_forms_star _ _ _ HT);
    destruct_things; subst;
    try solve [autorewrite with step in HMS; mpstep_inversion];
    auto.
  Qed.

  Theorem wf_star_steps_to_eq : forall G1 G2 t B r1 r2 a1 a2 k,
       typing (G1 ++ G2) t B exp_star
    -> mpstep (close r2 B (length r1)) (exp_eq a1 a2)
    -> env_wf k (G1 ++ G2) (r1 ++ r2)
    -> exists b1, exists b2, B = exp_eq b1 b2
         /\ mpstep (close r2 b1 (length r1)) a1
         /\ mpstep (close r2 b2 (length r1)) a2.
  Proof.
    intros G1 G2 t B r1 r2 A1 A2 k HT HMS HWF;
    assert (closed_rho (r1 ++ r2)) by eauto; assert (closed_rho r2) by eauto;
    assert (HCF := canonical_forms_star _ _ _ HT);
    destruct_things; subst;
    autorewrite with step in HMS; mpstep_inversion; eauto 8.
  Qed.

  Theorem wf_star_does_not_step_to_join : forall G1 G2 t B r1 r2 k,
       typing (G1 ++ G2) t B exp_star
    -> mpstep (close r2 B (length r1)) exp_join
    -> env_wf k (G1 ++ G2) (r1 ++ r2)
    -> False.
  Proof.
    intros G1 G2 t B r1 r2 k HT HMS HWF;
    assert (closed_rho (r1 ++ r2)) by eauto; assert (closed_rho r2) by eauto;
    assert (HCF := canonical_forms_star _ _ _ HT);
    destruct_things; subst;
    try solve [autorewrite with step in HMS; mpstep_inversion];
    auto.
  Qed.

  Hint Extern 0 =>
  match goal with
  | [ H : typing ((th_log,exp_star) :: _) th_log ?x exp_star
     |- typing _ th_log ({0 ~> exp_mu  ?x}?x) exp_star ] =>
    eapply typing_substitution_front_closed with (t' := th_log); try eassumption; eauto
  end.

  Theorem interp_respects_mpstep : forall k A cv B1 B2 t r1 r2 a G11 G12 G2,
         mpstep (close r2 B1 (length r1)) A
      -> mpstep (close r2 B2 (length r1)) A
      -> typing (G11 ++ G2) t B1 exp_star
      -> typing (G12 ++ G2) t B2 exp_star
      -> env_wf k (G11 ++ G2) (r1 ++ r2)
      -> env_wf k (G12 ++ G2) (r1 ++ r2)
      -> length G11 = length r1
      -> length G12 = length r1
      -> interp (k,B1,cv) t (r1 ++ r2) a
      -> interp (k,B2,cv) t (r1 ++ r2) a.
  Proof.
    intros k A; apply index_size_combined_ind with 
      (P := fun k A => forall cv B1 B2 t r1 r2 a G11 G12 G2,
               mpstep (close r2 B1 (length r1)) A
            -> mpstep (close r2 B2 (length r1)) A
            -> typing (G11 ++ G2) t B1 exp_star
            -> typing (G12 ++ G2) t B2 exp_star
            -> env_wf k (G11 ++ G2) (r1 ++ r2)
            -> env_wf k (G12 ++ G2) (r1 ++ r2)
            -> length G11 = length r1
            -> length G12 = length r1
            -> interp (k,B1,cv) t (r1 ++ r2) a
            -> interp (k,B2,cv) t (r1 ++ r2) a).
    clear k A; intros k A IH cv B1 B2 t r1 r2 a G11 G12 G2
                      HMS1 HMS2 HT1 HT2 HEWF1 HEWF2 HLen1 HLen2 HI;
    assert (mpstep (close (r1 ++ r2) B1 0) (close r1 A 0)) by
      (apply_close_congs; eauto using close_preserves_mpstep);
    assert (mpstep (close (r1 ++ r2) B2 0) (close r1 A 0)) by
      (apply_close_congs; eauto using close_preserves_mpstep);
    assert (closed_rho (r1 ++ r2)) by eauto;
    assert (closed_rho r1) by eauto; assert (closed_rho r2) by eauto;
    destruct cv; try match goal with
    | [ |- interp (_, _, Comp) _ _ _ ] =>
      unfold_interp_hyps; unfold_interp; intros; try inst_interp;
      try match goal with
      | [ H : mstep ?n ?a ?v |- exists v, exists n, mstep n ?a v /\ _ ] =>
        exists v; exists n; split; [ auto | ]
      end
    end; destruct A; try match goal with
    | [ HMS1 : mpstep (close r2 B1 (length r1)) (exp_var _) |- _ ] =>
      destruct (wf_star_does_not_step_to_var _ _ _ _ _ _ _ _
                   HT1 HMS1 HEWF1)
    | [ HMS1 : mpstep (close r2 B1 (length r1)) exp_star |- _ ] =>
      destruct (wf_star_does_not_step_to_star _ _ _ _ _ _ _
                   HT1 HMS1 HEWF1)
    | [ HMS1 : mpstep (close r2 B1 (length r1)) (exp_arr _ _) |- _ ] =>
      assert (HIS1 := wf_star_steps_to_arr _ _ _ _ _ _ _ _ _ HT1 HMS1 HEWF1);
      assert (HIS2 := wf_star_steps_to_arr _ _ _ _ _ _ _ _ _ HT2 HMS2 HEWF2);
      destruct_things; subst; try rewrite <- close_app in * by auto;
      unfold_interp_hyps; unfold_interp; intros; go_go_regularity;
      match goal with
      | [ HMS1 : mpstep (close _ (exp_arr ?B11 ?B12) _)
                        (close _ (exp_arr ?A1 ?A2) _),
          HMS2 : mpstep (close _ (exp_arr ?B21 ?B22) _)
                        (close _ (exp_arr ?A1 ?A2) _),
          HI2 : interp (_,?B21,_) _ _ ?v
         |- interp (?j,?B22,_) ?t (?v :: ?r1 ++ ?r2) _ ] =>
        require_neq B11 B21;
        assert (env_wf j (G11 ++ G2) (r1 ++ r2))
            by (eapply env_wf_downward_closure_leq; eauto);
        assert (env_wf j (G12 ++ G2) (r1 ++ r2))
            by (eapply env_wf_downward_closure_leq; eauto);
        assert (interp (j, B11, Val) t (r1 ++ r2) v)
            by (eapply IH with (A' := A1) (B1 := B21); eauto);
        rewrite app_comm_cons;
        eapply IH with (A' := A2) (B1 := B12)
                       (G11 := (t,B11)::G11) (G12 := (t,B21)::G12);
        simpl; eauto
      end
    | [ HMS1 : mpstep (close r2 B1 (length r1)) (exp_fun _) |- _ ] =>
      destruct (wf_star_does_not_step_to_fun _ _ _ _ _ _ _ _
                   HT1 HMS1 HEWF1)
    | [ HMS1 : mpstep (close r2 B1 (length r1)) (exp_lam _) |- _ ] =>
      destruct (wf_star_does_not_step_to_lam _ _ _ _ _ _ _ _
                   HT1 HMS1 HEWF1)
    | [ HMS1 : mpstep (close r2 B1 (length r1)) (exp_nrec _) |- _ ] =>
      destruct (wf_star_does_not_step_to_nrec _ _ _ _ _ _ _ _
                   HT1 HMS1 HEWF1)
    | [ HMS1 : mpstep (close r2 B1 (length r1)) (exp_app _ _) |- _ ] =>
      destruct (wf_star_does_not_step_to_app _ _ _ _ _ _ _ _ _
                   HT1 HMS1 HEWF1)
    | [ HMS1 : mpstep (close r2 B1 (length r1)) (exp_at _ _) |- _ ] =>
      assert (HIS1 := wf_star_steps_to_at _ _ _ _ _ _ _ _ _ HT1 HMS1 HEWF1);
      assert (HIS2 := wf_star_steps_to_at _ _ _ _ _ _ _ _ _ HT2 HMS2 HEWF2);
      destruct_things; subst; go_go_regularity; unfold_interp_hyps; unfold_interp;
      try rewrite <- close_app in * by auto; match goal with
      | [ HMS1 : mpstep (close _ (exp_at ?B1' _) _)
                        (close _ (exp_at ?A _) _),
          HMS2 : mpstep (close _ (exp_at ?B2' _) _)
                        (close _ (exp_at ?A _) _),
          HI2 : interp (_,?B1',_) _ _ ?v
         |- interp (?j,?B2',_) ?t (?r1 ++ ?r2) _ ] =>
        eapply IH with (A' := A) (B1 := B1') (B2 := B2'); eauto
      end
    | [ HMS1 : mpstep (close r2 B1 (length r1)) exp_nat |- _ ] =>
      assert (HIS1 := wf_star_steps_to_nat _ _ _ _ _ _ _ HT1 HMS1 HEWF1);
      assert (HIS2 := wf_star_steps_to_nat _ _ _ _ _ _ _ HT2 HMS2 HEWF2);
      subst; go_go_regularity; unfold_interp_hyps; unfold_interp; eauto
    | [ HMS1 : mpstep (close r2 B1 (length r1)) exp_zero |- _ ] =>
      destruct (wf_star_does_not_step_to_zero _ _ _ _ _ _ _
                   HT1 HMS1 HEWF1)
    | [ HMS1 : mpstep (close r2 B1 (length r1)) (exp_succ _) |- _ ] =>
      destruct (wf_star_does_not_step_to_succ _ _ _ _ _ _ _ _
                   HT1 HMS1 HEWF1)
    | [ HMS1 : mpstep (close r2 B1 (length r1)) (exp_ncase _ _ _) |- _ ] =>
      destruct (wf_star_does_not_step_to_ncase _ _ _ _ _ _ _ _ _ _
                   HT1 HMS1 HEWF1)
    | [ HMS1 : mpstep (close r2 B1 (length r1)) (exp_sum _ _) |- _ ] =>
      assert (HIS1 := wf_star_steps_to_sum _ _ _ _ _ _ _ _ _ HT1 HMS1 HEWF1);
      assert (HIS2 := wf_star_steps_to_sum _ _ _ _ _ _ _ _ _ HT2 HMS2 HEWF2);
      destruct_things; subst; go_go_regularity; unfold_interp_hyps; unfold_interp;
      try rewrite <- close_app in * by auto; match goal with
        | [ HMS1 : mpstep (close _ (exp_sum ?B11 ?B12) _)
                          (close _ (exp_sum ?A1 ?A2) _),
            HMS2 : mpstep (close _ (exp_sum ?B21 ?B22) _)
                          (close _ (exp_sum ?A1 ?A2) _) |- _ ] =>
        require_neq B11 B21; match goal with
        | [ HI2 : interp (_,B11,_) _ _ ?v
           |- interp (?j,B21,_) _ (?r1 ++ ?r2) _ ] =>
          eapply IH with (A' := A1) (B1 := B11) (B2 := B21); eauto
        | [ HI2 : interp (_,B12,_) _ _ ?v
           |- interp (?j,B22,_) _ (?r1 ++ ?r2) _ ] =>
          eapply IH with (A' := A2) (B1 := B12) (B2 := B22); eauto
        end
      end
    | [ HMS1 : mpstep (close r2 B1 (length r1)) (exp_sigma _ _)
       |- interp (?k, _, _) ?t ?r _ ] =>
      assert (HIS1 := wf_star_steps_to_sigma _ _ _ _ _ _ _ _ _ HT1 HMS1 HEWF1);
      assert (HIS2 := wf_star_steps_to_sigma _ _ _ _ _ _ _ _ _ HT2 HMS2 HEWF2);
      destruct_things; subst; go_go_regularity; unfold_interp_hyps;
      try rewrite <- close_app in * by auto;  unfold_interp; match goal with
        | [ HMS1 : mpstep (close _ (exp_sigma ?B11 ?B12) _)
                          (close _ (exp_sigma ?A1 ?A2) _),
            HMS2 : mpstep (close _ (exp_sigma ?B21 ?B22) _)
                          (close _ (exp_sigma ?A1 ?A2) _),
            HTpair : typing _ _ (exp_pair ?v1 ?v2) _ |- _ ] =>
        require_neq B11 B21;
        assert (interp (k, B11, Val) t r v1)
            by (eapply IH with (A' := A1) (B1 := B21) (B2 := B11); eauto);
        auto; rewrite app_comm_cons in *; 
        eapply IH with (A' := A2) (B1 := B22) (B2 := B12); eauto; simpl; eauto
      end
    | [ HMS1 : mpstep (close r2 B1 (length r1)) (exp_inl _) |- _ ] =>
      destruct (wf_star_does_not_step_to_inl _ _ _ _ _ _ _ _
                   HT1 HMS1 HEWF1)
    | [ HMS1 : mpstep (close r2 B1 (length r1)) (exp_inr _) |- _ ] =>
      destruct (wf_star_does_not_step_to_inr _ _ _ _ _ _ _ _
                   HT1 HMS1 HEWF1)
    | [ HMS1 : mpstep (close r2 B1 (length r1)) (exp_pair _ _) |- _ ] =>
      destruct (wf_star_does_not_step_to_pair _ _ _ _ _ _ _ _ _
                   HT1 HMS1 HEWF1)
    | [ HMS1 : mpstep (close r2 B1 (length r1)) (exp_scase _ _ _) |- _ ] =>
      destruct (wf_star_does_not_step_to_scase _ _ _ _ _ _ _ _ _ _
                   HT1 HMS1 HEWF1)
    | [ HMS1 : mpstep (close r2 B1 (length r1)) (exp_pcase _ _) |- _ ] =>
      destruct (wf_star_does_not_step_to_pcase _ _ _ _ _ _ _ _ _
                   HT1 HMS1 HEWF1)
    | [ HMS1 : mpstep (close r2 B1 (length r1)) (exp_mu _) |- _ ] =>
      assert (HIS1 := wf_star_steps_to_mu _ _ _ _ _ _ _ _ HT1 HMS1 HEWF1);
      assert (HIS2 := wf_star_steps_to_mu _ _ _ _ _ _ _ _ HT2 HMS2 HEWF2);
      destruct_things; subst; go_go_regularity; unfold_interp_hyps;
      try rewrite <- close_app in * by auto; unfold_interp; intros; match goal with
      | [ HMS1 : mpstep (close _ (exp_mu ?B1') _)
                        (close _ (exp_mu ?A) _),
          HMS2 : mpstep (close _ (exp_mu ?B2') _)
                        (close _ (exp_mu ?A) _)
         |- interp (?j,{0 ~> exp_mu ?B2'}?B2',_) _ _ _ ] =>
        require_neq B1' B2';
        assert (env_wf j (G11 ++ G2) (r1 ++ r2))
            by (eauto using env_wf_downward_closure);
        assert (env_wf j (G12 ++ G2) (r1 ++ r2))
            by (eauto using env_wf_downward_closure);
        eapply IH with (A' := {0 ~> exp_mu A}A)
                       (B1 := {0 ~> exp_mu B1'}B1');
        try rewrite close_open_commute by eauto; eauto
      end
    | [ HMS1 : mpstep (close r2 B1 (length r1)) (exp_roll _) |- _ ] =>
      destruct (wf_star_does_not_step_to_roll _ _ _ _ _ _ _ _
                   HT1 HMS1 HEWF1)
    | [ HMS1 : mpstep (close r2 B1 (length r1)) (exp_unroll _) |- _ ] =>
      destruct (wf_star_does_not_step_to_unroll _ _ _ _ _ _ _ _
                   HT1 HMS1 HEWF1)
    | [ HMS1 : mpstep (close r2 B1 (length r1)) (exp_eq _ _) |- _ ] =>
      assert (HIS1 := wf_star_steps_to_eq _ _ _ _ _ _ _ _ _ HT1 HMS1 HEWF1);
      assert (HIS2 := wf_star_steps_to_eq _ _ _ _ _ _ _ _ _ HT2 HMS2 HEWF2);
      destruct_things; subst; go_go_regularity; unfold_interp_hyps; unfold_interp;
      apply_close_congs; mpstep_inversion;
      repeat (rewrite <- close_app in * by auto);
      match goal with
      | [ HMSa1b1  : mpstep (close ?rs ?a1 _) ?b1,
          HMSa1'b1 : mpstep (close ?rs ?a1' _) ?b1,
          HMSa2b2  : mpstep (close ?rs ?a2 _) ?b2,
          HMSa2'b2 : mpstep (close ?rs ?a2' _) ?b2,
          HMSa1'b  : mpstep (close ?rs ?a1' _) ?b,
          HMSa2'b  : mpstep (close ?rs ?a2' _) ?b
         |- exists _, mpstep (close ?rs ?a1 _) _
                   /\ mpstep (close ?rs ?a2 _) _ ] =>
        require_neq a1 a1'; require_neq a2 a2';
        let d := fresh "d" in
        destruct (diamond_mpstep _ _ HMSa1'b1 _ HMSa1'b) as [? [? Hbc1]];
        destruct (diamond_mpstep _ _ HMSa2'b2 _ HMSa2'b) as [? [? Hbc2]];
        destruct (diamond_mpstep _ _ Hbc1 _ Hbc2) as [d [? ?]];
        exists d; split; eauto using MPTrans
      end

    | [ HMS1 : mpstep (close r2 B1 (length r1)) exp_join |- _ ] =>
      destruct (wf_star_does_not_step_to_join _ _ _ _ _ _ _
                   HT1 HMS1 HEWF1)
    end.
  Qed.
(*
  Theorem interp_rho_eval_swap : forall k A cv t rho a b1 t1 B1 v n,
       mstep n (close_e rho b1 0) v
    -> closed_rho rho
    -> typing nil t1 (close_e rho b1 0) B1
    -> typ_wf nil (close_t rho ({0 ~e> b1}A) 0)
    -> interp (k, {0 ~e> v}A, cv) t rho a
    -> interp (k, {0 ~e> b1}A, cv) t rho a.
  Proof.
    intros k A cv t rho a b1 t1 B1 v n; intros;
    assert (typing nil t1 v B1) by eauto;
    eapply interp_rho_convertible_swap with (b1 := v) (b2 := b1) (c := v); eauto.
  Qed. *)

  Theorem interp_respects_mpstep_full : forall k A cv B1 B2 t rho a G,
        mpstep (close rho B1 0) A
     -> mpstep (close rho B2 0) A
     -> typing G t B1 exp_star
     -> typing G t B2 exp_star
     -> env_wf k G rho
     -> interp (k, B1, cv) t rho a 
     -> interp (k, B2, cv) t rho a.
  Proof.
    intros; replace rho with (nil ++ rho) by auto;
    eapply interp_respects_mpstep with (B1 := B1) (A := A) (G11 := nil) (G12 := nil);
    simpl; eauto.
  Qed.
End ConvSwap.


Section Weakening.
  Hint Extern 1 =>
    match goal with
    | [ H : typing _ _ ?a (close (?r ++ _) _ _)
       |- typing _ _ ?a (close (?r ++ ?b :: _) _ _) ] =>
      rewrite close_app in * by eauto 3; simpl;
      repeat (rewrite open_under_lift by auto);
      autorewrite with step in H; autorewrite with step; eauto
    | [ H : typing _ _ ?a (close (?r ++ ?b :: _) _ _)
       |- typing _ _ ?a (close (?r ++ _) _ _) ] =>
      rewrite close_app in * by eauto 3; simpl in H;
      repeat (rewrite open_under_lift in H by auto);
      autorewrite with step in H; autorewrite with step; eauto
    | [ H : typing _ _ (close (?r ++ _) _ _) exp_star
       |- typing _ _ (close (?r ++ ?b :: _) _ _) exp_star ] =>
      rewrite close_app in * by eauto 3; simpl;
      repeat (rewrite open_under_lift by auto);
      autorewrite with step in H; autorewrite with step; eauto
    | [ H : typing _ _ (close (?r ++ ?b :: _) _ _) exp_star
       |- typing _ _ (close (?r ++ _) _ _) exp_star ] =>
      rewrite close_app in * by eauto 3; simpl in H;
      repeat (rewrite open_under_lift in H by auto);
      autorewrite with step in H; autorewrite with step; eauto
    end.
        
  Theorem interp_weakening_strengthening : forall k A,
       (forall cv t r1 r2 a,
           interp (k, A, cv) t (r1 ++ r2) a
        -> forall v,
                closed_rho (r1 ++ v :: r2)
             -> interp (k, lift A 1 (length r1), cv) t (r1 ++ v :: r2) a)
    /\ (forall cv B t r1 r2 v a,
            A = lift B 1 (length r1)
         -> closed_rho (r1 ++ v :: r2)
         -> interp (k, lift B 1 (length r1), cv) t (r1 ++ v :: r2) a
         -> interp (k, B, cv) t (r1 ++ r2) a).
  Proof.
    apply index_size_combined_ind;
    intros k A IH; subst; split; destruct cv; intros; simpl;
    try match goal with
    | [ |- interp (_, _, Comp) _ _ _ ] =>
      unfold_interp_hyps; unfold_interp; intros; try inst_interp;
      try match goal with
      | [ H : mstep ?n ?a ?v |- exists v, exists n, mstep n ?a v /\ _ ] =>
        exists v; exists n; split; [ auto | ]
      end
    end; destruct A;
    try match goal with
    | [ H : interp (_, exp_var _, _) _ _ _ |- _ ] => unfold_interp_hyps
    | [ H : interp (_, ?B, _) _ _ _, HEq : exp_var _ = ?B |- _ ] =>
      rewrite <- HEq in H; unfold_interp_hyps
    | [ H : interp (_, exp_fun _, _) _ _ _ |- _ ] => unfold_interp_hyps
    | [ H : interp (_, ?B, _) _ _ _, HEq : exp_fun _ = ?B |- _ ] =>
      rewrite <- HEq in H; unfold_interp_hyps
    | [ H : interp (_, exp_lam _, _) _ _ _ |- _ ] => unfold_interp_hyps
    | [ H : interp (_, ?B, _) _ _ _, HEq : exp_lam _ = ?B |- _ ] =>
      rewrite <- HEq in H; unfold_interp_hyps
    | [ H : interp (_, exp_nrec _, _) _ _ _ |- _ ] => unfold_interp_hyps
    | [ H : interp (_, ?B, _) _ _ _, HEq : exp_nrec _ = ?B |- _ ] =>
      rewrite <- HEq in H; unfold_interp_hyps
    | [ H : interp (_, exp_app _ _, _) _ _ _ |- _ ] => unfold_interp_hyps
    | [ H : interp (_, ?B, _) _ _ _, HEq : exp_app _ _ = ?B |- _ ] =>
      rewrite <- HEq in H; unfold_interp_hyps
    | [ H : interp (_, exp_zero, _) _ _ _ |- _ ] => unfold_interp_hyps
    | [ H : interp (_, ?B, _) _ _ _, HEq : exp_zero = ?B |- _ ] =>
      rewrite <- HEq in H; unfold_interp_hyps
    | [ H : interp (_, exp_succ _, _) _ _ _ |- _ ] => unfold_interp_hyps
    | [ H : interp (_, ?B, _) _ _ _, HEq : exp_succ _ = ?B |- _ ] =>
      rewrite <- HEq in H; unfold_interp_hyps
    | [ H : interp (_, exp_ncase _ _ _, _) _ _ _ |- _ ] => unfold_interp_hyps
    | [ H : interp (_, ?B, _) _ _ _, HEq : exp_ncase _ _ _ = ?B |- _ ] =>
      rewrite <- HEq in H; unfold_interp_hyps
    | [ H : interp (_, exp_inl _, _) _ _ _ |- _ ] => unfold_interp_hyps
    | [ H : interp (_, ?B, _) _ _ _, HEq : exp_inl _ = ?B |- _ ] =>
      rewrite <- HEq in H; unfold_interp_hyps
    | [ H : interp (_, exp_inr _, _) _ _ _ |- _ ] => unfold_interp_hyps
    | [ H : interp (_, ?B, _) _ _ _, HEq : exp_inr _ = ?B |- _ ] =>
      rewrite <- HEq in H; unfold_interp_hyps
    | [ H : interp (_, exp_scase _ _ _, _) _ _ _ |- _ ] => unfold_interp_hyps
    | [ H : interp (_, ?B, _) _ _ _, HEq : exp_scase _ _ _ = ?B |- _ ] =>
      rewrite <- HEq in H; unfold_interp_hyps
    | [ H : interp (_, exp_roll _, _) _ _ _ |- _ ] => unfold_interp_hyps
    | [ H : interp (_, ?B, _) _ _ _, HEq : exp_roll _ = ?B |- _ ] =>
      rewrite <- HEq in H; unfold_interp_hyps
    | [ H : interp (_, exp_unroll _, _) _ _ _ |- _ ] => unfold_interp_hyps
    | [ H : interp (_, ?B, _) _ _ _, HEq : exp_unroll _ = ?B |- _ ] =>
      rewrite <- HEq in H; unfold_interp_hyps
    | [ H : interp (_, exp_join, _) _ _ _ |- _ ] => unfold_interp_hyps
    | [ H : interp (_, ?B, _) _ _ _, HEq : exp_join = ?B |- _ ] =>
      rewrite <- HEq in H; unfold_interp_hyps
    end;
    try match goal with
    | [ H : _ = lift ?B _ _ |- _ ] => destruct B; inversion H; clear H; subst
    end; try match goal with
    | [ H : interp (_, lift _ _ _, _) _ _ _ |- _ ] => simpl in H
    end; simpl; unfold_interp_hyps; unfold_interp; intros; auto;
    match goal with
    (* exp_eq *)
    | [ H : closed_rho (_ ++ ?b :: _ ) |- exists _, mpstep _ _ /\ _ ] => 
      repeat rewrite close_app in * by auto 3; simpl in *;
      repeat (rewrite open_under_lift in * by auto);
      autorewrite with step in *; eauto
    (* exp_mu *)
    | [ HC : closed_rho (_ ++ ?b :: _),
        HT : typing _ _ _ (close _ (exp_mu _) _) |- _ ] =>
      let H := fresh "H" in match goal with
      | [ |- interp (?j, {_ ~> exp_mu (lift ?A ?n (S ?m))}_, _) _ _ _ ] =>
        replace (exp_mu (lift A n (S m))) with (lift (exp_mu A) n m)
             by eauto;
        rewrite <- lift_open_commute by eauto;
        destruct (IH j ({0 ~> exp_mu A}A)) as [H _]; try apply H; eauto
      | [ |- interp (?j, {_ ~> exp_mu ?B}_, _) _ (?r1 ++ _) _ ] =>
        destruct (IH j (lift ({0 ~> exp_mu B}B) 1 (length r1))) as [_ H];
        try eapply H; eauto
      end
    (* typ_sum, 1 *)
    | [ H : typing _ _ _ (close _ (exp_sum _ _) _)
       |- interp (?k, lift ?A _ _, _) _ _ _ ] =>
        apply_close_congs; go_go_regularity;
        repeat (rewrite open_under_lift in * by auto);
        repeat (rewrite lift_0_is_id in * ); destruct (IH k A) as [? _]; eauto
    | [ H : closed_rho (?r1 ++ ?b :: _),
        HT : typing _ _ _ (close _ (lift (exp_sum _ _) _ _) _)
       |- interp (?j, ?A, _) _ _ _ ] =>
      (* typ_sum, 2 *)
      apply_close_congs; go_go_regularity;
      repeat (rewrite open_under_lift in * by auto);
      repeat (rewrite lift_0_is_id in * );
      destruct (IH j (lift A 1 (length r1))) as [_ ?]; eauto
    (* sigma, second half, case 1 *)
    | [ H : typing _ _ _ (close _ (exp_sigma _ _) _)
       |- interp (?j, lift ?A _ (S (length ?r)), _) _ (?v :: ?r ++ _) _ ] =>
      rewrite app_comm_cons; let H2 := fresh "H2" in
      replace (S (length r)) with (length (v :: r)) by auto;
      destruct (IH j A) as [H2 _]; [ | apply H2]; simpl; eauto
    (* nat cases *)
    | [ |- looks_like_nat _ ] => eauto
    (* exp_arr *)
    | [ HIv : interp (_, lift ?A1 _ _, _) _ _ ?v
       |- interp (?j, lift ?A2 _ _, _) _ (?v :: ?r ++ ?r') _ ] =>
      let H2 := fresh "H2" in let H1 := fresh "H1" in
      destruct (IH j A2) as [H2 _]; eauto;
      destruct (IH j (lift A1 1 (length r))) as [_ H1]; eauto;
      replace (v :: r ++ r') with ((v :: r) ++ r') by auto;
      replace (S (length r)) with (length (v :: r)) by auto;
      apply H2; simpl; eauto
    | [ HIv : interp (_, ?A1, _) _ _ ?v
       |- interp (?j, ?A2, _) _ (?v :: ?r ++ ?r') _ ] =>
      let H2 := fresh "H2" in let H1 := fresh "H1" in
      destruct (IH j (lift A2 1 (S (length r)))) as [_ H2]; eauto;
      destruct (IH j A1) as [H1 _]; simpl; eauto;
      replace (v :: r ++ r') with ((v :: r) ++ r') by auto;
      eapply H2; simpl; eauto
    (* exp_at, and sigma first half *)
    | [ |- interp (?j, lift ?A _ _, _) _ _ _ ] => 
      destruct (IH j A) as [? _]; simpl; eauto
    | [ H : interp (?j, lift ?A ?n ?m, _) _ _ ?v
       |- interp (?j, ?A, _) _ _ ?v ] => 
      destruct (IH j (lift A n m)) as [_ ?]; simpl; eauto
    end.
  Qed.

  Theorem interp_weakening : forall k A cv t r1 r2 a,
       interp (k, A, cv) t (r1 ++ r2) a
    -> forall v,
           closed_rho (r1 ++ v :: r2)
        -> interp (k, lift A 1 (length r1), cv) t (r1 ++ v :: r2) a.
  Proof.
    intros; edestruct interp_weakening_strengthening; eauto.
  Qed.
  
  Theorem interp_weakening_cons : forall k A cv t a v r,
      interp (k, A, cv) t r a
   -> closed_rho (v :: r)
   -> interp (k, lift A 1 0, cv) t (v :: r) a.
  Proof.
    intros k A cv t a v r HI HC;
    apply interp_weakening with (r1 := nil) (v := v) in HI; eauto.
  Qed.

  Theorem interp_strengthening : forall k cv B t r1 r2 v a,
       closed_rho (r1 ++ v :: r2)
    -> interp (k, lift B 1 (length r1), cv) t (r1 ++ v :: r2) a
    -> interp (k, B, cv) t (r1 ++ r2) a.
  Proof.
    intros; edestruct interp_weakening_strengthening; eauto.
  Qed.
  
  Theorem interp_strengthening_cons : forall k cv B t rho v a,
       closed_rho (v :: rho)
    -> interp (k, lift B 1 0, cv) t (v :: rho) a
    -> interp (k, B, cv) t rho a.
  Proof.
    intros k cv B t rho v a HC HI;
    apply interp_strengthening with (r1 := nil) (v := v) in HI; eauto.
  Qed.
  
  Theorem interp_strengthening_cons_cons : forall k cv B t rho v1 v2 a,
       closed_rho (v1 :: v2 :: rho)
    -> interp (k, lift B 2 0, cv) t (v1 :: v2 :: rho) a
    -> interp (k, B, cv) t rho a.
  Proof.
    intros k cv B t rho v1 v2 a HC HI;
    replace (lift B 2 0) with (lift (lift B 1 0) 1 0) in HI
         by (autorewrite with step; auto);
    inversion HC; subst;
    repeat (apply interp_strengthening_cons in HI; auto).
  Qed.

  Theorem interp_strengthening_cons_cons_cons : forall k cv B t rho v1 v2 v3 a,
       closed_rho (v1 :: v2 :: v3 :: rho)
    -> interp (k, lift B 3 0, cv) t (v1 :: v2 :: v3 :: rho) a
    -> interp (k, B, cv) t rho a.
  Proof.
    intros k cv B t rho v1 v2 v3 a HC HI;
    replace (lift B 3 0) with (lift (lift (lift B 1 0) 1 0) 1 0) in HI
        by (autorewrite with step; auto);
    repeat (match goal with
    | [ H : closed_rho (_ :: _ :: _) |- _ ] => inversion H; clear H; subst
    end); repeat (apply interp_strengthening_cons in HI; auto).
  Qed.

End Weakening.
Hint Resolve interp_weakening_cons.

(* This is the var case of the main theorem *)
Theorem use_env_wf_trm' : forall k G rho,
     env_wf k G rho
  -> forall G1 t A G2, G = G1 ++ (t,A) :: G2 
       -> interp (k,lift A (S (length G1)) 0, Val) 
                 t rho (close rho (exp_var (length G1)) 0).
Proof.
  induction 1; intros; subst; eauto;
  match goal with
  | [ H : _ = ?G1 ++ _ |- _ ] =>
    destruct G1; inversion H; subst; simpl;
    try match goal with
    | [ |- context [ lift ?A (S (S (length ?G))) 0 ] ] =>
      replace (lift A (S (S (length G))) 0)
         with (lift (lift A (S (length G)) 0) 1 0)
           by (autorewrite with step; simpl; auto)
    end; eauto 6
  end.
Qed.

Theorem use_env_wf_trm : forall k G1 t A G2 rho,
     env_wf k (G1 ++ (t,A) :: G2) rho
  -> interp (k, lift A (S (length G1)) 0, Val) t rho 
            (close rho (exp_var (length G1)) 0).
Proof.
  intros; eauto using use_env_wf_trm'.
Qed.
Hint Resolve use_env_wf_trm.

