Require Import List.
Require Import Arith.

Require Import step_defs.
Require Import step_db_basics.
Require Import step_reduction_basics.
Require Import step_typing_basics.
Require Import step_substitution.

Require Import Program.Equality.
Require Import Omega.

(******************************************************************)
(* Some lemmas about changing consistency classifiers in the ctxt *)

(**** context theta weakening ***)

Theorem theta_weakening_combined :
     (forall G t a A, typing G t a A
        -> forall G1 B G2, G = G1 ++ (th_prog,B) :: G2
           -> typing G2 th_log B exp_star
           -> typing (G1 ++ (th_log,B) :: G2) t a A)
  /\  (forall G, ctx_wf G
        -> forall G1 B G2, G = G1 ++ (th_prog,B) :: G2
           -> typing G2 th_log B exp_star
           -> ctx_wf (G1 ++ (th_log,B) :: G2)).
Proof.
  apply typing_combined_ind; intros; subst;
  try solve [inst_equalities; eauto; mycrush; eauto];
  try match goal with
  | [ HEq : ?G1 ++ _ :: _ = ?G1' ++ _ :: _ |- context[length _] ] =>
    destruct_compares (length G1) (length G1');
    invert_list_equalities; subst; eauto; find_cons; inst_equalities;
    try (eauto using TSub);
    [ autorewrite with step in *; simpl in *; eauto
    | rewrite app_comm_cons; rewrite app_assoc; tconstructor;
      simpl in *; autorewrite with step in *; simpl in *; eauto ]
  end.
Qed.

Theorem theta_weakening_typing : forall G1 B G2 t a A,
     typing (G1 ++ (th_prog,B) :: G2) t a A
  -> typing G2 th_log B exp_star
  -> typing (G1 ++ (th_log,B) :: G2) t a A.
Proof.
  destruct theta_weakening_combined as [? ?]; eauto.
Qed.

Theorem theta_weakening_typing_cons : forall B G t a A,
     typing ((th_prog,B) :: G) t a A
  -> typing G th_log B exp_star
  -> typing ((th_log,B) :: G) t a A.
Proof.
  intros B G t a A HT HS;
  eapply theta_weakening_typing with (G1 := nil) in HT;
  simpl in *; eauto.
Qed.
Hint Resolve theta_weakening_typing_cons.
(*
Hint Extern 2 =>
  match goal with
  | [ |- typing (?b :: (th_log,?B) :: ?G) _ _ _ ] =>
    assert (FO B) by eauto;
    replace (b :: bnd_trm th_prog B :: G)
       with ((b :: nil) ++ bnd_trm th_log B :: G) by (simpl; auto);
    apply theta_weakening_typing
  end.
*)

(**** context theta strengthening ***)
Theorem theta_strengthening_fo_combined :
     (forall G t a A, typing G t a A
        -> forall G1 B G2, G = G1 ++ (th_log,B) :: G2
            -> FO B
            -> typing (G1 ++ (th_prog,B) :: G2) t a A)
  /\  (forall G, ctx_wf G
        -> forall G1 B G2, G = G1 ++ (th_log,B) :: G2
            -> FO B
            -> ctx_wf (G1 ++ (th_prog,B) :: G2)).
Proof.
  apply typing_combined_ind; intros; subst;
  inst_equalities; eauto; mycrush; eauto;

  match goal with
  | [ HEq : ?G1 ++ _ :: _ = ?G1' ++ _ :: _ |- context[length _] ] =>
    destruct_compares (length G1) (length G1');
    invert_list_equalities; subst; eauto; find_cons; inst_equalities;
    invert_ctx_wf; try match goal with
     | [ H : FO exp_star |- _ ] => inversion H
     end; try (eauto 7 using TSub, TFOVal);
    first [ simpl in *; autorewrite with step in *; tconstructor; eauto
          | rewrite app_comm_cons in *; rewrite app_assoc in *; tconstructor; eauto;
            autorewrite with step; eauto ]
  end.
Qed.

Theorem theta_strengthening_fo_typing : forall G1 B G2 t a A,
     typing (G1 ++ (th_log,B) :: G2) t a A -> FO B
  -> typing (G1 ++ (th_prog,B) :: G2) t a A.
Proof.
  destruct theta_strengthening_fo_combined as [? ?]; eauto.
Qed.

Theorem theta_strengthening_fo_typing_cons : forall B G t a A,
     FO B
  -> typing ((th_log, B) :: G) t a A
  -> typing ((th_prog, B) :: G) t a A.
Proof.
  intros B G t a A HF HT;
  apply theta_strengthening_fo_typing with (G1 := nil) in HT;
  simpl in *; eauto.
Qed.
Hint Immediate theta_strengthening_fo_typing_cons.

Hint Extern 2 =>
  match goal with
  | [ |- typing (?b :: (th_prog,?B) :: ?G) _ _ _ ] =>
    assert (FO B) by eauto;
    replace (b :: (th_prog,B) :: G)
       with ((b :: nil) ++ (th_prog,B) :: G) by (simpl; auto);
    apply theta_strengthening_fo_typing
  | [ HF : FO ?B, HT : typing ((th_log,?B) :: _) th_log ?A ?C
     |- typing ((?t,?B) :: _) ?t ?A ?C ] => is_var t; destruct t 
  | [ HF : FO ?B, HT : typing (_ :: (th_log,?B) :: _) th_log ?A ?C
     |- typing (_ :: (?t,?B) :: _) ?t ?A ?C ] => is_var t; destruct t
  | [ HT : typing ((_,?B) :: _) _ ?A ?C, HFO : FO ?B
     |- typing ((th_prog,?B) :: _) ?t ?A ?C ] =>
     eapply theta_strengthening_fo_typing_cons
  end.

Theorem theta_fo_swap : forall B G t1 t2 t a A,
     FO B
  -> typing G t2 B exp_star
  -> typing ((t1, B) :: G) t a A
  -> typing ((t2, B) :: G) t a A.
Proof.
  intros; destruct t1; destruct t2; eauto.
Qed.


(*************** Regularity ***************)
(* Finally, we can prove the rest of this *)
Ltac untyped a :=
  match goal with
  | [ H : typing _ _ a _ |- _ ] => fail 1
  | _ => idtac
  end.

Theorem regularity : forall G t a A,
     typing G t a A
  ->    (typing G t A exp_star)
     \/ (A = exp_star).
Proof.
  induction 1; invert_ctx_wf; destruct_things; eauto;
  try match goal with
  | [ |- _ \/ (exp_star = _) ] => destruct s; eauto
  end; try match goal with
  | [ H : typing (_ :: _) _ _ _ |- _ ] => 
    apply typing_regularity_ctx in H; invert_ctx_wf; eauto
  end.
Qed.

(******************************************************************)
(*  The judgemental equality is an equivalence relation           *)
(******************************************************************)
Theorem typing_eq_refl : forall G t a A,
     typing G t a A
  -> typing G th_log exp_join (exp_eq a a).
Proof.
  intros; eauto using TJoin.
Qed.
Hint Resolve typing_eq_refl.

Theorem typ_eq_regularity_gen : forall G t a A,
     typing G t a A 
  -> forall b1 b2, a = exp_eq b1 b2
      -> exists t1, exists t2, exists B1, exists B2,
          typing G t1 b1 B1 /\ typing G t2 b2 B2.
Proof.
  induction 1; intros; try match goal with
  | [ H : _ = exp_eq _ _ |- _] => inversion H; subst; clear H
  end; eauto; try destruct_ih; eauto 10.
Qed.

Theorem typ_eq_regularity : forall G t b1 b2 A,
     typing G t (exp_eq b1 b2) A
      -> exists t1, exists t2, exists B1, exists B2,
          typing G t1 b1 B1 /\ typing G t2 b2 B2.
Proof.
  intros; eauto using typ_eq_regularity_gen.
Qed.

Theorem typ_at_regularity_gen : forall G t a A,
     typing G t a A
  -> forall t' A', a = exp_at A' t'
      -> typing G t' A' exp_star.
Proof.
  induction 1; intros; try match goal with
  | [ H : _ = exp_at _ _ |- _] => inversion H; subst; clear H
  end; eauto.
Qed.

Theorem typ_at_regularity : forall G t B t' A,
     typing G t (exp_at B t') A
  -> typing G t' B exp_star.
Proof.
  intros; eauto using typ_at_regularity_gen.
Qed.

Ltac invert_eqs := repeat (match goal with
  | [ H : typing _ _ _ ?A |- _ ] =>
    let HNeq := fresh "HNeq" in let HContra := fresh "HContra" in
    let HReg := fresh "HReg" in
    untyped A; assert (A = exp_star -> False) as HNeq
                   by (intro HContra; inversion HContra);
    destruct (regularity _ _ _ _ H) as [HReg | HReg];
    [ | apply HNeq in HReg; destruct HReg]; clear HNeq
          end);
  repeat (match goal with
  | [ H : typing ?G _ (exp_eq ?a ?b) _ |- _ ] =>
    first [untyped a | untyped b]; let H' := fresh H in 
    assert (H' := typ_eq_regularity _ _ _ _ _ H)
  | [ H : typing ?G _ (exp_at ?b _) _ |- _ ] =>
    untyped b; let H' := fresh H in 
    assert (H' := typ_at_regularity _ _ _ _ _ H)
  end; destruct_things).

Theorem typing_eq_sym : forall G p a1 a2,
     typing G th_log p (exp_eq a1 a2)
  -> typing G th_log exp_join (exp_eq a2 a1).
Proof.
  intros G p a1 a2 ?.
  replace (exp_eq a2 a1) with ({0 ~> a2} (exp_eq (exp_var 0) (lift a1 1 0)))
       by (simpl; autorewrite with step; auto);
  invert_eqs; eapply TConv; simpl; autorewrite with step; eauto.
Qed.
Hint Immediate typing_eq_sym.

(******************************************************************)
(*      Context conversion                                        *)
(******************************************************************)

(* This is just to instantiate theta *)
Theorem eliminate_star_type : forall G G' t' a A A',
     typing G' t' exp_star A'
  -> typing G th_log a A.
Proof.  intros G G' t' a A A' H; apply star_has_no_type in H; inversion H. Qed.

Hint Extern 0 =>
  match goal with
  | [ H : typing _ _ exp_star _ |- _ ] =>
    first [ apply (eliminate_star_type _ _ _ _ _ _ H)
          | apply star_has_no_type in H; inversion H ]
  end.

Theorem eq_star : forall b A,
     {0 ~> b}A = exp_star
  -> (A = exp_star) \/ (b = exp_star /\ A = exp_var 0).
Proof.
  destruct A; intros; subst; eauto;
  (match goal with
  | [ H : _ = _ |- _ ] =>
    simpl in H; try solve [inversion H];
    db_var_case; inversion H
  end).
Qed.

Theorem ctx_conversion_combined :
     (forall G t a A, typing G t a A
        -> forall G1 t' B G2 b1 b2 p, G = G1 ++ (t',({0 ~> b1}B)) :: G2
            -> typing G2 th_log p (exp_eq b1 b2)
            -> typing G2 t' ({0 ~> b2}B) exp_star
            -> typing (G1 ++ (t',({0 ~> b2}B)) :: G2) t a A)
  /\  (forall G, ctx_wf G
        -> forall G1 t' B G2 b1 b2 p, G = G1 ++ (t',({0 ~> b1}B)) :: G2
            -> typing G2 th_log p (exp_eq b1 b2)
            -> typing G2 t' ({0 ~> b2}B) exp_star
            -> ctx_wf (G1 ++ (t',({0 ~> b2}B)) :: G2)).
Proof.
  apply typing_combined_ind; intros; subst; simpl in *;
  inst_equalities; eauto; try solve [mycrush].
  try match goal with
  | [ IH : forall _ _ _ _ _ _ _, _ = _ -> _,
      Heq : ?G1 ++ _ :: _ = ?G0 ++ _ :: _
     |- typing _ _ (exp_var (length ?G1)) _ ] =>
    let Heq' := fresh Heq in assert (Heq' := Heq);
    eapply IH in Heq'; eauto; clear IH; invert_ctx_wf;
    destruct_compares (length G1) (length G0); invert_list_equalities;
    try find_cons; try match goal with
    | [ H : exp_star = {_ ~> _}_ |- _ ] => apply eq_sym in H; apply eq_star in H;
      destruct_things; subst; simpl in *; try invert_eqs; eauto
    end; 
    mycrush; (* needs to be after the previous match, to avoid uninstantiated existential *)
    first [
     rewrite app_comm_cons, app_assoc; apply TVar;
     autorewrite with step; simpl; eauto
   | rewrite <- lift_open_commute; eauto ]
  end.
Qed.

Theorem ctx_conv_typing : forall G1 t' B G2 b1 b2 p t a A,
     typing (G1 ++ (t',{0 ~> b1}B) :: G2) t a A
  -> typing G2 th_log p (exp_eq b1 b2)
  -> typing G2 t' ({0 ~> b2}B) exp_star
  -> typing (G1 ++ (t',{0 ~> b2}B) :: G2) t a A.
Proof.
  intros; destruct ctx_conversion_combined as [? ?]; eauto.
Qed.
Hint Resolve ctx_conv_typing.
Hint Extern 4 =>
  match goal with
  | [   H : typing _ th_log _ (exp_eq ?b1' ?b2')
     |- typing (_ ++ (_,{0 ~> ?b2'}_) :: _) _ _ _ ] =>
    eapply ctx_conv_typing with (b1 := b1')
  end.

Theorem ctx_conv_ctx_wf : forall G1 t' B G2 b1 b2 p,
     ctx_wf (G1 ++ (t',{0 ~> b1}B) :: G2)
  -> typing G2 th_log p (exp_eq b1 b2)
  -> typing G2 t' ({0 ~> b2}B) exp_star
  -> ctx_wf (G1 ++ (t',{0 ~> b2}B) :: G2).
Proof.
  intros; destruct ctx_conversion_combined as [? ?]; eauto.
Qed.

Theorem ctx_conv_typing_cons : forall n B t' t a G b1 b2 p A,
     typing ((t',{n ~> b1}B) :: G) t a A
  -> typing G th_log p (exp_eq b1 b2)
  -> typing G t' ({n ~> b2}B) exp_star
  -> typing ((t',{n ~> b2}B) :: G) t a A.
Proof.
  intros n B; intros; match goal with
  | [ |- typing (?b1 :: ?G) _ _ _ ] =>
    replace (b1 :: G) with (nil ++ b1 :: G) by auto
  end;
  destruct (open_index_swap B n 0) as [? Heq];
  rewrite Heq in *; eauto.
Qed.
Hint Resolve ctx_conv_typing_cons.
(*
Theorem ctx_conv_typing_cons_cons : forall tc tb B C b1 b2 p G t a A,
     typing G th_log p (typ_eq b1 b2)
  -> typing (   bnd_trm tc (lift_t ({0 ~e> b1}C) 1 0) 
             :: bnd_trm tb ({0 ~e> b1}B) :: G) t a A
  -> typ_wf G ({0 ~e> b2}C)
  -> typ_wf G ({0 ~e> b2}B)
  -> typing (   bnd_trm tc (lift_t ({0 ~e> b2}C) 1 0) 
             :: bnd_trm tb ({0 ~e> b2}B) :: G) t a A.
Proof.
  intros;
  match goal with
  | [ H : typing (bnd_trm _ (lift_t ({0 ~e> ?b1'} ?C) 1 0) :: _) _ _ _ |- _ ] =>
    autorewrite with step in *;
    eapply ctx_conv_typing_cons with (b1 := lift_e b1' 1 0); eauto;
    try rewrite <- lift_open_commute_t_e; eauto
  end;
  match goal with
  | [ |- typing (?A :: ?B :: ?G) _ _ _ ] =>
    replace (A :: B :: G) with ((A :: nil) ++ B :: G) by (simpl; auto); eauto 6
  end.
Qed.
Hint Resolve ctx_conv_typing_cons_cons.
*)
Ltac trycconvs :=
  repeat (match goal with
  | [ H : pstep ?a ?b |- typing ((_,{_ ~> ?b} _) :: _) _ _ _ ] =>
    eapply ctx_conv_typing_cons with (b1 := a)
  | [ H : pstep ?a ?b |- typing ((_,?b) :: _) _ _ _ ] =>
    replace b with ({0 ~> b}(exp_var 0)) by eauto;
    eapply ctx_conv_typing_cons with (b1 := a)
  | [ H : pstep ?a ?b 
     |- typing ((_,{_ ~> lift ?b ?n ?m} _) :: _) _ _ _ ] =>
    eapply ctx_conv_typing_cons with (b1 := lift a n m);
    try eapply typing_weakening_cons_eq
  end).
Ltac convssubs := trycconvs; trysubs.
Ltac mycrush_cs := mycrush' convssubs.

(* TConv should really work at any n *)
(*
Theorem TConv_n : forall n A G p a b1 b2 t,
     typing G th_log p (typ_eq b1 b2)
  -> typing G t a ({n ~e> b1} A)
  -> typ_wf G ({n ~e> b2} A)
  -> typing G t a ({n ~e> b2} A).
Proof.
  intros n; intros; destruct (open_index_swap_t_e A n 0) as [? Heq];
  rewrite Heq in *; mycrush.
Qed.
Hint Extern 3 =>
  match goal with
  | [ H : typing _ _ ?p' (typ_eq _ ?b2) 
     |- typing _ _ _ ({?n ~e> ?b2} _) ] =>
    eapply (TConv_n n) with (p := p')
  | [ H : typing _ _ _ (typ_eq ?b1' ?b2') 
     |- typing _ _ _ ({?n ~e> lift_e ?b2' ?n ?m} _) ] =>
    eapply (TConv_n n) with (b1 := lift_e b1' n m)
  end.
*)

(******************************************************************)
(*             Inversion                                          *)
(******************************************************************)

Ltac pull_open_from_eq :=
  repeat (match goal with
  | [ |- context [exp_eq ({0 ~> ?a}?A) ?B] ] =>
      replace (exp_eq ({0 ~> a}A) B) with ({0 ~> a}(exp_eq A (lift B 1 0)))
           by (simpl; autorewrite with step; auto)
  | [ |- context [exp_eq ?B ({0 ~> ?a}?A)] ] =>
      replace (exp_eq B ({0 ~> a}A)) with ({0 ~> a}(exp_eq (lift B 1 0) A))
           by (simpl; autorewrite with step; auto)
  | [ H : context [exp_eq ({0 ~> ?a}?A) ?B] |- _ ] =>
      replace (exp_eq ({0 ~> a}A) B) with ({0 ~> a}(exp_eq A (lift B 1 0)))
           in H by (simpl; autorewrite with step; auto)
  | [ H : context [exp_eq ?B ({0 ~> ?a}?A)] |- _ ] =>
      replace (exp_eq B ({0 ~> a}A)) with ({0 ~> a}(exp_eq (lift B 1 0) A))
           in H by (simpl; autorewrite with step; auto)
  end).

Hint Extern 2 =>
  match goal with
  | [ HEq : typing _ th_log ?p' (exp_eq ?a ?b)
     |- typing _ _ _ ({0 ~> ?b} _) ] =>
    eapply TConv with (b1 := a) (p := p');
    simpl; autorewrite with step; simpl in *; autorewrite with step in *
  | [ HEq : typing _ th_log _ (exp_eq ?b ?a),
      HT : typing ?G _ ?e ?a
     |- typing ?G _ ?e ?b ] =>
    apply typing_eq_sym in HEq;
    replace b with ({0 ~> b}(exp_var 0)) by eauto;
    eapply TConv with (b1 := a) (p := exp_join);
    simpl; autorewrite with step; simpl in *; autorewrite with step in *
  | [ HEq : typing _ th_log ?p' (exp_eq ?a ?b),
      HT : typing ?G _ ?e ?a
     |- typing ?G _ ?e ?b ] =>
    replace b with ({0 ~> b}(exp_var 0)) by eauto;
    eapply TConv with (b1 := a) (p := p');
    simpl; autorewrite with step; simpl in *; autorewrite with step in *
  end.
Hint Extern 1 =>
  match goal with
  | [ H : typing _ th_log _ (exp_eq _ ?b) 
     |- typing _ _ _ (exp_eq ({0 ~> ?b}_) _) ] =>
    pull_open_from_eq
end.

(* prove equality is a congruence for At *)
Theorem eq_cong_At : forall G p A B t' tA tB CA CB,
     typing G th_log p (exp_eq A B)
  -> typing G tA (exp_at A t') CA
  -> typing G tB (exp_at B t') CB
  -> typing G th_log exp_join (exp_eq (exp_at A t') (exp_at B t')).
Proof.
  intros;
  replace (exp_at B t') with ({0 ~> B} (exp_at (exp_var 0) t'))
       by (simpl; auto);
  pull_open_from_eq; eapply TConv with (b1 := A) (p := p); simpl;
  autorewrite with step; eauto.
Qed.
Hint Resolve eq_cong_At.

Ltac invert_boxes := 
 try match goal with
 | [ H : Boxed _ _ ?x |- _ ] => is_var x; inversion H; subst
 end;
 try match goal with
 | [ HEqAt : typing _ _ _ (exp_eq (exp_at _ ?t1) (exp_at _ ?t2)) |- _ ] =>
   is_var t1; is_var t2; require_neq t1 t2; destruct t1; destruct t2
 end.

Theorem typ_arr_fo_regularity_gen : forall G t a A,
     typing G t a A
  -> forall B1 B2, a = exp_arr B1 B2 -> FO B1.
Proof.
  induction 1; intros; try match goal with
  | [ H : _ = exp_arr _ _ |- _] => inversion H; subst; clear H
  end; eauto.
Qed.

Theorem typ_arr_fo_regularity : forall G t B1 B2 A,
     typing G t (exp_arr B1 B2) A -> FO B1.
Proof.
  intros; eauto using typ_arr_fo_regularity_gen.
Qed.

Theorem boxed_star_has_no_type : forall t A,
  Boxed exp_star t A -> forall G t' B, typing G t' A B -> False.
Proof.
  induction 1; intros; invert_eqs; eauto using star_has_no_type.
Qed.

Ltac star_isnt_equal := 
  match goal with
  | [ H : typing _ th_log _ (exp_eq _ exp_star) |- _ ] =>
       apply regularity in H; destruct H as [ H | H ]; [ | inversion H];
       eapply typ_eq_regularity in H; eauto; destruct_things;
       match goal with
       | [ H : exp_eq _ _ = exp_star |- _ ] => inversion H
       | [ H : typing _ _ exp_star _ |- _ ] =>
         apply star_has_no_type in H; destruct H
       end
  | [ H : typing _ th_log _ (exp_eq exp_star _) |- _ ] =>
       apply regularity in H; destruct H as [ H | H ]; [ | inversion H];
       eapply typ_eq_regularity in H; eauto; destruct_things;
       match goal with
       | [ H : exp_eq _ _ = exp_star |- _ ] => inversion H
       | [ H : typing _ _ exp_star _ |- _ ] =>
         apply star_has_no_type in H; destruct H
       end
  | [ H : typing _ th_log _ (exp_eq _ (exp_at (exp_star) _)) |- _ ] =>
    apply regularity in H; (destruct H as [ H | H ]; [ | inversion H]);
    eapply typ_eq_regularity in H; eauto; destruct_things;
    match goal with
    | [ H : typing _ _ (exp_at exp_star _) _ |- _ ] =>
      apply typ_at_regularity in H; destruct_things;
      apply star_has_no_type in H; destruct H
    end
  | [ HB  : Boxed exp_star _ ?B,
      H : typing _ th_log _ (exp_eq _ ?B) |- _ ] =>
    apply regularity in H; (destruct H as [ H | H ]; [ | inversion H]);
    eapply typ_eq_regularity in H; eauto; destruct_things;
    match goal with | [ H : typing _ _ B _ |- _ ] => 
    destruct (boxed_star_has_no_type _ _ HB _ _ _ H)
    end
  end.

Theorem Boxed_typable_theta : forall A t B,
     Boxed A t B
  -> forall G t', typing G t A exp_star
       -> typing G t' B exp_star.
Proof.
  induction 1; intros; eauto.
Qed.
Hint Extern 2 =>
  match goal with
(*  | [ H : Boxed exp_nat ?t ?B |- typing _ _ (exp_at ?B _) _ ] =>
    apply (Boxed_typable_theta exp_nat t) *)
  | [ H : Boxed ?x ?t ?B |- typing _ _ ?B _ ] => apply (Boxed_typable_theta x t)
  | [ H : Boxed ?x ?t ?B |- typing _ _ (exp_at ?B _) _ ] => apply (Boxed_typable_theta x t)
  end.

Theorem Boxed_typ_inversion : forall A t B,
     Boxed A t B
  -> forall G t' C, typing G t' B C
     -> typing G t A exp_star.
Proof.
  induction 1; intros; eauto;
  match goal with
  | [ H : typing _ _ (exp_at _ _) _ |- _ ] =>
    apply typ_at_regularity in H; destruct_things; eauto
  end.
Qed.
Hint Resolve Boxed_typ_inversion.

Theorem typ_mu_inversion_gen : forall G t a A,
     typing G t a A
  -> forall b, a = exp_mu b
      ->    typing ((th_log,exp_star) :: G) th_log b exp_star.
Proof.
  induction 1; intros; try match goal with
  | [ H : _ = exp_mu _  |- _] => inversion H; subst; clear H
  end; repeat (match goal with
  | [ H : typing ?G _ _ _ |- _ ] =>
    notProvenYet (ctx_wf G); assert (ctx_wf G) by eauto
  end); eauto.
Qed.

Theorem typ_mu_inversion : forall G t a B,
     typing G t (exp_mu a) B
  -> typing ((th_log,exp_star) :: G) th_log a exp_star.
Proof.
  intros G t a s H; eapply typ_mu_inversion_gen in H; eauto.
Qed.

Theorem typ_sum_inversion_gen : forall G t a A,
     typing G t a A
  -> forall B1 B2, a = exp_sum B1 B2
     -> (   A = exp_star
         /\ typing G t B1 exp_star
         /\ typing G t B2 exp_star).
Proof.
  induction 1; intros; try match goal with
  | [ H : _ = exp_sum _ _  |- _] => inversion H; subst; clear H
  end; val_inversion; first [destruct_ih; subst | eauto 10]; eauto;
  try match goal with
  | [ H : typing _ _ exp_star _ |- _ ] =>
    apply star_has_no_type in H; inversion H
  | [ H : exp_eq _ _ = exp_star |- _ ] => inversion H
  | [ H : exp_at _ _ = exp_star |- _ ] => inversion H
  | [ H : {_ ~> _}_ = exp_star |- _ ] =>
    apply eq_star in H; destruct_things; subst; eauto; star_isnt_equal
  end.    
Qed.

Theorem typ_sum_inversion : forall G t A1 A2 B,
     typing G t (exp_sum A1 A2) B
  ->    B = exp_star
     /\ typing G t A1 exp_star
     /\ typing G t A2 exp_star.
Proof.
  intros G t A1 A2 s H; eapply typ_sum_inversion_gen in H; eauto;
  destruct_things; eauto.
Qed.
Hint Extern 2 => match goal with
| [ H : typing _ _ (exp_sum ?A _) _ |- typing _ _ ?A _ ] =>
  let H' := fresh H in 
  assert (H' := typ_sum_inversion _ _ _ _ _ H);
  destruct_things; subst
| [ H : typing _ _ (exp_sum _ ?A) _ |- typing _ _ ?A _ ] =>
  let H' := fresh H in 
  assert (H' := typ_sum_inversion _ _ _ _ _ H);
  destruct_things; subst
end.

Theorem typ_sigma_inversion_gen : forall G t a A,
     typing G t a A
  -> forall B1 B2, a = exp_sigma B1 B2
     -> (   FO B1
         /\ A = exp_star
         /\ typing G t B1 exp_star
         /\ typing ((t,B1)::G) t B2 exp_star).
Proof.
  induction 1; intros; try match goal with
  | [ H : _ = exp_sigma _ _  |- _] => inversion H; subst; clear H
  end; val_inversion; first [destruct_ih; subst | eauto 10]; eauto;
  try match goal with
  | [ H : typing _ _ exp_star _ |- _ ] =>
    apply star_has_no_type in H; inversion H
  | [ H : exp_eq _ _ = exp_star |- _ ] => inversion H
  | [ H : exp_at _ _ = exp_star |- _ ] => inversion H
  | [ H : {_ ~> _}_ = exp_star |- _ ] =>
    apply eq_star in H; destruct_things; subst; eauto; star_isnt_equal
  end.
Qed.

Theorem typ_sigma_inversion : forall G t A1 A2 B,
     typing G t (exp_sigma A1 A2) B
  ->     FO A1
      /\ B = exp_star
      /\ typing G t A1 exp_star
      /\ typing ((t,A1)::G) t A2 exp_star.
Proof.
  intros G t A1 A2 B H; eapply typ_sigma_inversion_gen in H; eauto;
  destruct_things; eauto.
Qed.
Hint Extern 2 => match goal with
| [ H : typing _ _ (exp_sigma ?A _) _ |- typing _ _ ?A _ ] =>
  let H' := fresh H in 
  assert (H' := typ_sigma_inversion _ _ _ _ _ H);
  destruct_things; subst
| [ H : typing _ _ (exp_sigma _ ?A) _ |- typing _ _ ?A _ ] =>
  let H' := fresh H in 
  assert (H' := typ_sigma_inversion _ _ _ _ _ H);
  destruct_things; subst
end.


Theorem typ_eq_inversion_gen : forall G t a A,
     typing G t a A
  -> forall b1 b2, a = exp_eq b1 b2
     -> exists B1, exists B2, 
            A = exp_star
         /\ typing G th_prog b1 B1
         /\ typing G th_prog b2 B2.
Proof.
  induction 1; intros; try match goal with
  | [ H : _ = exp_eq _ _  |- _] => inversion H; subst; clear H
  end; val_inversion; first [destruct_ih; subst | eauto 10]; eauto;
  try match goal with
  | [ H : typing _ _ exp_star _ |- _ ] =>
    apply star_has_no_type in H; inversion H
  | [ H : exp_eq _ _ = exp_star |- _ ] => inversion H
  | [ H : exp_at _ _ = exp_star |- _ ] => inversion H
  | [ H : {_ ~> _}_ = exp_star |- _ ] =>
    apply eq_star in H; destruct_things; subst; eauto; star_isnt_equal
  end.
Qed.

Theorem typ_eq_inversion : forall G t A b1 b2,
     typing G t (exp_eq b1 b2) A
  -> exists B1, exists B2, 
        A = exp_star
     /\ typing G th_prog b1 B1
     /\ typing G th_prog b2 B2.
Proof.
  intros G t A1 A2 B H; eapply typ_eq_inversion_gen in H; eauto;
  destruct_things; eauto.
Qed.
Hint Extern 2 => match goal with
| [ H : typing _ _ (exp_eq ?A _) _ |- typing _ _ ?A _ ] =>
  let H' := fresh H in 
  assert (H' := typ_eq_inversion _ _ _ _ _ H);
  destruct_things; subst
| [ H : typing _ _ (exp_eq _ ?A) _ |- typing _ _ ?A _ ] =>
  let H' := fresh H in 
  assert (H' := typ_eq_inversion _ _ _ _ _ H);
  destruct_things; subst
end.

Theorem typ_succ_inversion_gen : forall G t a A,
     typing G t a A
  -> forall b, a = exp_succ b
     -> (   (   (exists p, typing G th_log p (exp_eq A exp_nat))
             \/ (exists p, exists B, exists t', Boxed exp_nat t' B /\ typing G th_log p (exp_eq A B)))
         /\ typing G t b exp_nat).
Proof.
  induction 1; intros; try match goal with
  | [ H : _ = exp_succ _  |- _] => inversion H; subst; clear H
  end; first [destruct_ih | eauto 6]; try find_contra; val_inversion;
  try match goal with
  | [ H : _ <> _ |- _ ] => split; eauto 8
  | [ H : typing _ _ (exp_succ ?b) (exp_at _ _), HV : val ?b
     |- _ ] => (* UnboxVal *)
    invert_boxes; invert_eqs; try match goal with
    | [ H : typing _ _ _ (exp_eq (exp_at _ ?t) (exp_at _ ?t))
       |- _ ] => 
       split; eauto 8 using TEqAtInv; mycrush
    | [ H : typing _ _ _ (exp_eq (exp_at _ ?t) (exp_at _ ?t'))
       |- _ ] => require_neq t t'; try find_contra; split; eauto 9
    end
  | [ HFO : FO _, HV : val _ |- _ ] => (* FO VAL *)
    split; [ eauto 6 | eauto 6 using TFOVal]
  | [ HB : Boxed exp_nat _ ?x
     |- ((exists _, typing _ _ _ (exp_eq (exp_at _ ?t) _)) \/ _) /\ _
    ] => split; [ right; eexists; exists (exp_at x t); eauto 9 using BMany
                | eauto using TFOVal ]
  | [ |- ((exists _, typing _ _ _ (exp_eq (exp_at _ _) _)) \/ _) /\ _ ] =>
    split; [eauto 10 using BOne | eauto using TFOVal ]
  | [ H : typing _ _ _ (exp_eq ({0 ~> _}_) _) |- _ ] =>
    invert_eqs; eauto 10 (* Conv case *)
  end; eauto 7.
Qed.

Theorem typ_succ_inversion : forall G t a A,
     typing G t (exp_succ a) A
  -> typing G t a exp_nat.
Proof.
  intros G t a A H; eapply typ_succ_inversion_gen in H; destruct_things; eauto.
Qed.

Theorem typ_inl_inversion_gen : forall G t a A,
     typing G t a A
  -> forall b, a = exp_inl b
     -> (exists B1, exists B2,
            (   (exists p, typing G th_log p (exp_eq A (exp_sum B1 B2)))
             /\ typing G t (exp_sum B1 B2) exp_star
             /\ typing G t b B1)
         \/ (exists t', exists C,
                Boxed (exp_sum B1 B2) t' C
             /\ typing G t' (exp_sum B1 B2) exp_star
             /\ (exists p, typing G th_log p (exp_eq A C))
             /\ typing G t' b B1)).
Proof.
  induction 1; intros; try match goal with
  | [ H : _ = exp_inl _  |- _] => inversion H; subst; clear H
  end; val_inversion; first [destruct_ih | eauto 10];
  try match goal with
    | [ H : FO ?A, Hv : val _, HB : Boxed _ _ _ |- _ ] => eauto 10
    | [ H : FO ?A, Hv : val _, HT : typing _ _ (exp_sum ?B1 ?B2) _
       |- _ ] =>
      inversion H; subst;
           try (find_contra; exists B1; exists B2; right; exists th_prog;
                invert_eqs; eauto 11 using BOne);
      match goal with
      | [ HT : typing _ th_log (exp_sum ?A ?B) exp_star,
          HEq : typing _ _ _ (exp_eq (exp_sum _ _) (exp_sum _ _)) |- _ ] =>
        apply typ_sum_inversion in HT; destruct_things; subst;
        exists A; exists B; left; split; [ eauto | split;
        [ eauto | invert_eqs; eapply TFOVal; eauto;
                  eapply TEqSumInv1 in HEq; eauto ]]
      end
  end;
  match goal with
  | [ H : Boxed (exp_sum ?A ?B) _ _ |- _ ] => exists A; exists B
  | [ H : typing _ _ _ (exp_eq _ (exp_sum ?A ?B)) |- _ ] => exists A; exists B
  end; try find_contra; try match goal with
  | [ H : _ <> _, Ht : typing _ ?t ?b ?x
     |- (_ /\ _ /\ typing _ _ ?b _) \/ _ ] => (* Contra *)
    right; exists t; invert_eqs; eauto 11 using BOne
  | [ H : typing _ _ _ (exp_eq ({_ ~> _} _) (exp_sum _ _)) |- _ ] =>
    invert_eqs; left; split; eauto 8
  | [ H : typing _ _ _ (exp_eq ({_ ~> _} _) ?B),
      HB : Boxed _ _ ?B |- _ ] => invert_eqs; right; eauto 10 
  | [ H : typing _ _ (exp_inl ?b) (exp_at _ _), HV : val ?b
     |- _ ] => (* UnboxVal *)
    invert_boxes; invert_eqs; val_inversion; try match goal with
    | [ H : typing _ _ _ (exp_eq (exp_at _ ?t) (exp_at _ ?t)),
        HB : Boxed _ ?t' ?C
       |- _ ] => is_var C; right; exists t'; exists C; eauto 8 using TEqAtInv
    | [ H : typing _ _ _ (exp_eq (exp_at _ ?t) (exp_at _ ?t))
       |- _ ] => eauto 10 using TEqAtInv
    | [ H : typing _ _ _ (exp_eq (exp_at _ ?t) (exp_at _ ?t'))
       |- _ ] => require_neq t t'; try find_contra; right; eauto 11 using BOne
    end
  | [ HB : Boxed _ _ _, HEq : typing _ _ _ (exp_eq ?A ?B)
     |- _ \/ exists _, exists _, _ /\ _ 
                /\ (exists _, typing _ _ _ (exp_eq (exp_at ?A ?t) _)) /\ _ ] =>
    right; eexists; exists (exp_at B t); split; eauto 8 using BMany
  | [ |- _ \/ exists _, exists _, _ /\ _ 
                /\ (exists _, typing _ _ _ (exp_eq (exp_at ?A ?t) _)) /\ _ ] =>
    right; eauto 10 using BOne
  end; eauto 10.
Qed.

Theorem typ_inl_inversion : forall G t a A1 A2,
     typing G t (exp_inl a) (exp_sum A1 A2)
  -> typing G t a A1.
Proof.
  intros G t a A1 A2 H;
  assert (typing G t (exp_sum A1 A2) exp_star) as H'
    by (apply regularity in H; destruct_things; eauto; inversion H);
  apply typ_sum_inversion in H'; destruct_things;
  eapply typ_inl_inversion_gen in H; eauto;
  destruct_things; invert_eqs; match goal with
  | [ H : typing _ _ _ (exp_eq (exp_sum _ _) (exp_sum _ _)) |- _ ] =>
    eapply TEqSumInv1 in H; eauto
  | [ HB : Boxed _ _ ?A,
      HEq : typing _ _ _ (exp_eq (exp_sum _ _) ?A) |- _ ] =>
    find_contra; eauto
  end.
Qed.

Theorem typ_inr_inversion_gen : forall G t a A,
     typing G t a A
  -> forall b, a = exp_inr b
     -> (exists B1, exists B2,
            (   (exists p, typing G th_log p (exp_eq A (exp_sum B1 B2)))
             /\ typing G t (exp_sum B1 B2) exp_star
             /\ typing G t b B2)
         \/ (exists t', exists C,
                Boxed (exp_sum B1 B2) t' C
             /\ typing G t' (exp_sum B1 B2) exp_star
             /\ (exists p, typing G th_log p (exp_eq A C))
             /\ typing G t' b B2)).
Proof.
  induction 1; intros; try match goal with
  | [ H : _ = exp_inr _  |- _] => inversion H; subst; clear H
  end; val_inversion; first [destruct_ih | eauto 10];
  try match goal with
    | [ H : FO ?A, Hv : val _, HB : Boxed _ _ _ |- _ ] => eauto 10
    | [ H : FO ?A, Hv : val _, HT : typing _ _ (exp_sum ?B1 ?B2) _
       |- _ ] =>
    inversion H; subst;
         try (find_contra; exists B1; exists B2; right; exists th_prog;
              invert_eqs; eauto 11 using BOne);
    match goal with
    | [ HT : typing _ th_log (exp_sum ?A ?B) exp_star,
        HEq : typing _ _ _ (exp_eq (exp_sum _ _) (exp_sum _ _)) |- _ ] =>
      apply typ_sum_inversion in HT; destruct_things; subst;
      exists A; exists B; left; split; [ eauto | split;
      [ eauto | invert_eqs; eapply TFOVal; eauto;
                eapply TEqSumInv2 in HEq; eauto ]]
    end
  end;
  match goal with
  | [ H : Boxed (exp_sum ?A ?B) _ _ |- _ ] => exists A; exists B
  | [ H : typing _ _ _ (exp_eq _ (exp_sum ?A ?B)) |- _ ] => exists A; exists B
  end; try find_contra; try match goal with
  | [ H : _ <> _, Ht : typing _ ?t ?b ?x
     |- (_ /\ _ /\ typing _ _ ?b _) \/ _ ] => (* Contra *)
    right; exists t; invert_eqs; eauto 11 using BOne
  | [ H : typing _ _ _ (exp_eq ({_ ~> _} _) (exp_sum _ _)) |- _ ] =>
    invert_eqs; left; split; eauto 8
  | [ H : typing _ _ _ (exp_eq ({_ ~> _} _) ?B),
      HB : Boxed _ _ ?B |- _ ] => invert_eqs; right; eauto 10 
  | [ H : typing _ _ (exp_inr ?b) (exp_at _ _), HV : val ?b
     |- _ ] => (* UnboxVal *)
    invert_boxes; invert_eqs; val_inversion; try match goal with
    | [ H : typing _ _ _ (exp_eq (exp_at _ ?t) (exp_at _ ?t)),
        HB : Boxed _ ?t' ?C
       |- _ ] => is_var C; right; exists t'; exists C; eauto 8 using TEqAtInv
    | [ H : typing _ _ _ (exp_eq (exp_at _ ?t) (exp_at _ ?t))
       |- _ ] => eauto 10 using TEqAtInv
    | [ H : typing _ _ _ (exp_eq (exp_at _ ?t) (exp_at _ ?t'))
       |- _ ] => require_neq t t'; try find_contra; right; eauto 11 using BOne
    end
  | [ HB : Boxed _ _ _, HEq : typing _ _ _ (exp_eq ?A ?B)
     |- _ \/ exists _, exists _, _ /\ _ 
                /\ (exists _, typing _ _ _ (exp_eq (exp_at ?A ?t) _)) /\ _ ] =>
    right; eexists; exists (exp_at B t); split; eauto 8 using BMany
  | [ |- _ \/ exists _, exists _, _ /\ _ 
                /\ (exists _, typing _ _ _ (exp_eq (exp_at ?A ?t) _)) /\ _ ] =>
    right; eauto 10 using BOne
  end; eauto 10.
Qed.

Theorem typ_inr_inversion : forall G t a A1 A2,
     typing G t (exp_inr a) (exp_sum A1 A2)
  -> typing G t a A2.
Proof.
  intros G t a A1 A2 H;
  assert (typing G t (exp_sum A1 A2) exp_star) as H'
    by (apply regularity in H; destruct_things; eauto; inversion H);
  apply typ_sum_inversion in H'; destruct_things;
  eapply typ_inr_inversion_gen in H; eauto;
  destruct_things; invert_eqs; match goal with
  | [ H : typing _ _ _ (exp_eq (exp_sum _ _) (exp_sum _ _)) |- _ ] =>
    eapply TEqSumInv2 in H; eauto
  | [ HB : Boxed _ _ ?A,
      HEq : typing _ _ _ (exp_eq (exp_sum _ _) ?A) |- _ ] =>
    find_contra; eauto
  end.
Qed.

Theorem typ_arr_inversion_gen : forall G t a A,
     typing G t a A
  -> forall B1 B2, a = exp_arr B1 B2
     ->    FO B1
        /\ A = exp_star
        /\ typing G t B1 exp_star
        /\ typing ((t,B1)::G) t B2 exp_star.
Proof.
  induction 1; intros; try match goal with
  | [ H : _ = exp_arr _ _  |- _] => inversion H; subst; clear H
  end; eauto; try (destruct_ih; subst; (split; [ auto | ]));
  try find_contra; try match goal with
  | [ H : ?A = ?B |- _ ] => solve [inversion H]
  | [ H : typing _ _ exp_star _ |- _ ] => apply star_has_no_type in H; inversion H
  end; eauto; match goal with
  | [ H : {_ ~> _}_ = exp_star |- _ ] =>
    apply eq_star in H; destruct_things; subst; eauto; star_isnt_equal
  end.
Qed.

Theorem typ_arr_inversion : forall G t A1 A2 B,
     typing G t (exp_arr A1 A2) B
  ->    FO A1
     /\ B = exp_star
     /\ typing G t A1 exp_star
     /\ typing ((t,A1)::G) t A2 exp_star.
Proof.
  intros G t A1 A2 B H; eapply typ_arr_inversion_gen in H; eauto;
  destruct_things; eauto.
Qed.

(*
Theorem typ_arr_inversion_subst : forall G t B1 B2 s2,
     typing G t (exp_arr B1 B2) (exp_sort s2)
  -> forall v, val v -> typing G t v B1
  -> typing G t ({0 ~> v}B2) (exp_sort s2).
Proof.
  intros G t B1 B2 s2 H v HV HT; eapply typ_arr_inversion_gen in H; eauto;
  destruct_things; subst; repeat box_isnt_equal;
  try match goal with
  | [ H : exp_sort ?s = exp_sort box |- _ ] => inversion H; subst
  end; eauto; mycrush.
Qed.
*)
Ltac go_go_regularity :=
  simpl in *; repeat (match goal with
  | [ H : typing _ _ _ ?A |- _ ] =>
    let HNeq := fresh "HNeq" in let HContra := fresh "HContra" in
    let HReg := fresh "HReg" in
    untyped A; assert (A = exp_star -> False) as HNeq
                   by (intro HContra; inversion HContra);
    destruct (regularity _ _ _ _ H) as [HReg | HReg];
    [ | apply HNeq in HReg; destruct HReg]; clear HNeq
  end); repeat (match goal with
  | [ H : typing ((?t,?A) :: ?G) _ _ _ |- _ ] =>
    match A with 
    | exp_star => fail 1
    | _ => extract_open_lift; match goal with
      | [ H : typing G t A _ |- _ ] => fail 1
      | _ => is_var A; fail 1
      | _ => let HReg := fresh "HReg" in
        assert (HReg := typing_regularity_ctx _ _ _ _ H);
        inversion HReg; subst; try match goal with
        | [ H1 : ctx_wf G, H2 : ctx_wf G |- _ ] => clear H2
        end
      end
    end
  end); repeat (match goal with
  | [ H : Boxed ?A ?t ?B, Ht : typing _ _ ?B _ |- _ ] =>
    match goal with
    | [ Ht' : typing _ t A exp_star |- _ ] => fail 1
    | _ => let Ht' := fresh Ht in
      assert (Ht' := Boxed_typ_inversion _ _ _ H _ _ _ Ht)
    end
  | [ H : typing ?G _ (exp_at ?b _) _ |- _ ] =>
    untyped b; let H' := fresh H in 
    assert (H' := typ_at_regularity _ _ _ _ _ H)
  | [ H : typing ?G ?t (exp_arr ?A ?B) _ |- _ ] =>
    first [untyped A | untyped B]; let H' := fresh H in
    assert (H' := typ_arr_inversion _ _ _ _ _ H); destruct_things; subst
  | [ H : typing ?G ?t (exp_arr ?A ?B) ?x |- _ ] =>
    is_var x; let H' := fresh H in
    assert (H' := typ_arr_inversion _ _ _ _ _ H);
    destruct H' as [? [? [_ _]]]; subst
  | [ H : typing _ _ (exp_sum ?A ?B) _ |- _ ] =>
    first [untyped A | untyped B];
    apply typ_sum_inversion in H; destruct_things; subst
  | [ H : typing _ _ (exp_sigma ?A ?B) _ |- _ ] =>
    first [untyped A | untyped B]; let H' := fresh H in
    assert (H' := typ_sigma_inversion _ _ _ _ _ H);
    destruct_things; subst
  | [ H : typing _ _ (exp_mu ?A) _ |- _ ] =>
    match goal with
    | [ H : typing ((th_log,exp_star) :: _) th_log A exp_star |- _ ] =>
      fail 1
    | _ => let H' := fresh H in assert (H' := typ_mu_inversion _ _ _ _ H)
    end
  | [ H : typing ?G _ (exp_eq ?a ?b) _ |- _ ] =>
    let t := fresh "t" in 
    first 
      [  untyped a; untyped b; let H' := fresh H in 
         assert (H' := typ_eq_regularity _ _ _ _ _ H); destruct_things
      | untyped a; destruct (typ_eq_regularity _ _ _ _ _ H)
                         as [? [t [? [_ [? _]]]]]; clear t
      | untyped b; destruct (typ_eq_regularity _ _ _ _ _ H)
                         as [t [? [_ [? [_ ?]]]]]; clear t
      ]
  end).

Theorem typ_roll_inversion_gen : forall G t a A,
     typing G t a A
  -> forall b, a = exp_roll b
     ->    (exists p, exists B,
               typing G th_log p (exp_eq A (exp_mu B))
            /\ typing G th_log (exp_mu B) exp_star
            /\ typing G t b ({0 ~> exp_mu B}B))
        \/ (exists p, exists t', exists B, exists C,
               Boxed (exp_mu B) t' C
            /\ typing G th_log p (exp_eq A C)
            /\ typing G t' b ({0 ~> exp_mu B}B)).
Proof.
  induction 1; intros; try match goal with
  | [ H : _ = exp_roll _  |- _] => inversion H; subst; clear H
  end; first [destruct_ih | solve [go_go_regularity; eauto 10]];
  go_go_regularity; match goal with
  | [ H : typing ((?t,exp_star)::?G) ?t ?b exp_star |- _ ] =>
    assert (typing G t ({0 ~> exp_mu b}b) exp_star) by
      (apply typing_substitution_front with (v := exp_mu b) in H; eauto)
  end; try find_contra; try match goal with | [ H : _ <> _ |- _ ] => 
    right; eauto 13 using BOne
  end; val_inversion; try match goal with
  | [ H : typing _ _ (exp_roll ?b) (exp_at _ _), HV : val ?b
     |- _ ] => (* UnboxVal *)
    invert_boxes; invert_eqs; try match goal with
    | [ H : typing _ _ _ (exp_eq (exp_at _ ?t) (exp_at _ ?t))
       |- _ ] => eauto 11 using TEqAtInv
    | [ H : typing _ _ _ (exp_eq (exp_at _ ?t) (exp_at _ ?t'))
       |- _ ] => require_neq t t'; try find_contra; right; eauto 13 using BOne
    end
  | [ H : FO ?A, Hv : val _, HB : Boxed _ _ _ |- _ ] => eauto 10
  | [ H : FO ?A, Hv : val _, HT : typing _ _ (exp_roll _) _
     |- _ ] => inversion H; subst; find_contra; right; eauto 13 using BOne
  | [ H : typing _ _ _ (exp_eq ({_ ~> _} _) (exp_mu _)) |- _ ] =>
    left; eauto 9
  | [ H : typing _ _ _ (exp_eq ({_ ~> _} _) ?B),
      HB : Boxed _ _ ?B |- _ ] => right; eauto 11
  | [ |- (exists _, exists _, typing _ _ _ (exp_eq (exp_at _ _) _)
            /\ _) \/ _ ] =>
    right; eauto 14 using BOne, BMany
  end; eauto 9.
Qed.

Theorem typ_roll_inversion_prog : forall G a A,
     typing G th_prog (exp_roll a) (exp_mu A)
  -> typing G th_prog a ({0 ~> exp_mu A}A).
Proof.
  intros G a A H;
  assert (H' := typ_roll_inversion_gen _ _ _ _ H _ eq_refl); destruct_things;
  go_go_regularity;
  repeat (match goal with
  | [ H : typing ((?t,exp_star) :: _) ?t ?b exp_star |- _ ] =>
    apply typing_substitution_front with (v := exp_mu b) in H; auto; simpl in *
  end); try match goal with
  | [ H : typing _ _ _ (exp_eq (exp_mu _) (exp_mu _)) |- _ ] =>
    try eapply TEqMuInv in H; auto
  end; try find_contra; eauto;
  try match goal with
  | [ H : typing _ ?t _ ({0 ~> ?A} ?B),
      HC : typing _ _ ({0 ~> ?A} ?B) _ |- _ ] =>
    let H' := fresh H in
    assert (H' := regularity _ _ _ _ H); destruct_things;
    [ | rewrite H' in HC; apply star_has_no_type in HC; inversion HC ]
  end; eauto.
Qed.

Hint Extern 0 =>
  match goal with
  | [ H : FO exp_star |- _ ] => inversion H
  end.

Theorem typ_nrec_inversion_gen : forall G t a A,
     typing G t a A
  -> forall b, a = exp_nrec b
     -> (   (exists p, exists B2,
                typing G th_log p (exp_eq A (exp_arr exp_nat B2))
             /\ typing G th_log (exp_arr exp_nat B2) exp_star
             /\ typing (   (th_log, exp_arr exp_nat 
                             (exp_arr (exp_eq (exp_succ (exp_var 0)) (exp_var 1))
                                      (lift (lift B2 1 1) 1 0)))
                        :: (th_log,exp_nat) :: G)
                       th_log b (lift B2 1 0))
         \/ (exists p, exists B2, exists B, exists t',
                Boxed (exp_arr exp_nat B2) t' B
             /\ typing G th_log p (exp_eq A B)
             /\ typing G th_log (exp_arr exp_nat B2) exp_star
             /\ typing (   (th_log, exp_arr exp_nat
                             (exp_arr (exp_eq (exp_succ (exp_var 0)) (exp_var 1))
                                      (lift (lift B2 1 1) 1 0)))
                        :: (th_log,exp_nat) :: G) th_log b (lift B2 1 0))).
Proof.
  intros G t a A; induction 1; intros; try match goal with
  | [ H : _ = exp_nrec _  |- _] => inversion H; subst; clear H
  end; first [destruct_ih | solve [eauto 10]];
  try match goal with
  | [ H : FO _ |- _ ] => (* TFOVal *)
    inversion H; subst; find_contra; right; 
    repeat eexists; eauto 7 using BOne
  end; try find_contra; try match goal with
  | [ H : _ <> _ |- _ ]  => right; repeat eexists; eauto 7 using BOne
  | [ H : typing _ _ ?p (exp_eq ({_ ~> _}_) (exp_arr ?B1 ?B2)) |- _ ] =>
    (* TConv *) left; exists p; exists B1; exists B2; eauto 8
  | [ H : typing _ _ ?p (exp_eq ({_ ~> _}_) ?B), HB : Boxed _ _ ?B |- _ ] =>
    (* TConv *) right; eauto 14
  | [ HV : val (exp_nrec _), HT : typing _ _ (exp_nrec _) (exp_at _ _) |- _ ] =>
    (* unboxval *)  invert_boxes; try match goal with
    | [ H : typing _ _ _ (exp_eq (exp_at _ ?t) (exp_at ?B ?t))
       |- _ ] => is_var B; 
       right; eauto 14 using TEqAtInv
    | [ H : typing _ _ _ (exp_eq (exp_at _ ?t) (exp_at _ ?t))
       |- _ ] => left; repeat eexists; eauto 7 using TEqAtInv
    | [ H : typing _ _ _ (exp_eq (exp_at _ ?t) (exp_at _ ?t'))
       |- _ ] => require_neq t t';  try find_contra; right;
                 repeat eexists; eauto 7 using BOne
    end
  | [ |- (exists _, exists _,
            typing _ _ _ (exp_eq (exp_at _ _) _) /\ _) \/ _ ] =>
    go_go_regularity; auto 1; right; eexists; eexists; eexists; eexists;
    split; 
    [ match goal with
      | [ H : Boxed _ _ _ |- _ ] => eapply BMany; repeat split; eauto
      | _ => eapply BOne; eauto end | repeat split; eauto 7 using eq_cong_At]
  end; go_go_regularity; eauto 16.
Qed.

Theorem TVar0 : forall t A G,
     ctx_wf ((t,A) :: G)
  -> typing ((t,A) :: G) t (exp_var 0) (lift A 1 0).
Proof.
  intros t A G ?;
  replace ((t,A) :: G) with (nil ++ (t,A) :: G) in * by auto;
  eauto.
Qed.

Theorem TVar0_sub : forall t A G,
     ctx_wf ((t,A) :: G)
  -> typing ((t,A) :: G) th_prog (exp_var 0) (lift A 1 0).
Proof.
  intros t A G ?; destruct t; eauto using TSub, TVar0.
Qed.
Hint Resolve TVar0_sub.

Theorem TVar0_closed : forall t A G,
     ctx_wf ((t,A) :: G)
  -> fvs A = 0
  -> typing ((t,A) :: G) t (exp_var 0) A.
Proof.
  intros t A G ? ?;
  erewrite <- lift_closed_id; eauto using TVar0.
Qed.
Hint Resolve TVar0_closed.

Theorem TVar0_closed_sub : forall t A G,
     ctx_wf ((t,A) :: G)
  -> fvs A = 0
  -> typing ((t,A) :: G) th_prog (exp_var 0) A.
Proof.
  intros t A G ? ?; destruct t; eauto using TSub.
Qed.
Hint Resolve TVar0_closed_sub.

Theorem typing_weakening_second_cons : forall t1 t2 B1 B2 G3 t a A,
     typing ((t1,B1) :: G3) t a A 
  -> ctx_wf ((t2,B2) :: G3)
  -> fvs B1 = 0
  -> typing ((t1,B1) :: (t2,B2) :: G3) t 
            (lift a 1 1) 
            (lift A 1 1).
Proof.
  intros.
  replace ((t1,B1) :: (t2,B2) :: G3)
     with (   (lift_c ((t1,B1) :: nil) (length ((t2,B2) :: nil)) 0)
           ++ ((t2,B2) :: nil) ++ G3)
       by (simpl; autorewrite with step; auto).
  repeat (match goal with
  | [ |- context [lift ?a 1 1] ] =>
    replace (lift a 1 1)
       with (lift a (length ((t2,B2) :: nil)) (length ((t1,B1)::nil)))
         by (simpl; eauto)
          end).

  apply typing_weakening; eauto.
Qed.  

Theorem open_var_eq_gen : forall A n,
  {n ~> exp_var n} (lift A 1 (S n)) = A.
Proof.
  induction A; intros; eauto;
  try solve [simpl; db_var_case; rewrite_ih; eauto].
Qed.

Theorem typ_nrec_inversion_subst : forall G t b A1 A2,
     typing G t (exp_nrec b) (exp_arr A1 A2)
  -> forall v, val v -> typing G t v A1
       -> typing G t
                 ({0 ~> v} 
                   ({0 ~> exp_lam (exp_lam (
                             exp_app (exp_nrec (lift b 3 2))
                                     (exp_var 1)))} 
                      b))
                 ({0 ~> v}A2).
Proof.
  intros G t b A1 A2 HT; go_go_regularity; assert (HT' := HT);
  eapply typ_nrec_inversion_gen in HT; eauto;
  destruct_things; go_go_regularity; intros.
  match goal with
  | [ HArr1 : typing ?G _ (exp_arr ?A1 ?A2) exp_star,
      HArr2 : typing _ _ (exp_arr exp_nat ?B2) exp_star, 
      HEq : typing _ _ ?p (exp_eq (exp_arr ?A1 ?A2) (exp_arr exp_nat ?B2)),
      HbT : typing ((th_log,exp_arr exp_nat 
                             (exp_arr (exp_eq (exp_succ (exp_var 0))
                                             (exp_var 1))
                                     (lift (lift ?B2 1 1) 1 0))) :: (th_log,_) :: _)
                   th_log ?b (lift ?B2 1 0)
     |- typing _ ?t _ ({0 ~> ?v'}?A2) ] =>
    extract_open_lift;
    assert (typing G th_log p (exp_eq A1 exp_nat)) by eauto;
    assert (typing G t v' exp_nat) by eauto 2;
    assert (typing G th_log v' exp_nat) by (destruct t; eauto 3 using TFOVal);
    assert (typing G t ({0 ~> v'}A2) exp_star) by eauto;
    assert (typing G t ({0 ~> v'}B2) exp_star) by eauto;
    assert (typing G th_log p (exp_eq ({0 ~> v'}A2) ({0 ~> v'}B2)))
      by (eapply TEqArrInv2; eauto)
  end.


  apply typing_substitution_front_lifted
   with (v := exp_lam (exp_lam (exp_app (exp_nrec (lift b 3 2)) (exp_var 1))))
     in H5.

  eapply typing_substitution_front in H5; try eassumption; eauto.
  eauto.
  eapply TLam; eauto.  eapply TLam; eauto.
  extract_open_lift.

  match goal with
  | [ |- context[ exp_app (lift (exp_nrec ?a) (S ?n) 0) (exp_var 1) ] ] =>
    replace (exp_app (lift (exp_nrec a) (S n) 0) (exp_var 1))
       with (lift (exp_app (lift (exp_nrec a) n 0) (exp_var 0)) 1 0) in *
          by (simpl; autorewrite with step; auto)
  end.
  apply typing_weakening_cons; eauto.
  replace (lift (exp_nrec b) 2 0) with (lift (lift (exp_nrec b) 1 0) 1 1)
    by (autorewrite with step; eauto).


  replace (exp_app (lift (lift (exp_nrec b) 1 0) 1 1) (exp_var 0))
     with (lift (exp_app (lift (exp_nrec b) 1 0) (exp_var 0)) 1 1)
       by (simpl; eauto).
  apply typing_weakening_second_cons; eauto.
  rewrite <- (open_var_eq_gen x0 0).
  eapply TApp with (A := exp_nat).
  replace (exp_arr exp_nat (lift x0 1 1))
     with (lift (exp_arr exp_nat x0) 1 0)
       by eauto.
  apply typing_weakening_cons.
  tconstructor.  eauto.  eauto.  eauto.  eauto using TVar0_closed.
  rewrite open_var_eq_gen; eauto.
  (***** END CASE 1 *****)

  match goal with
  | [ HArr1 : typing ?G _ (exp_arr ?A1 ?A2) exp_star,
      HArr2 : typing _ _ (exp_arr exp_nat ?B2) exp_star,
      HB : Boxed (exp_arr exp_nat ?B2) _ _,
      HbT : typing ((th_log,_)::(th_log,exp_nat)::_) _ ?b (lift ?B2 1 0)
     |- typing _ ?t _ ({0 ~> ?v'}?A2) ] =>
    assert (typing G t ({0 ~> v'}A2) exp_star) by eauto;
    assert (typing G th_log exp_join (exp_eq A1 exp_nat)) 
        by (find_contra; eapply TContra; eauto);
    assert (typing G th_log v exp_nat) by eauto
  end.

  assert (typing ((th_log,exp_nat) :: G) th_log 
                 (exp_lam (exp_lam 
                   (exp_app (lift (exp_nrec b) 3 0) (exp_var 1))))
                 (exp_arr exp_nat 
                   (exp_arr (exp_eq (exp_succ (exp_var 0)) (exp_var 1))
                            (lift (lift x0 1 1) 1 0)))).
    repeat (apply TLam; try eassumption).

    match goal with
    | [ |- context[ exp_app (lift (exp_nrec ?a) (S ?n) 0) (exp_var 1) ] ] =>
      replace (exp_app (lift (exp_nrec a) (S n) 0) (exp_var 1))
         with (lift (exp_app (lift (exp_nrec a) n 0) (exp_var 0)) 1 0) in *
            by (simpl; autorewrite with step; auto)
    end.
    apply typing_weakening_cons; eauto.
    replace (lift (exp_nrec b) 2 0) with (lift (lift (exp_nrec b) 1 0) 1 1)
      by (autorewrite with step; eauto).
    replace (exp_app (lift (lift (exp_nrec b) 1 0) 1 1) (exp_var 0))
       with (lift (exp_app (lift (exp_nrec b) 1 0) (exp_var 0)) 1 1)
         by (simpl; eauto).
    apply typing_weakening_second_cons; eauto.
    rewrite <- (open_var_eq_gen x0 0).
    eapply TApp with (A := exp_nat); eauto using TVar0_closed.
      replace (exp_arr exp_nat (lift x0 1 1))
         with (lift (exp_arr exp_nat x0) 1 0)
           by eauto.
      apply typing_weakening_cons; eauto using TVar0_closed.  
      mycrush; eauto using TVar0_closed.
      rewrite open_var_eq_gen; eauto.

  (* end assertion *)
                          

  assert (H6' := H6).
  apply typing_substitution_front_lifted
   with (v := exp_lam (exp_lam (exp_app (lift (exp_nrec b) 3 0) (exp_var 1))))
     in H6'.  
  apply typing_substitution_front with (v := v) in H6'; eauto 3.

  find_contra; eauto.  eauto.  eauto.
Qed.

Theorem typ_fun_inversion_gen : forall G t a A,
     typing G t a A
  -> forall b, a = exp_fun b
      -> (   (exists p, exists B1, exists B2,
                 typing G th_log p (exp_eq A (exp_arr B1 B2))
              /\ typing G t (exp_arr B1 B2) exp_star
              /\ typing (   (th_prog, lift (exp_arr B1 B2) 1 0)
                         :: (t,B1) :: G)
                        t b (lift B2 1 0))
          \/ (exists p, exists B1, exists B2, exists B, exists t',
                 Boxed (exp_arr B1 B2) t' B
              /\ typing G th_log p (exp_eq A B)
              /\ typing G t' (exp_arr B1 B2) exp_star
              /\ typing (   (th_prog, lift (exp_arr B1 B2) 1 0)
                         :: (t',B1) :: G) t' b (lift B2 1 0))).
Proof.
  intros G t a A; induction 1; intros; try match goal with
  | [ H : _ = exp_fun _  |- _] => inversion H; subst; clear H
  end; first [destruct_ih | eauto 10];
  try match goal with
  | [ H : FO _ |- _ ] => (* TFOVal *)
    inversion H; subst; find_contra; right; eauto 14 using BOne
  end; try find_contra; try match goal with
  | [ H : _ <> _ |- _ ]  => right; eauto 14 using BOne
  | [ H : typing _ _ ?p (exp_eq ({_ ~> _}_) (exp_arr ?B1 ?B2)) |- _ ] =>
    (* TConv *) left; exists p; exists B1; exists B2; eauto 8
  | [ H : typing _ _ ?p (exp_eq ({_ ~> _}_) ?B), HB : Boxed _ _ ?B |- _ ] =>
    (* TConv *) right; eauto 14
  | [ HV : val (exp_fun _), HT : typing _ _ (exp_fun _) (exp_at _ _) |- _ ] =>
    (* unboxval *)  invert_boxes; try match goal with
    | [ H : typing _ _ _ (exp_eq (exp_at _ ?t) (exp_at _ ?t))
       |- _ ] => eauto 14 using TEqAtInv
    | [ H : typing _ _ _ (exp_eq (exp_at _ ?t) (exp_at _ ?t'))
       |- _ ] => require_neq t t';  try find_contra; right; eauto 14 using BOne
    end
  | [ |- (exists _, exists _, exists _,
            typing _ _ _ (exp_eq (exp_at _ _) _) /\ _) \/ _ ] =>
    go_go_regularity; auto 1; right; eexists; eexists; eexists; eexists; eexists;
    split; 
    [ match goal with
      | [ H : Boxed _ _ _ |- _ ] => eapply BMany; eauto
      | _ => eapply BOne; eauto end | split; eauto 7 using eq_cong_At]
  end; go_go_regularity; eauto 14.
Qed.

Theorem typ_fun_inversion_subst : forall G t b A1 A2,
     typing G t (exp_fun b) (exp_arr A1 A2)
  -> forall v, val v -> typing G t v A1 
           -> typing G t ({0 ~> v} ({0 ~> exp_fun (lift b 1 2)} b))
                       ({0 ~> v}A2).
Proof.
  intros G t b A1 A2 HT; go_go_regularity; assert (HT' := HT);
  eapply typ_fun_inversion_gen in HT; eauto;
  destruct_things; go_go_regularity; intros;
  match goal with
  | [ HArr1 : typing ?G _ (exp_arr ?A1 ?A2) exp_star,
      HArr2 : typing _ _ (exp_arr ?B1 ?B2) exp_star,
      HEq : typing _ _ ?p (exp_eq (exp_arr ?A1 ?A2) (exp_arr ?B1 ?B2)),
      HbT : typing ((th_prog,lift (exp_arr ?B1 ?B2) 1 0) :: (?t,?B1) :: _)
                   ?t ?b (lift ?B2 1 0)
     |- typing _ ?t _ ({0 ~> ?v'}?A2) ] =>  extract_open_lift;
    assert (typing G th_log p (exp_eq A1 B1)) by eauto;
    assert (typing G t v' B1) by eauto;
    assert (typing G t ({0 ~> v'}A2) exp_star) by eauto;
    assert (typing G t ({0 ~> v'}B2) exp_star) by eauto;
    assert (typing G th_log p (exp_eq ({0 ~> v'}A2) ({0 ~> v'}B2)))
      by (eapply TEqArrInv2; eauto);
    apply typing_substitution_front_lifted
     with (v := (lift (exp_fun b) 1 0)) in HbT;
    [ eapply typing_substitution_front in HbT; try eassumption; eauto
    | eauto | eauto]
  | [ HArr1 : typing ?G _ (exp_arr ?A1 ?A2) exp_star,
      HArr2 : typing _ _ (exp_arr ?B1 ?B2) exp_star,
      HB : Boxed (exp_arr ?B1 ?B2) _ _,
      HbT : typing ((th_prog,_)::(?t,?B1)::_) _ ?b (lift ?B2 1 0)
     |- typing _ ?t' _ ({0 ~> ?v'}?A2) ] =>
    assert (typing G t' ({0 ~> v'}A2) exp_star) by eauto;
    apply typing_substitution_front_lifted with (v := lift (exp_fun b) 1 0) in HbT;
      [ apply typing_substitution_front with (v := v') in HbT;
        find_contra; eauto
      | eauto | find_contra; eauto ]
  end.
Qed.

Theorem typ_lam_inversion_gen : forall G t a A,
     typing G t a A
  -> forall b, a = exp_lam b
      -> (   (exists p, exists B1, exists B2,
                 typing G th_log p (exp_eq A (exp_arr B1 B2))
              /\ typing G t (exp_arr B1 B2) exp_star
              /\ typing ((t,B1) :: G) t b B2)
          \/ (exists p, exists B1, exists B2, exists B, exists t',
                 Boxed (exp_arr B1 B2) t' B
              /\ typing G th_log p (exp_eq A B)
              /\ typing G t' (exp_arr B1 B2) exp_star
              /\ typing ((t',B1) :: G) t' b B2)).
Proof.
  intros G t a A; induction 1; intros; try match goal with
  | [ H : _ = exp_lam _  |- _] => inversion H; subst; clear H
  end; first [destruct_ih | eauto 10]; try match goal with
  | [ H : FO _ |- _ ] => (* TFOVal *)
    inversion H; subst; find_contra; right; eauto 14 using BOne
  end; try find_contra; try match goal with
  | [ H : _ <> _ |- _ ]  => right; eauto 14 using BOne
  end; try match goal with
  | [ H : typing _ _ ?p (exp_eq ({_ ~> _}_) (exp_arr ?B1 ?B2)) |- _ ] =>
    (* TConv *) left; exists p; exists B1; exists B2; eauto 8
  | [ H : typing _ _ ?p (exp_eq ({_ ~> _}_) ?B), HB : Boxed _ _ ?B |- _ ] =>
    (* TConv *) right; eauto 14
  | [ HV : val (exp_lam _), HT : typing _ _ (exp_lam _) (exp_at _ _) |- _ ] =>
    (* unboxval *)  invert_boxes; try match goal with
    | [ H : typing _ _ _ (exp_eq (exp_at _ ?t) (exp_at _ ?t))
       |- _ ] => eauto 14 using TEqAtInv
    | [ H : typing _ _ _ (exp_eq (exp_at _ ?t) (exp_at _ ?t'))
       |- _ ] => require_neq t t';  try find_contra; right; eauto 14 using BOne
    end
  | [ |- (exists _, exists _, exists _,
            typing _ _ _ (exp_eq (exp_at _ _) _) /\ _) \/ _ ] =>
    go_go_regularity; auto; right; eauto 16 using BOne, BMany, eq_cong_At
  end; go_go_regularity; eauto 14.
Qed.

Theorem typ_lam_inversion_subst : forall G t b A1 A2,
     typing G t (exp_lam b) (exp_arr A1 A2)
  -> forall v, val v -> typing G t v A1 
           -> typing G t ({0 ~> v}b) ({0 ~> v}A2).
Proof.
  intros G t b A1 A2 HT; go_go_regularity;
  eapply typ_lam_inversion_gen in HT; eauto;
  destruct_things; go_go_regularity; intros;
  match goal with
  | [ HArr1 : typing ?G _ (exp_arr ?A1 ?A2) exp_star,
      HArr2 : typing _ _ (exp_arr ?B1 ?B2) exp_star,
      HEq : typing _ _ ?p (exp_eq (exp_arr ?A1 ?A2) (exp_arr ?B1 ?B2)),
      HbT : typing ((?t,?B1)::_) ?t _ ?B2
     |- typing _ ?t _ ({0 ~> ?v}?A2) ] =>
    (* Ugly explicit forward reasoning, but automating will be
       so painful *)
    assert (typing G th_log p (exp_eq A1 B1)) by eauto;
    assert (typing G t v B1) by eauto;
    assert (typing G t ({0 ~> v}A2) exp_star) by eauto;
    assert (typing G t ({0 ~> v}B2) exp_star) by eauto;
    assert (typing G th_log p (exp_eq ({0 ~> v}A2) ({0 ~> v}B2)))
      by (eapply TEqArrInv2; eauto);
    eapply typing_substitution_front in HbT; try eassumption;
    eauto
  | [ HArr1 : typing ?G _ (exp_arr ?A1 ?A2) exp_star,
      HArr2 : typing _ _ (exp_arr ?B1 ?B2) exp_star,
      HB : Boxed (exp_arr ?B1 ?B2) _ _,
      HbT : typing ((?t',?B1)::_) ?t' ?b ?B2
     |- typing _ ?t _ ({0 ~> ?v}?A2) ] =>
    assert (typing G t' ({0 ~> v}b) ({0 ~> v}B2))
       by (eapply typing_substitution_front in HbT; try eassumption;
          find_contra; eauto);
    assert (typing G t ({0 ~> v}A2) exp_star) by eauto;
    find_contra; eauto
  end.
Qed.

(*
Ltac typing_inversion :=
  repeat (match goal with
  | [ H : typing ?G _ ?v (typ_at ?A ?t) |- _ ] =>
    notProvenYet (typing G t v A);
    let HV := fresh "HV" in assert (val v) as HV by eauto;
    let H' := fresh H in assert (H' := TUnboxVal _ _ _ _ _ H HV)
  end); repeat (match goal with
  | [ H : typing _ _ (exp_succ _) typ_nat |- _ ] =>
    apply typ_succ_inversion in H; destruct_things; subst
  | [ H : typing _ _ (exp_inl _) _ |- _ ] =>
    apply typ_inl_inversion in H; destruct_things; subst
  | [ H : typing _ _ (exp_inr _) _ |- _ ] =>
    apply typ_inr_inversion in H; destruct_things; subst
  | [ H : typing _ _ (exp_lam _) (typ_arr _ _) |- _] =>
    apply typ_lam_inversion in H; destruct_things
  | [ H : typing _ _ (exp_fun _) _ |- _] =>
    apply typ_fun_inversion in H; destruct_things
  | [ H : typing _ _ (exp_app _ _) _ |- _ ] =>
    apply typ_app_inversion in H; destruct_things
  | [ H : typing _ _ (exp_natcase _ _ _) _ |- _ ] =>
    apply typ_natcase_inversion in H; destruct_things; subst
  | [ H : typing _ _ (exp_sumcase _ _ _) _ |- _ ] =>
    apply typ_sumcase_inversion in H; destruct_things; subst
  | [ H : typing _ _ (exp_roll _) _ |- _ ] =>
    apply typ_roll_inversion in H; destruct_things; subst
  | [ H : typing _ _ (exp_unroll _) _ |- _ ] =>
    apply typ_unroll_inversion in H; destruct_things; subst
  end); subst.
*)

Theorem typ_pair_inversion_gen : forall G t a A,
     typing G t a A
  -> forall b1 b2, a = exp_pair b1 b2
      -> (   (exists p, exists B1, exists B2,
                 typing G th_log p (exp_eq A (exp_sigma B1 B2))
              /\ typing G t (exp_sigma B1 B2) exp_star
              /\ typing G t b1 B1
              /\ typing G t ({0~>b1}B2) exp_star
              /\ typing G t b2 ({0~>b1}B2))
          \/ (exists p, exists B1, exists B2, exists B, exists t',
                 Boxed (exp_sigma B1 B2) t' B
              /\ typing G th_log p (exp_eq A B)
              /\ typing G t' (exp_sigma B1 B2) exp_star
              /\ typing G t' b1 B1
              /\ typing G t' ({0~>b1}B2) exp_star
              /\ typing G t' b2 ({0~>b1}B2))).
Proof.
  intros G t a A; induction 1; intros; try match goal with
  | [ H : _ = exp_pair _ _  |- _] => inversion H; subst; clear H
  end; first [destruct_ih | solve [eauto 10]]; try match goal with
    | [ H : FO ?A, Hv : val _, HB : Boxed _ _ _ |- _ ] => eauto 15
    | [ H : FO ?A, Hv : val (exp_pair ?b1 ?b2),
        HT : typing ?G _ (exp_sigma ?B1 ?B2) _ |- _ ] =>
      inversion H; subst; try (find_contra; right; eauto 16 using BOne);
      match goal with
      | [ HT : typing _ th_log (exp_sigma ?A ?B) exp_star,
          HEq : typing _ _ _ (exp_eq (exp_sigma _ _) (exp_sigma _ _)) |- _ ] =>
        apply typ_sigma_inversion in HT; destruct_things; subst;
        inversion Hv; subst; left; eexists; exists A; exists B;
        assert (typing G th_log b1 A) by
            (eapply TFOVal; eauto; eapply TEqSigmaInv1 in H3; eauto);
        assert (typing G th_log (({0 ~> b1}) B) exp_star) by eauto;
        repeat (split; [ eauto | ]); apply TFOVal; eauto;
        eapply TEqSigmaInv2 with (v := b1) in H3; eauto
      end
  end; try find_contra; try match goal with
  | [ H : _ <> _ |- _ ]  => right; eauto 16 using BOne
  end; try match goal with
  | [ H : typing _ _ ?p (exp_eq ({_ ~> _}_) (exp_sigma ?B1 ?B2)) |- _ ] =>
    (* TConv *) left; exists p; exists B1; exists B2; eauto 8
  | [ H : typing _ _ ?p (exp_eq ({_ ~> _}_) ?B), HB : Boxed _ _ ?B |- _ ] =>
    (* TConv *) right; eauto 15
  | [ HV : val (exp_pair _ _),
      HT : typing _ _ (exp_pair _ _) (exp_at _ _) |- _ ] =>
    (* unboxval *)  invert_boxes; try match goal with
    | [ H : typing _ _ _ (exp_eq (exp_at _ ?t) (exp_at _ ?t))
       |- _ ] => eauto 15 using TEqAtInv
    | [ H : typing _ _ _ (exp_eq (exp_at _ ?t) (exp_at _ ?t'))
       |- _ ] => require_neq t t';  try find_contra; right; eauto 16 using BOne
    end
  | [ H : Boxed _ _ _ |- (exists _, exists _, exists _,
            typing _ _ _ (exp_eq (exp_at _ _) _) /\ _) \/ _ ] =>
    go_go_regularity; auto; right;
    eauto 16 using BOne, BMany, eq_cong_At

  | [ |- (exists _, exists _, exists _,
            typing _ _ _ (exp_eq (exp_at _ _) _) /\ _) \/ _ ] =>
    go_go_regularity; auto; right; repeat eexists; repeat split;
    eauto 11 using BOne, BMany, eq_cong_At
  end; go_go_regularity; eauto 14.
Qed.

(* The "val" hypothesis here is unfortunate *)
Theorem typ_pair_val_inversion : forall G t a b A B,
     typing G t (exp_pair a b) (exp_sigma A B)
  -> val a
  ->    typing G t a A
     /\ typing G t b ({0 ~> a}B).
Proof.
  intros G t a b A B H Hv;
  assert (typing G t (exp_sigma A B) exp_star) as H'
    by (apply regularity in H; destruct_things; eauto; inversion H);
  apply typ_sigma_inversion in H'; destruct_things;
  eapply typ_pair_inversion_gen in H; eauto; destruct_things; invert_eqs;
  match goal with
  | [ HB : typing (_ :: _) ?t' ?B exp_star,
      H : typing ?G ?t ?p (exp_eq (exp_sigma ?A ?B)
                                  (exp_sigma ?A' ?B')) |- _ ] =>
    assert (typing G t p (exp_eq A A')) by eauto;
    assert (typing G t' ({0 ~> a} B) exp_star) by eauto;
    assert (typing G t p (exp_eq ({0 ~>a}B) ({0 ~> a}B'))) by eauto;
    eauto
  | [ HB : Boxed _ _ ?A,
      HEq : typing _ _ _ (exp_eq (exp_sigma _ _) ?A) |- _ ] =>
    find_contra; eauto 8
  end.
Qed.


(* Preservation proper *)
Theorem pstep_preservation :
    forall G t a A, typing G t a A
 -> forall a', pstep a a' -> typing G t a' A.
Proof.
  induction 1; intros; subst; simpl in *;
  eauto using TSub, TEqAtInv, TEqArrInv1, TEqArrInv2,
              TEqSumInv1, TEqSumInv2, TEqMuInv,
              TEqSigmaInv1, TEqSigmaInv2;
  try match goal with
  | [ H : pstep _ _ |- typing _ _ _ _ ] => inversion H; subst; eauto
  end; repeat (match goal with
  | [ Hstep : pstep ?a ?b, Hv : val ?a |- _ ] =>
    notProvenYet (val b); assert (val b) by eauto 3
  | [ Hstep : pstep ?a ?b, IH : forall b, pstep ?a b -> _ |- _ ] =>
    let Hstep' := fresh "Hstep" in
    assert (pstep a b) as Hstep' by eauto;
    apply IH in Hstep'; clear IH
  | [ Hstep1 : pstep ?a1 ?b1, Hstep2 : pstep ?a2 ?b2,
      IH : forall b, pstep (?P ?a1 ?a2) b -> _ |- _ ] =>
    let Hstep' := fresh "Hstep" in
    assert (pstep (P a1 a2) (P b1 b2)) as Hstep' by eauto;
    apply IH in Hstep'; clear IH
  | [ Hstep : pstep ?a ?b, IH : forall b, pstep (?P (?Q ?a)) b -> _ |- _ ] =>
    let Hstep' := fresh "Hstep" in
    assert (pstep (P (Q a)) (P (Q b))) as Hstep' by eauto;
    apply IH in Hstep'; clear IH
  | [ Hstep : pstep ?a ?b, IH : forall b, pstep (?P ?a) b -> _ |- _ ] =>
    let Hstep' := fresh "Hstep" in
    assert (pstep (P a) (P b)) as Hstep' by eauto;
    apply IH in Hstep'; clear IH
  end); val_inversion; try solve [ mycrush_cs ].

  assert (HFoo := typ_fun_inversion_subst _ _ _ _ _ Hstep0 _ H3 Hstep).
  mycrush_cs.

  assert (HFoo := typ_lam_inversion_subst _ _ _ _ _ Hstep0 _ H3 Hstep).
  mycrush_cs.
  
  assert (HFoo := typ_nrec_inversion_subst _ _ _ _ _ Hstep0 _ H3 Hstep).
  mycrush_cs.

  (* Messy messy really need to automate this idea *)
  invert_eqs; apply typ_sigma_inversion in HReg0;  destruct_things.
  apply TUnboxVal in Hstep0; eauto.
  apply typ_pair_val_inversion in Hstep0; eauto; destruct_things.
  match goal with
  | [ H : typing (?b1 :: ?b2 :: (?t',?A1) :: ?G) ?t ?a _,
      Hval1 : val ?v1', Hv1 : typing ?G ?t' ?v1' ?A1
     |- typing _ _ ({_ ~> ?v1'} ({_ ~> _} ({_ ~> exp_join} ?a))) _ ] =>
    replace (b1 :: b2 :: (t',A1) :: G)
       with ((b1 :: b2 :: nil) ++ (t',A1) :: G) in H by (simpl; auto);
    apply (typing_substitution _ _ _ _ _ _ _ _ Hval1 Hv1) in H;
    simpl in H; autorewrite with step in *
  end;
  match goal with
  | [ H : typing (?b1 :: (?t',?A2) :: ?G) ?t ({_ ~> _}?a) _,
      Hval2 : val ?v2', Hv2 : typing ?G ?t' ?v2' ?A2
     |- typing _ _ ({_ ~> ?v1'} ({_ ~> lift ?v2' 1 0} ({_ ~> exp_join} ?a))) _ ] =>
    replace (b1 :: (t',A2) :: G)
       with ((b1 :: nil) ++ (t',A2) :: G) in H by (simpl; auto);
    apply (typing_substitution _ _ _ _ _ _ _ _ Hval2 Hv2) in H;
    simpl in H; autorewrite with step in *
  end; rewrite open_open_swap by auto;
  repeat (rewrite open_open_swap with (a2 := exp_join) by auto; simpl);
  mycrush_cs.

  apply TPCase with (t' := t') (A1 := A1) (A2 := A2); eauto.
  go_go_regularity.
  match goal with
  | [ H : pstep ?a ?b
     |- typing ((th_log,exp_eq (?P (exp_var 1) (exp_var 0)) (lift ?b 2 0))
                 :: _) _ _ _] =>
    replace (exp_eq (P (exp_var 1) (exp_var 0)) (lift b 2 0))
       with ({0 ~> lift b 2 0} (exp_eq (P (exp_var 2) (exp_var 1)) (exp_var 0)))
         by (simpl; auto);
    eapply ctx_conv_typing_cons with (b1 := lift a 2 0); eauto
  end; mycrush.

  apply typ_succ_inversion in Hstep0.
  match goal with
  | [ H : typing (?b1 :: (?t',?A) :: ?G) ?t ?a _, Hval : val ?v',
      Hv : typing ?G ?t' ?v' ?A
     |- typing _ _ ({_ ~> ?v'} ({_ ~> exp_join} ?a)) _ ] =>
    replace (b1 :: (t,A) :: G) with ((b1 :: nil) ++ (t,A) :: G)
         in H by (simpl; auto);
    apply (typing_substitution _ _ _ _ _ _ _ _ Hval Hv) in H;
    simpl in H; rewrite open_open_swap; mycrush_cs
  end.

  apply TNCase; eauto;
  match goal with
  | [ H : pstep ?a ?b
     |- typing ((th_log,exp_eq exp_zero ?b) :: _) _ _ _ ] =>
    replace (exp_eq exp_zero b) with ({0 ~> b} (exp_eq exp_zero (exp_var 0)))
         by (simpl; auto); mycrush_cs
  | [ H : pstep ?a ?b
     |- typing ((th_log,exp_eq (exp_succ (exp_var 0)) (lift ?b 1 0))
                 :: _) _ _ _] =>
    replace (exp_eq (exp_succ (exp_var 0)) (lift b 1 0))
       with ({0 ~> lift b 1 0} (exp_eq (exp_succ (exp_var 1)) (exp_var 0)))
         by (simpl; auto); mycrush_cs
  end.

  invert_eqs.
  eapply TUnboxVal in Hstep0; eauto.
  eapply typ_inl_inversion in Hstep0.
  match goal with
  | [ H : typing (?b1 :: (?t',?A) :: ?G) ?t ?a _, Hval : val ?v',
      Hv : typing ?G ?t' ?v' ?A
     |- typing _ _ ({_ ~> ?v'} ({_ ~> exp_join} ?a)) _ ] =>
    replace (b1 :: (t',A) :: G) with ((b1 :: nil) ++ (t',A) :: G)
         in H by (simpl; auto);
    apply (typing_substitution _ _ _ _ _ _ _ _ Hval Hv) in H;
    simpl in H; rewrite open_open_swap; mycrush_cs
  end.

  invert_eqs.
  eapply TUnboxVal in Hstep0; eauto.
  eapply typ_inr_inversion in Hstep0.
  match goal with
  | [ H : typing (?b1 :: (?t',?A) :: ?G) ?t ?a _, Hval : val ?v',
      Hv : typing ?G ?t' ?v' ?A
     |- typing _ _ ({_ ~> ?v'} ({_ ~> exp_join} ?a)) _ ] =>
    replace (b1 :: (t',A) :: G) with ((b1 :: nil) ++ (t',A) :: G)
         in H by (simpl; auto);
    apply (typing_substitution _ _ _ _ _ _ _ _ Hval Hv) in H;
    simpl in H; rewrite open_open_swap; mycrush_cs
  end.

  apply TSCase with (t' := t') (B1 := B1) (B2 := B2); eauto.
  match goal with
  | [ H : pstep ?a ?b
     |- typing ((th_log,exp_eq (?P (exp_var 0)) (lift ?b 1 0))
                 :: _) _ _ _] =>
    replace (exp_eq (P (exp_var 0)) (lift b 1 0))
       with ({0 ~> lift b 1 0} (exp_eq (P (exp_var 1)) (exp_var 0)))
         by (simpl; auto)
  end.
    eapply ctx_conv_typing_cons with (b1 := lift a 1 0).  eauto.
    eapply typing_weakening_cons_eq.  mycrush. go_go_regularity; eauto.
    simpl; go_go_regularity; eauto.


  match goal with
  | [ H : pstep ?a ?b
     |- typing ((th_log,exp_eq (?P (exp_var 0)) (lift ?b 1 0))
                 :: _) _ _ _] =>
    replace (exp_eq (P (exp_var 0)) (lift b 1 0))
       with ({0 ~> lift b 1 0} (exp_eq (P (exp_var 1)) (exp_var 0)))
         by (simpl; auto)
  end.
    eapply ctx_conv_typing_cons with (b1 := lift a 1 0).  eauto.
    try eapply typing_weakening_cons_eq. mycrush. go_go_regularity; eauto.
    simpl; go_go_regularity; eauto.

  assert (HFoo := typ_roll_inversion_prog _ _ _ Hstep).
  eauto.
Qed.

Theorem preservation : forall G t a b A,
  typing G t a A -> step a b -> typing G t b A.
Proof.
  intros; eauto using pstep_preservation.
Qed.  

Theorem mpreservation : forall G t a b A n,
  typing G t a A -> mstep n a b -> typing G t b A.
Proof.
  induction 2; eauto using preservation.
Qed.
Hint Immediate preservation mpreservation.

