Require Import List.
Require Import Arith.

Require Import step_defs.
Require Import step_db_basics.
Require Import step_reduction_basics.
Require Import step_substitution.
Require Import step_preservation.
Require Import step_interp.
Require Import step_interp_facts.
Require Import step_interp_soundness. 

Theorem equality_soundness : forall a A1 A2,
  typing nil th_log a (exp_eq A1 A2) ->
  exists A,
    mpstep A1 A /\ mpstep A2 A.
Proof.
  intros a A1 A2 H.
  assert (interp (0, exp_eq A1 A2, Comp) th_log nil (close nil a 0))
    as HInterp
    by eauto using interp_soundness.
  rewrite ui_comp in HInterp.
  destruct HInterp as [_ [v [n [_ HInterp]]]].
  rewrite ui_val_eq in HInterp.
  simpl in *.
  destruct HInterp; destruct v; eauto; tauto.
Qed.

(* Because of the implicit box/unbox rules, 
   the canonical forms lemmas don't talk about the top type constructor,
   but rather the first non-@ type constructor. *)
Inductive boxed_head_form : exp -> head -> Prop :=
  | BHStar : boxed_head_form exp_star head_star
  | BHArr : forall A B, boxed_head_form (exp_arr A B) head_arr
  | BHNat : boxed_head_form exp_nat head_nat
  | BHSum : forall A B, boxed_head_form (exp_sum A B) head_sum
  | BHSigma : forall A B, boxed_head_form (exp_sigma A B) head_sigma
  | BHMu : forall A, boxed_head_form (exp_mu A) head_mu
  | BHEq : forall a b, boxed_head_form (exp_eq a b) head_eq
(* Only this line is different from head_form: *)
  | HAt : forall A t h, boxed_head_form A h -> boxed_head_form (exp_at A t) h.
Hint Constructors boxed_head_form.

Lemma pstep_preserves_head : forall A A' h,
  pstep A A' 
  -> head_form A h 
  -> head_form A' h.
Proof.
  intros A A' h STEP HF.
  inversion STEP; inversion HF; (congruence||subst); eauto.

  inversion H2; subst; auto.
Qed.

Lemma mpstep_preserves_head : forall A A' h,
  mpstep A A'
  -> head_form A h
  -> head_form A' h.
Proof.
  induction 1; eauto using pstep_preserves_head.
Qed.

Lemma pstep_preserves_boxed_head : forall A A' h,
  pstep A A' 
  -> boxed_head_form A h 
  -> boxed_head_form A' h.
Proof.
  intros A A' h STEP HF.
  induction STEP; inversion HF; (congruence||subst); eauto.
Qed.

Lemma mpstep_preserves_boxed_head : forall A A' h,
  mpstep A A'
  -> boxed_head_form A h
  -> boxed_head_form A' h.
Proof.
  induction 1; eauto using pstep_preserves_boxed_head.
Qed.

Lemma head_form_functional : forall A h1 h2,
  head_form A h1
  -> head_form A h2 
  -> h1 = h2.
Proof.
  intros A h1 h2 H1 H2.
  inversion H1; inversion H2; congruence.
Qed.

Lemma boxed_head_form_functional : forall A h1 h2,
  boxed_head_form A h1
  -> boxed_head_form A h2 
  -> h1 = h2.
Proof.
  induction A;
  intros h1 h2 H1 H2;
  inversion H1; inversion H2; try congruence; eauto.
Qed.

Lemma head_subst : forall B b1 b2 h1 h2, 
    head_form ({0~>b1} B) h1 
 -> head_form ({0~>b2} B) h2
 -> h1 <> h2
 -> head_form b1 h1 /\ head_form b2 h2.
Proof.
 induction B; intros b1 b2 h1 h2; simpl;
   try clarify_compares; auto;
   inversion 1; inversion 1; congruence.
Qed.

Lemma boxed_head_subst : forall B b1 b2 h1 h2, 
    boxed_head_form ({0~>b1} B) h1 
 -> boxed_head_form ({0~>b2} B) h2
 -> h1 <> h2
 -> boxed_head_form b1 h1 /\ boxed_head_form b2 h2.
Proof.
 induction B; intros b1 b2 h1 h2; simpl;
   try clarify_compares; auto;
   inversion 1; inversion 1; try congruence; eauto.
Qed.

Lemma equal_same_head : forall p B1 B2 h1 h2,
  typing nil th_log p (exp_eq B1 B2) 
  -> head_form B1 h1
  -> head_form B2 h2
  -> h1 = h2.
Proof.
  intros p B1 B2 h1 h2 H H1 H2.
  apply equality_soundness in H.
  destruct H as [B [HB1 HB2]].
  apply mpstep_preserves_head with (A':=B) in H1;
  apply mpstep_preserves_head with (A':=B) in H2;
  eauto using head_form_functional.
Qed.

Lemma equal_same_boxed_head : forall p B1 B2 h1 h2,
  typing nil th_log p (exp_eq B1 B2) 
  -> boxed_head_form B1 h1
  -> boxed_head_form B2 h2
  -> h1 = h2.
Proof.
  intros p B1 B2 h1 h2 H H1 H2.
  apply equality_soundness in H.
  destruct H as [B [HB1 HB2]].
  apply mpstep_preserves_boxed_head with (A':=B) in H1;
  apply mpstep_preserves_boxed_head with (A':=B) in H2;
  eauto using boxed_head_form_functional.
Qed.

(* This theorem except with head_form instead of boxed_head_form is a very
   easy corollary of canonical_forms_star, which we proved earlier.
   Now we prove it again for boxed_head_form. Oh well. *)
Lemma star_defined_boxed_head_form :  forall G t A C,
     typing G t A C -> C = exp_star
  -> (   (exists G1, exists G2, exists t',
              G = (G1 ++ (t',exp_star) :: G2))
      \/ (exists h, boxed_head_form A h)).
Proof.
  induction 1; intros; subst; eauto;
  try solve [match goal with
  | [ H : ?A = exp_star, HT : typing _ _ ?A _ |- _ ] =>
    rewrite H in HT; auto
  | [ H : lift ?A _ _ = exp_star |- _ ] =>
    destruct A; simpl in H; inversion H; eauto 6
  | [ H : _ = exp_star |- _ ] => inversion H
  end].
(* Last case: *)
 destruct IHtyping; eauto; destruct_things; eauto.
Qed.

Lemma typable_defined_boxed_head_form :  forall G t a A,
     typing G t a A 
  -> (   (exists G1, exists G2, exists t',
              G = (G1 ++ (t',exp_star) :: G2))
      \/ (exists h, boxed_head_form A h)).
Proof.
  intros G t a A H.
  destruct (regularity _ _ _ _ H);
   subst; eauto using star_defined_boxed_head_form.
Qed.

Ltac nil_is_empty :=
 try match goal with
    | [ H : nil = ?G1 ++ _ :: _ |- _ ] => destruct G1; simpl in H; inversion H
    | [ H : ?G1 ++ _ :: _ = nil |- _ ] => destruct G1; simpl in H; inversion H
 end.

Ltac rule_out_contra :=
  match goal with
  | [ H : typing nil th_log _ (exp_eq ?B1 ?B2),
      H1 : head_form ?B1 ?h1,
      H2 : head_form ?B2 ?h2 |- _ ] =>
    pose (equal_same_head _ _ _ _ _ H H1 H2); solve [congruence]
  end.

Hint Constructors head_form.

Ltac rule_out_bad_conv :=
  match goal with 
  | [ H1 : boxed_head_form ({0~>?b1} _) ?h1,
      H2 : boxed_head_form ({0~>?b2} _) ?h2 |- _] =>
    assert (h1 <> h2) as NEQ by congruence;
    destruct (boxed_head_subst _ _ _ _ _ H1 H2 NEQ);
    match goal with
    | [ H : typing nil th_log _ (exp_eq ?B1 ?B2),
        H1 : boxed_head_form ?B1 ?h1,
        H2 : boxed_head_form ?B2 ?h2 |- _ ] =>
      pose (equal_same_boxed_head _ _ _ _ _ H H1 H2); solve [congruence]
    end
  end.

Ltac generic_canonical_forms_lemma :=
  induction 1; intros; subst; eauto;
  val_inversion;
  try solve [match goal with
  | [ H : _ = _ |- _ ] => inversion H
  | [ H : boxed_head_form _ _ |- _ ] => inversion H; subst; eauto
  end];
  (   rule_out_contra (* The TContra case *)
   || nil_is_empty   (* The TVar case *)
   ||  match goal with   (* The TConv case *)
       | [ H : typing nil _ _ ({0~>?b1}?A) |- _ ] =>
         destruct (typable_defined_boxed_head_form _ _ _ _ H);
         destruct_things;
              nil_is_empty 
           || match goal with [H : boxed_head_form _ ?x |- _ ] => is_var x; destruct x end;
                solve [rule_out_bad_conv ||  eauto]
       end).

Lemma canonical_forms_nat' : forall G t a A,
     typing G t a A 
  -> G = nil
  -> val a -> boxed_head_form A head_nat
  -> a = exp_zero \/ (exists a', a = exp_succ a'). 
Proof.
  generic_canonical_forms_lemma.
Qed.

Theorem canonical_forms_nat_closed : forall t a,
     typing nil t a exp_nat -> val a
  -> a = exp_zero \/ exists a', a = exp_succ a'.
Proof.
  intros; eauto using canonical_forms_nat'.
Qed.

Theorem canonical_forms_arrow' : forall G t a A,
     typing G t a A 
  -> G = nil
  -> val a -> boxed_head_form A head_arr
  ->    (exists b, a = exp_lam b)
     \/ (exists b, a = exp_fun b)
     \/ (exists b, a = exp_nrec b)
    .
Proof.
  generic_canonical_forms_lemma.
Qed.

Theorem canonical_forms_arrow_closed : forall t a A B,
     typing nil t a (exp_arr A B) -> val a
  ->    (exists b, a = exp_lam b)
     \/ (exists b, a = exp_fun b)
     \/ (exists b, a = exp_nrec b).
Proof.
  eauto using canonical_forms_arrow'.
Qed.

Theorem canonical_forms_mu' : forall G t a A,
     typing G t a A 
  -> G = nil
  -> val a -> boxed_head_form A head_mu
  ->    (exists b, a = exp_roll b).
Proof.
  (* this proof is slightly different from the other ones,
     because I destruct x one level deeper (and also destruct t0).
     The generic_canonical_forms_lemma is not quite generic enough. *)
  induction 1; intros; subst; eauto;
  val_inversion;
  try solve [match goal with
  | [ H : _ = _ |- _ ] => inversion H
  | [ H : boxed_head_form _ _ |- _ ] => inversion H; subst; eauto
  end].

  rule_out_contra.
  nil_is_empty.
 
  destruct (typable_defined_boxed_head_form _ _ _ _ H0);
  destruct_things.
    nil_is_empty.

destruct x as [ | | | [ | ] | | | [ | ] | ];
  solve [rule_out_bad_conv ||  eauto].
Qed.

Theorem canonical_forms_mu_closed : forall t a A,
     typing nil t a (exp_mu A) -> val a
  -> (exists b, a = exp_roll b).
Proof.
  eauto using canonical_forms_mu'.
Qed.

Theorem canonical_forms_sum' : forall G t a A,
     typing G t a A 
     -> G = nil
     -> val a -> boxed_head_form A head_sum
       ->    (exists a', a = exp_inl a')
          \/ (exists a', a = exp_inr a'). 
Proof.
  generic_canonical_forms_lemma.
Qed.

Theorem canonical_forms_sum_closed : forall t a B1 B2,
     typing nil t a (exp_sum B1 B2) -> val a
  -> (exists b, a = exp_inl b) \/ (exists b, a = exp_inr b).
Proof.
  eauto using canonical_forms_sum'.
Qed.

Theorem canonical_forms_sum_closed_at_type : forall t a B1 B2 t0,
     typing nil t a (exp_at (exp_sum B1 B2) t0) -> val a
  -> (exists b, a = exp_inl b) \/ (exists b, a = exp_inr b).
Proof.
  eauto using canonical_forms_sum'.
Qed.

Theorem canonical_forms_sigma' : forall G t a A,
     typing G t a A 
     -> G = nil
     -> val a -> boxed_head_form A head_sigma
     -> (exists b1 b2, a = exp_pair b1 b2).
Proof.
  generic_canonical_forms_lemma.
Qed.

Theorem canonical_forms_sigma_closed : forall t a B1 B2,
     typing nil t a (exp_sigma B1 B2) -> val a
  -> (exists b1 b2, a = exp_pair b1 b2).
Proof.
  eauto using canonical_forms_sigma'.
Qed.

Theorem canonical_forms_sigma_closed_at_type : forall t a B1 B2 t0,
     typing nil t a (exp_at (exp_sigma B1 B2) t0) -> val a
  -> (exists b1 b2, a = exp_pair b1 b2).
Proof.
  eauto using canonical_forms_sigma'.
Qed.


Ltac go_go_canonical_forms :=
  match goal with
  | [ HT : typing nil _ ?a exp_nat, HVal : val ?a |- _ ] =>
    apply canonical_forms_nat_closed in HT; auto;
    destruct HT as [? | [? ?]]; subst
  | [ HT : typing nil _ ?a (exp_arr _ _), HVal : val ?a |- _ ] =>
    apply canonical_forms_arrow_closed in HT; auto;
    destruct_things; subst
  | [ HT : typing nil _ ?a (exp_mu _), HVal : val ?a |- _ ] =>
    apply canonical_forms_mu_closed in HT; auto;
    destruct HT as [? ?]; subst
  | [ HT : typing nil _ ?a (exp_sum _ _), HVal : val ?a |- _ ] =>
    apply canonical_forms_sum_closed in HT; auto;
    destruct HT as [[? ?] | [? ?]]; subst
  | [ HT : typing nil _ ?a (exp_at (exp_sum _ _) _), HVal : val ?a |- _ ] =>
    apply canonical_forms_sum_closed_at_type in HT; auto;
    destruct HT as [[? ?] | [? ?]]; subst
  | [ HT : typing nil _ ?a (exp_sigma _ _), HVal : val ?a |- _ ] =>
    apply canonical_forms_sigma_closed in HT; auto;
    destruct_things; subst
  | [ HT : typing nil _ ?a (exp_at (exp_sigma _ _) _), HVal : val ?a |- _ ] =>
    apply canonical_forms_sigma_closed_at_type in HT; auto;
    destruct_things; subst
  end.

Theorem progress' : forall G t a A,
     typing G t a A 
  -> G = nil -> val a \/ exists a', step a a'.
Proof.
  induction 1; intros;
  try match goal with
  | [ H : ?G1 ++ _ :: _ = nil |- _ ] => destruct G1; simpl in H; inversion H
  end; subst;
  repeat (match goal with
  | [ H : nil = nil -> _ \/ _ |- _ ] => destruct (H eq_refl); clear H
  | [ H : exists _, step ?a _ |- _ ] => 
    destruct H as [? ?]
  end); val_inversion; eauto;
  go_go_canonical_forms; val_inversion; eauto.
Qed.

Theorem progress : forall t a A,
     typing nil t a A 
  -> val a \/ exists a', step a a'.
Proof.
  intros; eauto using progress'.
Qed.
