Require Import Arith.
Require Import Omega.
Require Import Bool.
Require Import List.

Require Import step_defs.
Require Import step_db_basics.
(********************)

(* Some facts about values *)
Theorem values_dont_step : forall a,
  val a -> forall b, step a b -> False.
Proof.
  induction 1; intros ? HStep;
  inversion HStep; subst; eauto.
Qed.

Theorem val_mstep : forall j a b,
  val a -> mstep j a b -> j = 0 /\ a = b.
Proof.
  inversion 2; eauto; subst;
  edestruct values_dont_step; eauto.
Qed.

Ltac step_val_elim :=
  match goal with
  | [ HS : step ?a _ |- _ ] =>
    let HVal := fresh "HVal" in assert (val a) as HVal by eauto;
    destruct (values_dont_step _ HVal _ HS)
  | [ HS : mstep _ ?a _ |- _ ] =>
    let HVal := fresh "HVal" in assert (val a) as HVal by eauto;
    destruct (val_mstep _ _ _ HVal HS); subst
  end.

Theorem step_pstep : forall a b,
  step a b -> pstep a b.
Proof.
  induction 1; eauto.
Qed.
Hint Immediate step_pstep. 

Theorem MPSMu : forall A B,
  mpstep A B -> mpstep (exp_mu A) (exp_mu B).
Proof.
  induction 1; eauto.
Qed.
Hint Resolve MPSMu.

(* Congruence rules for MS *)

Theorem MSTrans : forall j1 j2 a b c,
  mstep j1 a b -> mstep j2 b c -> mstep (j1 + j2) a c.
Proof.
  induction 1; intros; simpl; eauto.
Qed.

Theorem MSApp1 : forall j a a' b,
  mstep j a a' -> mstep j (exp_app a b) (exp_app a' b).
Proof.
  induction 1; intros; simpl; eauto.
Qed.

Theorem MSApp2 : forall j a b b',
  mstep j b b' -> val a -> mstep j (exp_app a b) (exp_app a b').
Proof.
  induction 1; intros; simpl; eauto.
Qed.

Theorem MSSucc : forall j a a',
  mstep j a a' -> mstep j (exp_succ a) (exp_succ a').
Proof. induction 1; intros; simpl; eauto. Qed.

Theorem MSRoll : forall j a a',
  mstep j a a' -> mstep j (exp_roll a) (exp_roll a').
Proof. induction 1; intros; simpl; eauto. Qed.

Theorem MSInl : forall j a b,
  mstep j a b -> mstep j (exp_inl a) (exp_inl b).
Proof. induction 1; intros; eauto. Qed.

Theorem MSInr : forall j a b,
  mstep j a b -> mstep j (exp_inr a) (exp_inr b).
Proof. induction 1; intros; eauto. Qed.

Theorem MSPair2 : forall j a b b',
     mstep j b b' -> val a
  -> mstep j (exp_pair a b) (exp_pair a b').
Proof. induction 1; intros; eauto. Qed.

Theorem MSPair : forall j k a a' b b',
     mstep j a a' -> mstep k b b' -> val a'
  -> mstep (j + k) (exp_pair a b) (exp_pair a' b').
Proof.
  induction 1; intros; [ eauto using MSPair2
                       | eapply MSStep; eauto ].
Qed.

Theorem MSNatCase1 : forall j a a' b1 b2,
  mstep j a a' -> mstep j (exp_ncase a b1 b2) (exp_ncase a' b1 b2).
Proof.
  induction 1; eauto.
Qed.

Theorem MSSumCase1 : forall j a a' b1 b2,
  mstep j a a' -> mstep j (exp_scase a b1 b2) (exp_scase a' b1 b2).
Proof.
  induction 1; eauto.
Qed.

Theorem MSSuccInversion : forall j a b',
  mstep j (exp_succ a) b' -> exists b, b' = exp_succ b /\ mstep j a b.
Proof.
  intros j a b' HStep; remember (exp_succ a) as a';
  revert a Heqa'; induction HStep; eauto.
  match goal with
  | [ HStep : step _ _, H : forall _, _ -> exists _, _ |- _ ] =>
    intros; subst; inversion HStep; subst;
    edestruct IHHStep as [? [? ?]]; eauto
  end.
Qed.

Theorem MSRollInversion : forall j a b',
     mstep j (exp_roll a) b' 
 -> exists b, exp_roll b = b' /\ mstep j a b.
Proof.
  intros j a b' HMS; remember (exp_roll a) as rolla;
  revert a Heqrolla; induction HMS; intros; subst; eauto;
  match goal with
  | [ IH : forall _, _ = _ -> _, Hstep : step (exp_roll _) _ |- _ ] =>
    inversion Hstep; subst; edestruct IHHMS as [? [? ?]]; eauto
  end.
Qed.

Theorem MSInlInversion : forall j a b',
     mstep j (exp_inl a) b' 
  -> exists b, exp_inl b = b' /\ mstep j a b.
Proof.
  intros j a b' HMS; remember (exp_inl a) as inla;
  revert a Heqinla; induction HMS; intros; subst; eauto.
  match goal with
  | [ IH : forall _, _ = _ -> _, Hstep : step (exp_inl _) _ |- _ ] =>
    inversion Hstep; subst; edestruct IH as [? [? ?]]; eauto
  end.
Qed.

Theorem MSInrInversion : forall j a b',
     mstep j (exp_inr a) b' 
  -> exists b, exp_inr b = b' /\ mstep j a b.
Proof.
  intros j a b' HMS; remember (exp_inr a) as inra;
  revert a Heqinra; induction HMS; intros; subst; eauto.
  match goal with
  | [ IH : forall _, _ = _ -> _, Hstep : step (exp_inr _) _ |- _ ] =>
    inversion Hstep; subst; edestruct IH as [? [? ?]]; eauto
  end.
Qed.

Theorem MSPairInversion : forall j a1 a2 b,
     mstep j (exp_pair a1 a2) b
  -> exists j1, exists j2, exists b1, exists b2,
        exp_pair b1 b2 = b
     /\ j = j1 + j2
     /\ mstep j1 a1 b1 /\ mstep j2 a2 b2.
Proof.
  intros j a1 a2 b HMS; remember (exp_pair a1 a2) as paira;
  revert a1 a2 Heqpaira; induction HMS; intros; subst; eauto 8;
  match goal with
  | [ IH : forall _ _, _ = _ -> _, Hstep : step (exp_pair _ _) _ |- _ ] =>
    let j1' := fresh "j1'" in let j2' := fresh "j2'" in
    inversion Hstep; subst; edestruct IH as [j1' [j2' ?]]; eauto;
    clear IH; destruct_things; subst;
    [ exists (S j1'); exists j2' | exists j1'; exists (S j2') ];
    eauto 10
  end.
Qed.


(* reduction is deterministic *)

Theorem step_deterministic : forall a b,
  step a b -> forall b', step a b' -> b = b'.
Proof.
  induction 1; subst; inversion 1; subst; eauto;
  try solve [step_val_elim];
  match goal with
  | [ IHstep : forall _, step _ _ -> _ = _ |- _ ] =>
    erewrite IHstep; eauto
  end.
Qed.

Theorem mstep_val_deterministic : forall b b' a j j',
     mstep j a b 
  -> mstep j' a b'
  -> val b -> val b' -> j = j' /\ b = b'.
Proof.
  intros b b' a j j' HMS1 HMS2; revert b' j' HMS2;
  induction HMS1; subst; inversion 1; subst; intros; eauto;
  try solve [step_val_elim];
  match goal with
  | [ HStep1 : step ?a ?b, HStep2 : step ?a ?b',
      HMS1 : mstep _ ?b ?c, HMS2 : mstep _ ?b' ?c',
      IH : forall _ _, mstep _ _ _ -> _
     |- _ /\ ?c = ?c'] => 
    rewrite <- (step_deterministic _ _ HStep1 _ HStep2) in *;
    destruct (IH  _ _ HMS2); subst; eauto
  end.
Qed.

Theorem mstep_val_deterministic_exp : forall b b' a j j',
     mstep j a b 
  -> mstep j' a b'
  -> val b -> val b' -> b = b'.
Proof.
  intros b b' a j j' HMS1 HMS2 HV1 HV2;
  destruct (mstep_val_deterministic _ _ _ _ _ HMS1 HMS2 HV1 HV2);
  subst; eauto.
Qed.

(* Inversion lemmas for the reduction forms.  We need to be very
   explicit about the number of steps for the main soundness
   theorem. *)
Theorem mstep_unroll_val_inversion : forall a c j,
     mstep j (exp_unroll a) c
  -> val c
  -> exists j1,
         j = S j1
      /\ mstep j1 a (exp_roll c).
Proof.
  intros a c j HMS; remember (exp_unroll a) as eroll;
  revert a Heqeroll; induction HMS; intros; subst;
  val_inversion;
  match goal with
  | [ H  : step (exp_unroll _) _,
      IH : forall _, _ = _ -> val _ -> _
     |- exists _, S ?n = _ /\ _ ] =>
    inversion H; subst;
    match goal with
    | [ HS : step (exp_unroll (exp_roll ?b)) ?b |- _ ] =>
      exists n; eauto using MSRoll
    | [ HS : step (exp_unroll _) (exp_unroll _) |- _ ] =>
      edestruct IH as [? [? ?]]; subst; eauto
    end
  end.
Qed.

Theorem mstep_app_val_inversion : forall a b c j,
     mstep j (exp_app a b) c
  -> val c
  -> exists j1, exists j2, exists j3, exists a', exists vb,
         val vb
      /\ j = 1 + j1 + j2 + j3
      /\ mstep j1 b vb
      /\ (   (   mstep j2 a (exp_lam a') 
              /\ mstep j3 ({0 ~> vb} a') c)
          \/ (   mstep j2 a (exp_fun a') 
              /\ mstep j3 ({0 ~> vb} 
                           ({0 ~> exp_fun (lift a' 1 2)} a')) c)
          \/ (   mstep j2 a (exp_nrec a')
              /\ mstep j3 ({0 ~> vb}
                           ({0 ~> exp_lam (exp_lam (exp_app
                                    (exp_nrec (lift a' 3 2)) (exp_var 1)))} a'))
                          c)).
Proof.
  intros a b c j HMS; remember (exp_app a b) as eapp;
  revert a b Heqeapp; induction HMS; intros; subst;
  val_inversion.
  match goal with
  | [ H : step (exp_app _ _) _,
      IH : forall _ _, _ = _ -> val _ -> _
     |- exists _, exists _, exists _, exists _, exists _,
        val _ /\ S ?n = _ /\ _ ] =>
    inversion H; subst;
    try match goal with
    | [ HMS : mstep ?n ({_ ~> _} _) _,
        HS : step (exp_app (?P ?a) ?b) ({_ ~> _} _) |- _ ] =>
      exists 0; exists 0; exists n; exists a; exists b; eauto 7
    | [ HS : step (exp_app ?a1 ?b) (exp_app ?a2 ?b) |- _ ] =>
      edestruct IH as [j1 [j2 [j3 [abody [vb ?]]]]];
      destruct_things; auto; subst; 
      exists j1; exists (S j2); exists j3; exists abody; exists vb;
      repeat (split; eauto; try omega)
    | [ HS : step (exp_app ?a ?b1) (exp_app ?a ?b2) |- _ ] =>
      edestruct IH as [j1 [j2 [j3 [abody [vb ?]]]]];
      destruct_things; auto; subst; 
      exists (S j1); exists j2; exists j3; exists abody; exists vb;
      repeat (split; eauto; try omega)
    end
  end.
Qed.

Theorem mstep_ncase_val_inversion : forall a b1 b2 c j,
     mstep j (exp_ncase a b1 b2) c
  -> val c
  -> exists j1, exists j2,
         j = 1 + j1 + j2
      /\ (   (   mstep j1 a exp_zero
              /\ mstep j2 ({0 ~> exp_join} b1) c)
          \/ (exists va,
                 val va
              /\ mstep j1 a (exp_succ va)
              /\ mstep j2 ({0 ~> va} ({0 ~> exp_join} b2)) c)).
Proof.
  intros a b1 b2 c j HMS; remember (exp_ncase a b1 b2) as ecase;
  revert a b1 b2 Heqecase; induction HMS; intros; subst;
  val_inversion;
  match goal with
  | [ H : step (exp_ncase _ _ _) _,
      IH : forall _ _ _, _ = _ -> val _ -> _
     |- exists _, exists _, S ?n = _ /\ _ ] =>
    inversion H; subst;
    try match goal with
    | [ HS : step (exp_ncase exp_zero _ _) _ |- _ ] =>
      exists 0; exists n; eauto 6
    | [ HS : step (exp_ncase (exp_succ _) _ _) _ |- _ ] =>
      exists 0; exists n; eauto 6
    | [ HS : step (exp_ncase _ _ _) (exp_ncase _ _ _) |- _ ] =>
      edestruct IH as [j1 [j2 [? [ [? ?] | [? [? [? ?]]] ] ]]]; auto;
      subst; exists (S j1); exists j2;
      (repeat (split; eauto 6; try omega))
    end
  end.  
Qed.

Theorem mstep_scase_val_inversion : forall a b1 b2 c j,
     mstep j (exp_scase a b1 b2) c
  -> val c
  -> exists j1, exists j2,
         j = 1 + j1 + j2
      /\ (   (exists va,
                 val va
              /\ mstep j1 a (exp_inl va)
              /\ mstep j2 ({0 ~> va} ({0 ~> exp_join} b1)) c)
          \/ (exists vb,
                 val vb
              /\ mstep j1 a (exp_inr vb)
              /\ mstep j2 ({0 ~> vb} ({0 ~> exp_join} b2)) c)).
Proof.
  intros a b1 b2 c j HMS; remember (exp_scase a b1 b2) as ecase;
  revert a b1 b2 Heqecase; induction HMS; intros; subst;
  val_inversion;
  match goal with
  | [ H : step (exp_scase _ _ _) _,
      IH : forall _ _ _, _ = _ -> val _ -> _
     |- exists _, exists _, S ?n = _ /\ _ ] =>
    inversion H; subst;
    try match goal with
    | [ HS : step (exp_scase (exp_inl _) _ _) _ |- _ ] =>
      exists 0; exists n; eauto 6
    | [ HS : step (exp_scase (exp_inr _) _ _) _ |- _ ] =>
      exists 0; exists n; eauto 6
    | [ HS : step (exp_scase _ _ _) (exp_scase _ _ _) |- _ ] =>
      edestruct IH as [j1 [j2 ?]]; destruct_things; auto;
      subst; exists (S j1); exists j2;
      split; try omega; [left | right]; eauto 6
    end
  end.
Qed.

Theorem mstep_pcase_val_inversion : forall a b c j,
     mstep j (exp_pcase a b) c
  -> val c
  -> exists j1, exists j2, exists va1, exists va2,
         j = 1 + j1 + j2
      /\ val va1 /\ val va2
      /\ mstep j1 a (exp_pair va1 va2)
      /\ mstep j2 ({0 ~> va1} ({0 ~> lift va2 1 0} ({0~> exp_join}b))) c.
Proof.
  intros a b c j HMS; remember (exp_pcase a b) as ecase;
  revert a b Heqecase; induction HMS; intros; subst;
  val_inversion.
  match goal with
  | [ H : step (exp_pcase _ _) _,
      IH : forall _ _, _ = _ -> val _ -> _
     |- exists _, exists _, exists _, exists _, S ?n = _ /\ _ ] =>
    inversion H; subst;
    match goal with
    | [ HMS : mstep n ({0 ~> ?va1} ({0 ~> lift ?va2 _ _} _)) _ |- _ ] =>
      exists 0; exists n; exists va1; exists va2; eauto
    | [ HS : step (exp_pcase _ _) (exp_pcase _ _) |- _ ] =>
      let j1 := fresh "j1" in let j2 := fresh "j2" in
      let va1 := fresh "va1" in let va2 := fresh "va2" in
      edestruct IH as [j1 [j2 [va1 [va2 ?]]]]; destruct_things; auto;
      subst; exists (S j1); exists j2; exists va1; exists va2; eauto 6
    end
  end.
Qed.



(******************)
(***** mpstep *****)
Theorem mpstep_zero_inversion' : forall A B,
  mpstep A B -> A = exp_zero -> B = exp_zero.
Proof.
  induction 1; intros; subst; auto;
  match goal with
  | [ H : pstep exp_zero _ |- _ ] => inversion H; subst; eauto
  end.
Qed.

Theorem mpstep_zero_inversion : forall A,
  mpstep exp_zero A -> A = exp_zero.
Proof.  intros; eapply mpstep_zero_inversion'; eauto.  Qed.

Theorem mpstep_succ_inversion' : forall A B,
     mpstep A B -> forall A1, A = exp_succ A1
  -> exists B1,
       B = exp_succ B1 /\ mpstep A1 B1.
Proof.
  induction 1; intros; subst; eauto;
  match goal with
  | [ H : pstep (exp_succ _) _ |- _ ] =>
    inversion H; subst; destruct_ih; subst; eauto 8
  end.
Qed.

Theorem mpstep_succ_inversion : forall A B,
     mpstep (exp_succ A) B
  -> exists B1,
       B = exp_succ B1 /\ mpstep A B1.
Proof. eauto using mpstep_succ_inversion'. Qed.

Theorem mpstep_nat_inversion' : forall A B,
  mpstep A B -> A = exp_nat -> B = exp_nat.
Proof.
  induction 1; intros; subst; auto;
  match goal with
  | [ H : pstep exp_nat _ |- _ ] => inversion H; subst; eauto
  end.
Qed.

Theorem mpstep_nat_inversion : forall A,
  mpstep exp_nat A -> A = exp_nat.
Proof.  intros; eapply mpstep_nat_inversion'; eauto.  Qed.


Theorem mpstep_arr_inversion' : forall A B,
     mpstep A B -> forall A1 A2, A = exp_arr A1 A2
  -> exists B1, exists B2,
       B = exp_arr B1 B2 /\ mpstep A1 B1 /\ mpstep A2 B2.
Proof.
  induction 1; intros; subst; eauto;
  match goal with
  | [ H : pstep (exp_arr _ _) _ |- _ ] =>
    inversion H; subst; destruct_ih; subst; eauto 8
  end.
Qed.

Theorem mpstep_arr_inversion : forall A1 A2 B,
     mpstep (exp_arr A1 A2) B
  -> exists B1, exists B2,
       B = exp_arr B1 B2 /\ mpstep A1 B1 /\ mpstep A2 B2.
Proof. eauto using mpstep_arr_inversion'. Qed.

Theorem mpstep_sigma_inversion' : forall A B,
     mpstep A B -> forall A1 A2, A = exp_sigma A1 A2
  -> exists B1, exists B2,
       B = exp_sigma B1 B2 /\ mpstep A1 B1 /\ mpstep A2 B2.
Proof.
  induction 1; intros; subst; eauto;
  match goal with
  | [ H : pstep (exp_sigma _ _) _ |- _ ] =>
    inversion H; subst; destruct_ih; subst; eauto 8
  end.
Qed.

Theorem mpstep_sigma_inversion : forall A1 A2 B,
     mpstep (exp_sigma A1 A2) B
  -> exists B1, exists B2,
       B = exp_sigma B1 B2 /\ mpstep A1 B1 /\ mpstep A2 B2.
Proof. eauto using mpstep_sigma_inversion'. Qed.

Theorem mpstep_sum_inversion' : forall A B,
    mpstep A B -> forall A1 A2, A = exp_sum A1 A2
 -> exists B1, exists B2,
      B = exp_sum B1 B2 /\ mpstep A1 B1 /\ mpstep A2 B2.
Proof.
  induction 1; intros; subst; eauto;
  match goal with
  | [ H : pstep (exp_sum _ _) _ |- _ ] =>
    inversion H; subst; destruct_ih; subst; eauto 8
  end.
Qed.

Theorem mpstep_sum_inversion : forall A1 A2 B,
     mpstep (exp_sum A1 A2) B
  -> exists B1, exists B2,
       B = exp_sum B1 B2 /\ mpstep A1 B1 /\ mpstep A2 B2.
Proof. eauto using mpstep_sum_inversion'. Qed.


Theorem mpstep_mu_inversion' : forall A B,
     mpstep A B -> forall A1, A = exp_mu A1
  -> exists B1, B = exp_mu B1 /\ mpstep A1 B1.
Proof.
  induction 1; intros; subst; eauto;
  match goal with
  | [ H : pstep (exp_mu _) _ |- _ ] =>
    inversion H; subst; destruct_ih; subst; eauto 8
  end.
Qed.

Theorem mpstep_mu_inversion : forall A B,
     mpstep (exp_mu A) B
  -> exists B1, B = exp_mu B1 /\ mpstep A B1.
Proof.  eauto using mpstep_mu_inversion'.  Qed.


Theorem mpstep_eq_inversion' : forall A B,
     mpstep A B -> forall a1 a2, A = exp_eq a1 a2
  -> exists b1, exists b2,
       B = exp_eq b1 b2 /\ mpstep a1 b1 /\ mpstep a2 b2.
Proof.
  induction 1; intros; subst; eauto;
  match goal with
  | [ H : pstep (exp_eq _ _) _ |- _ ] =>
    inversion H; subst; destruct_ih; subst; eauto 8
  end.
Qed.

Theorem mpstep_eq_inversion : forall a1 a2 B,
     mpstep (exp_eq a1 a2) B
  -> exists b1, exists b2,
       B = exp_eq b1 b2 /\ mpstep a1 b1 /\ mpstep a2 b2.
Proof.  eauto using mpstep_eq_inversion'.  Qed.


Theorem mpstep_at_inversion' : forall A B,
     mpstep A B -> forall A1 t, A = exp_at A1 t
  -> exists B1, B = exp_at B1 t /\ mpstep A1 B1.
Proof.
  induction 1; intros; subst; eauto;
  match goal with
  | [ H : pstep (exp_at _ _) _ |- _ ] =>
    inversion H; subst; destruct_ih; eauto 8
  end.
Qed.

Theorem mpstep_at_inversion : forall A t B,
     mpstep (exp_at A t) B
  -> exists B1, B = exp_at B1 t /\ mpstep A B1.
Proof.  eauto using mpstep_at_inversion'.  Qed.


Theorem mpstep_star_inversion' : forall A B,
     mpstep A B -> A = exp_star -> B = exp_star.
Proof.
  induction 1; intros; subst; eauto;
  match goal with
  | [ H : pstep exp_star _ |- _ ] =>
    inversion H; subst; destruct_ih; eauto 8
  end.
Qed.

Theorem mpstep_star_inversion : forall B,
     mpstep exp_star B -> B = exp_star.
Proof.  intros; eauto using mpstep_star_inversion'.  Qed.



Ltac mpstep_inversion :=
  repeat (match goal with
  | [ H : mpstep exp_nat ?B |- _ ] => let H' := fresh H in
    match B with
    | exp_nat => fail 1
    | _ => assert (H' := mpstep_nat_inversion _ H); subst; try solve [inversion H']
    end
  | [ H : mpstep exp_zero ?B |- _ ] => let H' := fresh H in
    match B with
    | exp_zero => fail 1
    | _ => assert (H' := mpstep_zero_inversion _ H); subst; try solve [inversion H']
    end
  | [ H : mpstep (exp_succ ?A') ?B |- _ ] => let H' := fresh H in
    match B with 
    | exp_succ ?B' => 
      match goal with
      | [ H : mpstep A' B' |- _ ] => fail 2
      | _ => idtac
      end
    | _ =>
      destruct (mpstep_succ_inversion _ _ H) as [? [HEq ?]]; subst; 
      try solve [inversion HEq]
    end
  | [ H : mpstep (exp_arr ?A1 _) ?B |- _ ] =>
    match B with | exp_arr ?B1 _ =>
      match goal with 
        | [ H : mpstep A1 B1 |- _ ] => fail 2
        | _ => idtac 
      end
      | _ => idtac
    end;
    destruct (mpstep_arr_inversion _ _ _ H) as [? [? [HEq [? ?]]]]; subst;
    try solve [inversion HEq];
    match B with
    | exp_arr _ _ => inversion HEq; subst; clear HEq
    | _ => idtac
    end
  | [ H : mpstep (exp_sigma ?A1 _) ?B |- _ ] =>
    match B with | exp_sigma ?B1 _ =>
      match goal with 
        | [ H : mpstep A1 B1 |- _ ] => fail 2
        | _ => idtac 
      end
      | _ => idtac
    end;
    destruct (mpstep_sigma_inversion _ _ _ H) as [? [? [HEq [? ?]]]]; subst;
    try solve [inversion HEq];
    match B with
    | exp_sigma _ _ => inversion HEq; subst; clear HEq
    | _ => idtac
    end
  | [ H : mpstep (exp_sum _ _) ?B |- _ ] =>
    let HEq := fresh "HEq" in
    destruct (mpstep_sum_inversion _ _ _ H) as [? [? [HEq [? ?]]]]; subst; clear H;
    try solve [inversion HEq];
    match B with
    | exp_sum _ _ => inversion HEq; subst; clear HEq
    | _ => idtac
    end
  | [ H : mpstep exp_star ?B |- _ ] =>
    match B with
    | exp_star => fail 1
    | _ =>
      let HEq := fresh "HEq" in assert (HEq := mpstep_star_inversion _ H);
      try solve [inversion HEq]; subst
    end
  | [ H : mpstep (exp_eq _ _) ?B |- _ ] =>
    let HEq := fresh "HEq" in
    destruct (mpstep_eq_inversion _ _ _ H) as [? [? [HEq [? ?]]]]; subst; clear H;
    try solve [inversion HEq];
    match B with
    | exp_eq _ _ => inversion HEq; subst; clear HEq
    | _ => idtac
    end
  | [ H : mpstep (exp_at _ _) ?B |- _ ] =>
    let HEq := fresh "HEq" in
    destruct (mpstep_at_inversion _ _ _ H) as [? [HEq ?]]; subst; clear H;
    try solve [inversion HEq];
    match B with
    | exp_at _ _ => inversion HEq; subst; clear HEq
    | _ => idtac
    end
  | [ H : mpstep (exp_mu _) ?B |- _ ] =>
    let HEq := fresh "HEq" in
    destruct (mpstep_mu_inversion _ _ H) as [? [HEq ?]]; subst; clear H;
    try solve [inversion HEq];
    match B with
    | exp_mu _ => inversion HEq; subst; clear HEq
    | _ => idtac
    end
  end).
