Require Import List.
Require Import Arith.
Require Import step_defs.
Require Program.Equality.  (* This adds nice hints, apparently *)
Require Omega.

(* This file contains basic facts about the substitution and lift
   operations on terms.  It's all De Bruijn index wrangling *)

(*************************************)
(******** Tactics and Hints **********)
Hint Extern 2 =>
  match goal with
  | [ |- _ <= _ ] => omega
  | [ |- _ < _ ] => omega
  end.

(* These next three seek out uses of nat_compare and break them
   down into several cases. *)
Ltac compares_cleanup :=
    match goal with
    | [ H : nat_compare _ _ = Eq |- _ ] => apply nat_compare_eq in H
    | [ H : nat_compare _ _ = Lt |- _ ] => apply nat_compare_lt in H
    | [ H : nat_compare _ _ = Gt |- _ ] => apply nat_compare_gt in H
    end.
  
Ltac destruct_compares n m :=
  remember (nat_compare n m);
  match goal with | [ H : ?comp = nat_compare n m |- _ ] =>
    apply eq_sym in H;
    destruct comp
  end;
  compares_cleanup.
  
Ltac clarify_compares :=
  match goal with
  | [ |- context[match nat_compare ?n ?m with
                 | Lt => _ | Eq => _ | Gt => _ 
                 end]
    ] => destruct_compares n m
  | [ H : ?P (match nat_compare ?n ?m with
              | Lt => _ | Eq => _ | Gt => _ 
              end) _ |- _
    ] => destruct_compares n m
  | [ H : ?P _ (match nat_compare ?n ?m with
                | Lt => _ | Eq => _ | Gt => _ 
                end) |- _
    ] => destruct_compares n m
  end.

(* Most of the proofs in this file involve equalities.  The proofs go like this: 
   
   - In the var case, break down all the comparisons, reduce, observe
     the two sides are equal
   - In the inductive cases, rewrite by the IHs and declare victory.

   The next two pieces of ltac just do these two cases.
 *)
Ltac db_var_case := 
  repeat clarify_compares; simpl in *; intuition;
  repeat clarify_compares; simpl in *; autorewrite with steps in *; intuition.

Ltac rewrite_ih :=
  repeat (match goal with
  | [ H : _ = _ |- _ ] => rewrite H by eauto; clear H
  | [ H : forall _, _ = _ |- _ ] => rewrite H by eauto; clear H
  | [ H : forall _ _, _ = _ |- _ ] => rewrite H by eauto; clear H
  | [ H : forall _ _ _, _ = _ |- _ ] => rewrite H by eauto; clear H
  | [ H : forall _ _ _ _, _ = _ |- _ ] => rewrite H by eauto; clear H
  | [ H : forall _ _ _ _ _, _ = _ |- _ ] => rewrite H by eauto; clear H
  | [ H : forall _ _ _ _ _ _, _ = _ |- _ ] => rewrite H by eauto; clear H
  | [ H : forall _ _ _ _ _ _ _, _ = _ |- _ ] => rewrite H by eauto; clear H
  | [ H : forall _ _ _ _ _ _ _ _, _ = _ |- _ ] => rewrite H by eauto; clear H
  end).


(*************************************)
(****** Now, onto the theorems *******)
Theorem lift_lift_swap_gen : forall b m1 m2 n1 n2,
    m2 <= m1
  ->   lift_e (lift_e b n1 m1) n2 m2
     = lift_e (lift_e b n2 m2) n1 (m1 + n2).
Proof.
  induction b; intros; simpl in *; auto;
  rewrite_ih; eauto; try db_var_case.
Qed.

Theorem lift_lift_swap_0 : forall b m1 n1 n2,
       lift_e (lift_e b n1 m1) n2 0
     = lift_e (lift_e b n2 0) n1 (n2 + m1).
Proof.
  intros; subst; rewrite lift_lift_swap_gen; eauto with arith.
Qed.
Hint Rewrite <- lift_lift_swap_0 : steps.
  
Hint Extern 4 =>
  match goal with
  | [ |- context[lift_e (lift_e _ _ ?m1) ?n2 0] ] =>
    rewrite lift_lift_swap_0
  | [ H : context[lift_e (lift_e _ _ ?m1) ?n2 0] |- _ ] =>
    rewrite lift_lift_swap_0 in H
  end.

Theorem lift_lift_collapse : forall b n1 n2 m1 m2,
     m1 <= m2
  -> m2 <= m1 + n1
  ->   lift_e (lift_e b n1 m1) n2 m2
     = lift_e b (n1 + n2) m1.
Proof.
  induction b; intros; simpl in *; auto;
  rewrite_ih; auto; try db_var_case.
Qed.
Hint Rewrite lift_lift_collapse using auto : steps.

(* Special case of the last theorem that comes up *)
Theorem lift_lift_collapse_1 : forall b n1,
       lift_e (lift_e b n1 0) 1 0
     = lift_e b (S n1) 0.
Proof.
  intros; autorewrite with steps; rewrite plus_comm; simpl; auto.
Qed.
Hint Rewrite lift_lift_collapse_1 : steps.

Theorem lift_open_commute : forall B b i n m,
     i <= m
  ->   lift_e ({i ~> b}B) n m
     = {i ~> lift_e b n m}(lift_e B n (S m)).
Proof.
  induction B; intros; simpl in *;
  rewrite_ih; auto; try db_var_case.
Qed.

Theorem open_lift_commute : forall b a n1 n2 i m,
     m <= n2
  -> m <= i 
  -> {n1 + i ~> (lift_e a (n1 + n2) 0)} (lift_e b n1 m) 
   = lift_e ({i ~> lift_e a n2 0} b) n1 m.
Proof.
  induction b; intros; simpl in *; eauto;
  autorewrite with steps;
  repeat (match goal with
  | [ |- context [ S (_ + _) ] ] =>
    repeat (rewrite plus_n_Sm)
  | [ |- context [ (_ + _ + _) ] ] =>
    repeat (rewrite <- plus_assoc)
  end); rewrite_ih; auto; try db_var_case.
Qed.

Hint Extern 4 =>
  match goal with
  | [ |- context [lift_e ({_ ~> _}_) _ _] ] =>
    rewrite lift_open_commute by auto
  end.

Theorem open_under_lift : forall b a i n m,
     i <= n + m
  -> m <= i
  -> {i ~> a} (lift_e b (S n) m) = lift_e b n m.
Proof.
  induction b; intros; simpl in *; eauto; subst;
  rewrite_ih; auto; try db_var_case.
Qed.
Hint Rewrite open_under_lift using auto : steps.

Theorem lift_0_is_id : forall (b : exp) (m : nat),
  lift_e b 0 m = b.
Proof.
  induction b; intros; simpl in *; eauto;
  rewrite_ih; auto; try db_var_case.
Qed.
Hint Rewrite lift_0_is_id : steps.

Theorem open_open_swap : forall b n1 n2 a1 a2,
    n2 <= n1
  ->   {n1 ~> a1}({n2 ~> a2}b)
     = {n2 ~> ({n1 ~> a1}a2)}({S n1 ~> lift_e a1 1 n2}b).
Proof.
  induction b; intros; simpl in *; eauto;
  rewrite_ih; auto;
  (* a little bit of fanciness so that we can see open_lift_commute applies *)
  try (match goal with
  | [ |- context[lift_e ({?i ~> ?a}?b) ?n 0] ] =>
    replace (lift_e ({i ~> a}b) n 0) with (lift_e ({i ~> lift_e a 0 0}b) n 0)
      by (autorewrite with steps; eauto);
    rewrite <- open_lift_commute; eauto
  end);
  try db_var_case.
Qed.

(* The de bruijn operations meet reduction *)
Theorem lift_preserves_val : forall v,
  val v -> forall n m, val (lift_e v n m).
Proof.
  induction 1; intros; simpl in *; eauto.
Qed.
Hint Resolve lift_preserves_val.

Theorem open_preserves_val : forall v,
  val v -> forall v' n, val v' -> val ({n ~> v'} v).
Proof.
  induction 1; intros; simpl in *; eauto;
  clarify_compares; simpl; eauto.
Qed.
Hint Resolve open_preserves_val.

Ltac extract_lift_fun :=
  match goal with
  | [ |- context[ exp_fun (lift_e ?a 1 2) ] ] =>
    replace (exp_fun (lift_e a 1 2)) with (lift_e (exp_fun a) 1 0)
         by (simpl; auto)
  end.

Theorem lift_preserves_step : forall A B,
     step A B
  -> forall n m, 
       step (lift_e A n m) (lift_e B n m).
Proof.
  induction 1; intros; simpl in *; eauto.
  match goal with
  | [ |- step _ (lift_e ({_ ~> _} ({0 ~> exp_fun (lift_e ?a 1 2)} ?a)) _ ?m) ] 
    => repeat (rewrite lift_open_commute; auto);
       extract_lift_fun;
       replace (S m) with (1 + m) by (simpl; eauto);
       rewrite <- lift_lift_swap_0; simpl; eauto
  end.
Qed.
Hint Resolve lift_preserves_step.  

Hint Rewrite <- plus_n_O : steps.
Hint Rewrite app_length : steps.

(* Opening closed terms does nothing.  Here we use the typing relation
   instead of defining something else to count free variables, just because
   I'm lazy. *)
Theorem open_closed_id_gen : forall G t a A, 
  typing G t a A -> forall n b, length G <= n -> {n ~> b}a = a.
Proof.
  induction 1; intros; subst; simpl in *; eauto;
  rewrite_ih; auto;
  (* Deal with the var case *)
  match goal with
  | [ |- context[ exp_var _ ] ] =>
    db_var_case; simpl in *;
    match goal with
    | [ H0 : length _ + S _ <= _ |- _ ] => 
      apply le_not_gt in H0; destruct H0; auto
    end
  end.
Qed.

Theorem open_closed_id : forall t a A, 
  typing nil t a A -> forall n b, {n ~> b}a = a.
Proof.
  intros; erewrite open_closed_id_gen; eauto; simpl; auto.
Qed.
Hint Extern 3 =>
  match goal with
  | [ H : typing nil _ ?a _ |- context [{?n ~> ?b}?a] ] =>
    erewrite (open_closed_id _ _ _ H n b) by assumption
  end.


(* Same thing with lifting closed terms *)
Theorem lift_closed_id_gen : forall G t a A, 
     typing G t a A 
  -> forall n m, length G <= m -> lift_e a n m = a.
Proof.
  induction 1; intros; subst; simpl in *; eauto;
  rewrite_ih; auto;
  (* Deal with the var case *)
  match goal with
  | [ |- context[ exp_var _ ] ] =>
    db_var_case; simpl in *;
    match goal with
    | [ H0 : length _ + S _ <= _ |- _ ] => 
      apply le_not_gt in H0; destruct H0; auto
    end
  end.
Qed.

Theorem lift_closed_id : forall t a A, 
  typing nil t a A -> forall n m, lift_e a n m = a.
Proof.
  intros; erewrite lift_closed_id_gen; eauto; simpl; auto.
Qed.
Hint Extern 3 =>
  match goal with
  | [ H : typing nil _ ?a _ |- context [lift_e ?a ?n ?m] ] =>
    erewrite (lift_closed_id _ _ _ H n m) by assumption
  end.

