Require Import Arith.
Require Import Bool.
Require Import List.

(* This file was generated with help from OTT, but heavily modified.
   OTT does not currently support de Bruijn indicies. *)

Inductive th : Set := 
 | th_log : th
 | th_prog : th.

Inductive exp : Set := 
 | exp_var    (n:nat)
 | exp_zero   
 | exp_succ   (a:exp)
 | exp_fun    (a:exp)
 | exp_lam    (a:exp)
 | exp_app    (a:exp) (b:exp)
 | exp_natcase (a:exp) (b1:exp) (b2:exp)
 | exp_sumcase (a:exp) (b1:exp) (b2:exp)
 | exp_roll   (a:exp)
 | exp_unroll (a:exp)
 | exp_inl    (a:exp)
 | exp_inr    (b:exp).

Inductive typ : Set := 
 | typ_nat
 | typ_arr (A1:typ) (A2:typ)
 | typ_at  (A:typ)  (t:th)
 | typ_sum (A B : typ)
 | typ_var (n:nat)
 | typ_mu  (A:typ).

Definition ctx : Set := list (th * typ).

Inductive val : exp -> Prop :=
  | val_var  : forall n, val (exp_var n)
  | val_zero : val (exp_zero)
  | val_succ : forall v, val v -> val (exp_succ v)
  | val_fun  : forall a, val (exp_fun a)
  | val_lam  : forall a, val (exp_lam a)
  | val_inl  : forall a, val a -> val (exp_inl a)
  | val_inr  : forall b, val b -> val (exp_inr b)
  | val_roll : forall v, val v -> val (exp_roll v).

(** definitions *)
(* lift_e e n m  -  adds n to every var which is at least as big as m.

   so if     G1 ++ G3 |- e : A
   and       length G2 = n      length G1 = m
   then 
        G1 ++ G2 ++ G3 |- lift_e e n m : A
 *)
Fixpoint lift_e (e : exp) (n m : nat) : exp :=
  match e with
  | exp_var n'       => exp_var (match nat_compare n' m with
                                 | Lt => n'
                                 | _  => (n' + n)
                                 end)
  | exp_zero         => exp_zero
  | exp_succ a       => exp_succ (lift_e a n m)
  | exp_fun  a       => exp_fun  (lift_e a n (S (S m)))
  | exp_lam  a       => exp_lam  (lift_e a n (S m))
  | exp_app  a b     => exp_app  (lift_e a n m) (lift_e b n m)
  | exp_natcase a b1 b2 => exp_natcase (lift_e a n m) (lift_e b1 n m)
                                 (lift_e b2 n (S m))
  | exp_sumcase a b1 b2 => exp_sumcase (lift_e a n m) (lift_e b1 n (S m))
                                                      (lift_e b2 n (S m))
  | exp_inl a        => exp_inl (lift_e a n m)
  | exp_inr b        => exp_inr (lift_e b n m)
  | exp_roll   a     => exp_roll (lift_e a n m)
  | exp_unroll a     => exp_unroll (lift_e a n m)
  end.

Fixpoint lift_t (T : typ) (n m : nat) : typ :=
  match T with
  | typ_nat       => typ_nat
  | typ_arr A1 A2 => typ_arr (lift_t A1 n m) (lift_t A2 n m)
  | typ_at A t    => typ_at (lift_t A n m) t
  | typ_sum A B   => typ_sum (lift_t A n m) (lift_t B n m)
  | typ_var n'    => typ_var (match nat_compare n' m with
                              | Lt => n'
                              | _  => (n' + n)
                              end)
  | typ_mu A      => typ_mu (lift_t A n (S m))
  end.

Reserved Notation "{ n ~> a }" (at level 60, no associativity).
Fixpoint open_e (n : nat) (u : exp) (e : exp) {struct e} : exp :=
  match e with
  | exp_var  n'   => 
    match nat_compare n' n with
      | Lt => exp_var n'
      | Eq => u
      | Gt => exp_var (pred n')
    end
  | exp_zero      => exp_zero
  | exp_succ a    => exp_succ ({n ~> u} a)
  | exp_fun  a    => exp_fun ({S (S n) ~> lift_e u 2 0} a)
  | exp_lam  a    => exp_lam ({S n ~> lift_e u 1 0} a)
  | exp_app  a b  => exp_app ({n ~> u}a) ({n ~> u}b)
  | exp_natcase a b1 b2 => exp_natcase ({n ~> u} a) ({n ~> u}b1) 
                                 ({S n ~> lift_e u 1 0} b2)
  | exp_sumcase a b1 b2 => exp_sumcase ({n ~> u} a) 
                                  ({S n ~> lift_e u 1 0} b1) 
                                  ({S n ~> lift_e u 1 0} b2)
  | exp_inl a     => exp_inl ({n ~> u} a)
  | exp_inr b     => exp_inr ({n ~> u} b)
  | exp_roll  a   => exp_roll ({n ~> u} a)
  | exp_unroll a  => exp_unroll ({n ~> u} a)
  end
  where "{ n ~> a }" := (open_e n a).

Reserved Notation "{ n ~~> A }" (at level 60, no associativity).
Fixpoint open_t (n : nat) (U : typ) (T : typ) {struct T} : typ :=
  match T with
  | typ_nat       => typ_nat
  | typ_arr A B   => typ_arr ({n ~~> U} A) ({n ~~> U} B)
  | typ_at A t    => typ_at ({n ~~> U} A) t
  | typ_sum A B   => typ_sum ({n ~~> U} A) ({n ~~> U} B)
  | typ_var n'    =>
    match nat_compare n' n with
      | Lt => typ_var n'
      | Eq => U
      | Gt => typ_var (pred n')
    end
  | typ_mu A      => typ_mu ({S n ~~> lift_t U 1 0} A)
  end
  where "{ n ~~> A }" := (open_t n A).


(* defns JStepCBV *)
Inductive step : exp -> exp -> Prop :=    (* defn step *)
 | SCaseZ : forall (a1 a2:exp),
     step (exp_natcase exp_zero a1 a2) a1
 | SCaseS : forall (v a1 a2:exp),
     val v ->
     step (exp_natcase (exp_succ v) a1 a2)  ({0 ~> v} a2) 
 | SCaseInl : forall (v a1 a2:exp),
     val v ->
     step (exp_sumcase (exp_inl v) a1 a2) ({0 ~> v} a1)
 | SCaseInr : forall (v a1 a2:exp),
     val v ->
     step (exp_sumcase (exp_inr v) a1 a2)  ({0 ~> v} a2) 
 | SBetaFun : forall (a v:exp),
     val v ->
     step (exp_app (exp_fun a) v)
          ({0 ~> v} ({0 ~> exp_fun (lift_e a 1 2)} a))
 | SBetaLam : forall (a v:exp),
     val v ->
     step (exp_app (exp_lam a) v) ({0 ~> v} a) 
 | SUnroll : forall (v:exp),
     val v ->
     step (exp_unroll (exp_roll v)) v
 | SSucc1 : forall a a',
     step a a' ->
     step (exp_succ a) (exp_succ a')
 | SApp1 : forall (a a' b : exp),
     step a a' -> 
     step (exp_app a b) (exp_app a' b)
 | SApp2 : forall (v b b' : exp),
     val v -> 
     step b b' -> 
     step (exp_app v b) (exp_app v b')
 | SInl1 : forall a a',
     step a a' ->
     step (exp_inl a) (exp_inl a')
 | SInr1 : forall a a',
     step a a' ->
     step (exp_inr a) (exp_inr a')
 | SNatCase1 : forall (a a' b1 b2 : exp),
     step a a' -> 
     step (exp_natcase a b1 b2) (exp_natcase a' b1 b2)
 | SSumCase1 : forall (a a' b1 b2 : exp),
     step a a' -> 
     step (exp_sumcase a b1 b2) (exp_sumcase a' b1 b2)
 | SRoll1 : forall a a',
     step a a' ->
     step (exp_roll a) (exp_roll a')
 | SUnroll1 : forall a a',
     step a a' ->
     step (exp_unroll a) (exp_unroll a').


(** definitions *)
(* defns JMultiStep *)
Inductive mstep : nat -> exp -> exp -> Prop :=    (* defn mstep *)
 | MSRefl : forall (a:exp),
     mstep 0 a a
 | MSStep : forall (n:nat) (a:exp) (b' b:exp),
     step a b -> 
     mstep n b b' -> 
     mstep (S n) a b'.

(* defns JArbStep *)
Inductive astep : exp -> exp -> Prop :=    (* defn astep *)
 | ASAny : forall n (a b:exp),
     mstep n a b ->
     astep a b.

(* defns Jetyp *)
Inductive FO : typ -> Prop :=    (* defn fo *)
 | FONat : 
     FO typ_nat
 | FOAt : forall (A:typ) (t:th),
     FO (typ_at A t)
 | FOSum : forall (A B:typ),
     FO A -> FO B ->
     FO (typ_sum A B).

Inductive typing : ctx -> th -> exp -> typ -> Prop :=    (* defn typing *)
 | TVar : forall (G1 G3:ctx) (t:th) (n:nat) (A:typ),
     n = length G1 ->
     typing (G1 ++ (t,A) :: G3) t (exp_var n) A
 | TApp : forall (G:ctx) (t:th) (a b:exp) (A B:typ),
     typing G t a (typ_arr A B) ->
     typing G t b A ->
     typing G t (exp_app a b) B
 | TLam : forall (G:ctx) (b:exp) (A B:typ),
     typing ((th_log,A) :: G) th_log b B ->
     FO A ->
     typing G th_log (exp_lam b)  (typ_arr A B)
 | TRec : forall (G:ctx) (a:exp) (A B:typ),
     typing ((th_prog, typ_arr A B) :: (th_prog, A) :: G) th_prog a B ->
     FO A ->
     typing G th_prog (exp_fun a) (typ_arr A B)
 | TSub : forall (G:ctx) (a:exp) (A:typ),
     typing G th_log a A ->
     typing G th_prog a A
 | TFOVal : forall (G:ctx) (v:exp) (A:typ),
     val v ->
     FO A ->
     typing G th_prog v A ->
     typing G th_log v A
 | TZero : forall (G:ctx) (t:th),
     typing G t exp_zero typ_nat
 | TSuc : forall (G:ctx) (t:th) (b:exp),
     typing G t b typ_nat ->
     typing G t (exp_succ b) typ_nat
 | TNatCase : forall (G:ctx) (t:th) (a b1 b2:exp) (A:typ),
     typing G t a typ_nat ->
     typing G t b1 A ->
     typing ((t, typ_nat) :: G) t b2 A ->
     typing G t (exp_natcase a b1 b2) A
 | TInl : forall (G:ctx) (t:th) (a:exp) (A B:typ),
     typing G t a A ->
     typing G t (exp_inl a) (typ_sum A B)
 | TInr : forall (G:ctx) (t:th) (b:exp) (A B:typ),
     typing G t b B ->
     typing G t (exp_inr b) (typ_sum A B)
 | TSumCase : forall (G:ctx) (t t':th) (a b1 b2:exp) (A B C:typ),
     typing G t a (typ_at (typ_sum A B) t') ->
     typing ((t',A) :: G) t b1 C ->
     typing ((t',B) :: G) t b2 C ->
     typing G t (exp_sumcase a b1 b2) C
 | TBoxL : forall (G:ctx) (t:th) (a:exp) (A:typ),
     typing G th_log a A ->
     typing G th_log a (typ_at A t)
 | TBoxLV : forall (G:ctx) (v:exp) (A:typ),
     val v ->
     typing G th_prog v A ->
     typing G th_log v (typ_at A th_prog)
 | TBoxP : forall (G:ctx) (t:th) (a:exp) (A:typ),
     typing G t a A ->
     typing G th_prog a (typ_at A t)
 | TUnboxVal : forall (G:ctx) (t t':th)  (a b:exp) (A B:typ),
     typing G t a (typ_at A t') ->
     val a ->
     typing G t' a A
 | TRoll : forall (G:ctx) (t:th) (a:exp) (A:typ),
     typing G t a ({0 ~~> typ_mu A} A)  ->
     typing G t (exp_roll a) (typ_mu A)
 | TUnroll : forall (G:ctx) (a:exp) (A:typ),
     typing G th_prog a (typ_mu A) ->
     typing G th_prog (exp_unroll a) ({0 ~~> typ_mu A} A).
(** definitions *)


(* A little infrastruture for dealing with nested @ types.
   Used in both progress and preservation *)
Inductive Boxed (A : typ) (t : th) : typ -> Prop :=
  | BOne : Boxed A t (typ_at A t)
  | BMany : forall B t',
      Boxed A t B -> Boxed A t (typ_at B t').
Hint Immediate BOne BMany.

Ltac is_nonbox_constr A :=
  match A with
  | typ_nat => idtac
  | typ_arr _ _ => idtac
  | typ_at _ _ => fail 1
  | typ_sum _ _  => idtac
  | typ_var _ => idtac
  | typ_mu _ => idtac
  end.

Hint Extern 1 =>
  match goal with
  | [ H : Boxed _ _ ?A |- _ ] => 
    is_nonbox_constr A; inversion H
  end.

Hint Constructors val step mstep astep.
Hint Immediate FONat FOAt FOSum.
Hint Resolve TVar TApp TLam TRec TZero TSuc
             TInl TInr TBoxL TBoxLV TBoxP
             TRoll TUnroll.
Hint Immediate TSub.
Hint Extern 4 =>
  match goal with
  | [ HFO : FO ?A |- typing _ th_log _ ?A ] =>
    apply TFOVal
  | [ |- typing _ th_log _ typ_nat ] =>
    apply TFOVal
  | [ |- typing _ th_log _ (typ_at _ _)] =>
    apply TFOVal
  | [ H1 : FO ?A1, H2 : FO ?A2 |- typing _ th_log _ (typ_sum ?A1 ?A2)] =>
    apply TFOVal
  | [ HFO : FO ?A |- typing _ th_log _ ({_ ~> _}?A) ] =>
    apply TFOVal
  | [ HFO : FO ?A |- typing _ th_log _ ({_ ~> _}?A) ] =>
    apply TFOVal
  | [ |- typing _ _ (exp_sumcase _ _ _) _ ] => eapply TSumCase
  | [ |- typing _ _ (exp_natcase _ _ _) _ ] => eapply TNatCase
  | [ H : typing _ _ ?v (typ_at ?A _)
     |- typing _ _ ?v ?A ] => eapply TUnboxVal
  | [ H : typing _ _ ?v (typ_at ?A _)
     |- typing _ _ (lift_e ?v _ _) ?A ] => eapply TUnboxVal
  | [ H : typing _ _ ?v (typ_at ?A _)
     |- typing _ _ ({_ ~> _} ?v) ?A ] => eapply TUnboxVal
  | [ H : typing _ th_log ?a _ |-
        typing _ th_prog ?a _ ] => apply TSub
  | [ H : typing _ th_log ?a _ |-
        typing _ th_prog (lift_e ?a _ _) _ ] => apply TSub
  | [ H : typing _ th_log ?a _ |-
        typing _ th_prog ({_ ~> _}?a) _ ] => apply TSub
      
  end.

Ltac destruct_things :=
  repeat (match goal with
  | [ H : exists _, _ |- _ ] => destruct H
  | [ H : _ /\ _ |- _ ] => destruct H
  | [ H : _ \/ _ |- _ ] => destruct H
  end).

