
Require Import Arith.
Require Import Omega.
Require Import Bool.
Require Import List.

Require Import Wf_nat.

Require Import step_defs.

(* We start by defining the ordering that gets smaller in each step of
   the definition of the interpretation.  It's called "IOrder". *)
Inductive CV :=
  | Comp : CV
  | Val  : CV.

Fixpoint typ_size (A : typ) : nat :=
  match A with
  | typ_nat       => 0
  | typ_arr B1 B2 => 1 + typ_size B1 + typ_size B2
  | typ_at  A' _  => 1 + typ_size A'
  | typ_var _     => 0
  | typ_mu A'     => 1 + typ_size A'
  | typ_sum B1 B2 => 1 + typ_size B1 + typ_size B2
  end.

(* Some lemmas about type size to help automation *)
Lemma ts_arr_sub1 : forall A B,
  typ_size A < typ_size (typ_arr A B).
Proof. intros; simpl; omega. Defined.

Lemma ts_arr_sub2 : forall A B,
  typ_size B < typ_size (typ_arr A B).
Proof. intros; simpl; omega. Defined.

Lemma ts_at_sub : forall A t,
  typ_size A < typ_size (typ_at A t).
Proof. intros; simpl; omega. Defined.

Lemma ts_mu_sub : forall A,
  typ_size A < typ_size (typ_mu A).
Proof. intros; simpl; omega. Defined.

Lemma ts_sum_sub1 : forall A B,
  typ_size A < typ_size (typ_sum A B).
Proof. intros; simpl; omega. Defined.

Lemma ts_sum_sub2 : forall A B,
  typ_size B < typ_size (typ_sum A B).
Proof. intros; simpl; omega. Defined.
Implicit Arguments ts_arr_sub1 [A B].
Implicit Arguments ts_arr_sub2 [A B].
Implicit Arguments ts_at_sub [A t].
Implicit Arguments ts_sum_sub1 [A B].
Implicit Arguments ts_sum_sub2 [A B].


Notation ntc := (nat * typ * CV) % type.

Inductive IOrder : ntc -> ntc -> Prop :=
  | IONat : forall n1 n2 A1 A2 cv1 cv2,
     n1 < n2 -> IOrder (n1,A1,cv1) (n2,A2,cv2)
  | IOTyp : forall n A1 A2 cv1 cv2,
     typ_size A1 < typ_size A2 -> IOrder (n,A1,cv1) (n,A2,cv2)
  | IOCV  : forall n A,
      IOrder (n,A,Val) (n,A,Comp)
.
Hint Constructors IOrder.
Implicit Arguments IONat [n1 n2 A1 A2 cv1 cv2].
Implicit Arguments IOTyp [n A1 A2 cv1 cv2].

(* Next we prove that IOrder is well-founded, then we prove that some
   particular things are related by IOrder (these will come in handy
   in the definition of the interpretation). *)
Theorem wfIO : well_founded IOrder.
Proof.
  intros p1; destruct p1 as [[n A] cv]. revert A cv.

  apply (lt_wf_rec n (fun n => forall A cv, Acc IOrder (n, A, cv)));
  clear n; intros n Hwfnat A.

  apply (induction_ltof1 _ typ_size 
                        (fun A => forall cv, Acc IOrder (n,A,cv)));
  clear A; intros A Hwftyp cv.

  apply Acc_intro; intros p2 HOrd; inversion HOrd; subst; eauto;
  apply Acc_intro; intros p3 HOrd'; inversion HOrd'; subst; eauto.
Defined.
Hint Resolve wfIO.

Theorem IOrder_minus : forall k j A,
  IOrder (k-j, A, Val) (k, A, Comp).
Proof.
  intros; destruct j;
  [ rewrite <- minus_n_O; apply IOCV |
    destruct k; simpl; eauto; apply IONat; omega ].
Defined.
Hint Resolve IOrder_minus.

Theorem IO_le_sub : forall k j A B cv cv',
     j <= k 
  -> typ_size A < typ_size B
  -> IOrder (j, A, cv) (k, B, cv').
Proof.
  intros; match goal with
  | [ Hleq : _ <= _ |- _] => destruct (le_lt_or_eq _ _ Hleq)
  end; subst; eauto.
Defined. 
Implicit Arguments IO_le_sub [k j A B cv cv'].

(* Next, the interpretation of nats *)
Fixpoint looks_like_nat (a : exp) : Prop :=
  match a with
  | exp_zero => True
  | exp_succ a' => looks_like_nat a'
  | _ => False
  end.

(********************************************************************)
(* The interpretation ***********************************************)

(* interp_innards is the main definition for the interpretation.  It's
   not recursive---instead, it takes a parameter that represents the
   recursive call.  Every time we use this parameter, we must prove
   that we are calling it on "smaller" things, according to IOrder.
   After, we use coq magic to tie the knot.
 *)
Definition interp_innards p1 :
     (forall p2, IOrder p2 p1 -> th -> exp -> Prop)
  -> th -> exp -> Prop := 
  match p1 as p1
    return    (forall p2, IOrder p2 p1 -> th -> exp -> Prop) 
           -> th -> exp -> Prop
    with
  | (k,A,cv) =>
    match cv as cv
      return   (forall y, IOrder y (k, A, cv) -> th -> exp -> Prop)
            -> th -> exp -> Prop
    with 
    | Comp => fun f t a => 
          typing nil t a A
       /\ match t with
          | th_prog => forall j v, j <= k -> mstep j a v -> val v
                         -> f (k - j, A, Val) (IOrder_minus _ _ _)
                              th_prog v
          | th_log => exists v, astep a v
                             /\ f (k,A,Val) (IOCV _ _) th_log v
          end
    | Val =>
      match A as A
        return (forall y, IOrder y (k, A, Val) -> th -> exp -> Prop)
            -> th -> exp -> Prop
      with
      | typ_var _    => fun _ _ _ => False
      | typ_nat      => fun _ _ a => looks_like_nat a
      | typ_at A' t' => fun f t a => f (k,A',Val) (IOTyp ts_at_sub) t' a
      | typ_arr B1 B2 => fun f t a =>
            typing nil t a (typ_arr B1 B2)
         /\ match t with
            | th_log =>
              match a with
              | exp_lam b => 
                  forall j v (Hle : j <= k),
                              f (j,B1,Val) (IO_le_sub Hle ts_arr_sub1)
                                th_log v
                           -> f (j,B2,Comp) (IO_le_sub Hle ts_arr_sub2)
                                th_log ({0 ~> v}b)
              | _ => False
              end
            | th_prog =>
              match a with
              | exp_lam b =>
                  forall j v (Hlt : j < k),
                              f (j,B1,Val) (IONat Hlt) th_prog v
                           -> f (j,B2,Comp) (IONat Hlt) 
                                th_prog ({0 ~> v}b)
              | exp_fun b =>
                  forall j v (Hlt : j < k),
                              f (j,B1,Val) (IONat Hlt) th_prog v
                           -> f (j,B2,Comp) (IONat Hlt) th_prog
                                     ({0 ~> v} ({0 ~> exp_fun b} b))
              | _ => False
              end 
            end
      | typ_mu B => fun f t a => 
        match a with
        | exp_roll a' => 
             val a' /\ typing nil t (exp_roll a') (typ_mu B)
          /\ forall j (Hlt : j < k),
               f (j,{0 ~~> typ_mu B}B,Val) (IONat Hlt) t a'
        | _ => False
        end
      | typ_sum B1 B2 => fun f t a =>
        match a with
        | exp_inl a' => f (k,B1,Val) (IOTyp ts_sum_sub1) t a'
        | exp_inr a' => f (k,B2,Val) (IOTyp ts_sum_sub2) t a'
        | _ => False
        end
      end
    end
  end.

(* Here's the magic knot tying *)
Definition interp : ntc -> th -> exp -> Prop :=
  Fix wfIO (fun _ => th -> exp -> Prop) interp_innards.

(* We need a nice induction principle for reasoning about this
   definition.  Sadly, built in coq magic like Functional Scheme
   doesn't work.  So, we're going to have to define and prove our
   own principle.  There is a little support in the coq libraries
   for this kind of reasoning, but first we must demonstrate that
   interp is "extensional" in the sense that it doesn't care what
   proof of well-ordering it is passed.

   The problem is that the proof of extensionality that the
   coq library demands uses the propositional equality.  This makes
   sense, since it doesn't know my function actually defines
   an (exp -> Prop) for which a more extensional equality would
   work, but it's annoying.  As a result, I have to assume
   functional extensionality.  *)
Axiom extensionality : forall (A : Type) (B : A -> Type)
                              (f g : forall a : A, B a),
     (forall a : A, f a = g a) -> f = g.

Theorem interp_extensional : 
  forall (p1 : ntc) (f g : forall p2 : ntc, IOrder p2 p1 
                                         -> th -> exp -> Prop),
        (forall (p2 : ntc) (io : IOrder p2 p1), f p2 io = g p2 io) 
     -> interp_innards p1 f = interp_innards p1 g.
Proof.
  intros;
  assert (f = g) by (eauto using extensionality); subst; eauto.
Qed.
Hint Resolve interp_extensional.

(* Now a few tactics for unfolding the interpretation and cleaning up the
   resulting mess *)
Ltac unfold_cleanup :=
  repeat (match goal with
  | [ H : False |- _ ] => inversion H
  | [ H : val _ /\ _ |- _ ] =>
    let HVal := fresh "HVal" in destruct H as [HVal H]
  | [ H : typing _ _ _ _ /\ _ |- _ ] =>
    let HTyp := fresh "HTyp" in destruct H as [HTyp H]
  | [ H : match ?t with | th_prog => _ | th_log => _ end |- _ ] =>
    destruct t
  | [ H : exists v, astep _ _ /\ _ |- _] =>
    let v := fresh "v" in let Hastep := fresh "Hastep" in
    destruct H as [v [Hastep H]]
  | [ H : match ?v with | exp_var _ => _ | exp_zero => _ | exp_succ _ => _
            | exp_fun _ => _ | exp_lam _ => _ | exp_app _ _ => _
            | exp_natcase _ _ _ => _ | exp_sumcase _ _ _ => _ 
            | exp_roll _ => _ | exp_unroll _ => _
            | exp_inl _ => _ | exp_inr _ => _
          end |- _ ] =>
    destruct v; try solve [inversion H]
  end).

Tactic Notation "unfold_interp" "in" hyp(H) := 
  unfold interp in H; rewrite Fix_eq in H;
  [ simpl in H; fold interp in H | apply interp_extensional ];
  unfold_cleanup.

Ltac unfold_interp_hyps :=
  repeat (match goal with
  | [ H : interp (_,_,Comp) _ _ |- _ ] => unfold_interp in H
  | [ H : interp (_,typ_nat,Val) _ _ |- _ ] => unfold_interp in H
  | [ H : interp (_,?p _,Val) _ _ |- _ ] => unfold_interp in H
  | [ H : interp (_,?p _ _,Val) _ _ |- _ ] => unfold_interp in H
  | [ H : interp (_,?p _ _ _,Val) _ _ |- _ ] => unfold_interp in H
  end).

Ltac unfold_interp :=
  unfold interp; rewrite Fix_eq;
  [ simpl; fold interp | apply interp_extensional ];
  repeat match goal with
  | [ |- typing _ _ _ _ /\ _ ] => split; [ eauto | ]
  | [ |- match ?t with | th_log => _ | th_prog => _ end ] =>
    destruct t
  end.


(* Induction Scheme doesn't seem to work for interp.  So!  We define our
   own induction principle by hand ... *)
Theorem interp_induction :
     forall (P : nat -> typ -> CV -> th -> exp -> Prop),
     (forall k A a, typing nil th_prog a A
        -> (forall j v, j <= k -> mstep j a v -> val v
                     -> P (k - j) A Val th_prog v)
        -> (forall j v, j <= k -> mstep j a v -> val v
                     -> interp ((k - j),A,Val) th_prog v)
        -> P k A Comp th_prog a)
  -> (forall k A a v, typing nil th_log a A
        -> astep a v
        -> interp (k,A,Val) th_log v
        -> P k A Val th_log v
        -> P k A Comp th_log a)
  -> (forall k t a, looks_like_nat a -> P k typ_nat Val t a)
  -> (forall k A t' t a,
           interp (k,A,Val) t' a
        -> P k A Val t' a
        -> P k (typ_at A t') Val t a)
  -> (forall k A B a, typing nil th_log (exp_lam a) (typ_arr A B)
        -> (forall j v, j <= k
                -> interp (j,A,Val) th_log v 
                -> interp (j,B,Comp) th_log ({0 ~> v}a))
        -> (forall j v, j <= k
                -> interp (j,A,Val) th_log v
                -> P j A Val th_log v 
                -> P j B Comp th_log ({0 ~> v}a))
        -> (forall j v, j <= k
                -> interp (j,A,Val) th_log v
                -> P j A Val th_log v)
        -> P k (typ_arr A B) Val th_log (exp_lam a))
  -> (forall k A B a, typing nil th_prog (exp_lam a) (typ_arr A B)
        -> (forall j v, j < k
                -> interp (j,A,Val) th_prog v 
                -> interp (j,B,Comp) th_prog ({0 ~> v}a))
        -> (forall j v, j < k
                -> interp (j,A,Val) th_prog v
                -> P j A Val th_prog v 
                -> P j B Comp th_prog ({0 ~> v}a))
        -> (forall j v, j < k 
              -> interp (j,A,Val) th_prog v -> P j A Val th_prog v)
        -> P k (typ_arr A B) Val th_prog (exp_lam a))
  -> (forall k A B a, typing nil th_prog (exp_fun a) (typ_arr A B)
        -> (forall j v, j < k
                -> interp (j,A,Val) th_prog v 
                -> interp (j,B,Comp) th_prog ({0 ~> v}({0 ~> exp_fun a} a)))
        -> (forall j v, j < k
                -> interp (j,A,Val) th_prog v 
                -> P j A Val th_prog v
                -> P j B Comp th_prog ({0 ~> v}({0 ~> exp_fun a} a)))
        -> (forall j v, j < k 
              -> interp (j,A,Val) th_prog v -> P j A Val th_prog v)
        -> P k (typ_arr A B) Val th_prog (exp_fun a))
  -> (forall k A t a,
           val a
        -> typing nil t (exp_roll a) (typ_mu A)
        -> (forall j, j < k -> interp (j,{0 ~~> typ_mu A}A,Val) t a)
        -> (forall j, j < k -> P j ({0 ~~> typ_mu A}A) Val t a)
        -> P k (typ_mu A) Val t (exp_roll a))
  -> (forall k A B t a,
           interp (k,A,Val) t a
        -> P k A Val t a
        -> P k (typ_sum A B) Val t (exp_inl a))
  -> (forall k A B t a,
           interp (k,B,Val) t a
        -> P k B Val t a
        -> P k (typ_sum A B) Val t (exp_inr a))
  -> (forall k A cv t a, interp (k,A,cv) t a -> P k A cv t a).
Proof.
  intros P HCProg HCLog HVNat HVAt HVLamL HVLamP HVFun HVMu HVInl HVInr.
  induction k as [k IHWf] using lt_wf_ind; induction A; destruct cv; intros;
  unfold_interp_hyps;
  try match goal with
  | [ |- P _ typ_nat Comp th_log _ ] => eauto
  | [ |- P _ typ_nat Val _ _ ] => eauto
  | [ Hstep : astep ?a ?v'
     |- P _ _ Comp th_log ?a ] =>
    apply HCLog with (v := v'); eauto;
    try unfold_interp; eauto
  | [ HRed : forall _ _, _ <= _ -> mstep _ ?a _ -> val _ -> interp _ _ _
     |- P _ _ Comp th_prog ?a ] =>
    apply HCProg; eauto; intros ? ? HLe HMs HVal;
    let HInterp := fresh "HInterp" in
    assert (HInterp := HRed _ _ HLe HMs HVal); clear HRed;
    unfold_interp_hyps; eauto
  end;
  try match goal with
  | [ |- P _ (typ_arr _ _) Val th_prog (exp_fun _) ] =>
    apply HVFun; eauto; intros; apply IHWf; eauto; try omega
  | [ |- P _ (typ_arr _ _) Val th_prog (exp_lam _) ] =>
    apply HVLamP; eauto; intros; apply IHWf; eauto; try omega
  | [ |- P _ (typ_arr _ _) Val th_log _ ] => 
    apply HVLamL; eauto; intros;
    match goal with
    | [ Hle : _ <= _ |- _ ] => 
      apply le_lt_or_eq in Hle; destruct Hle; subst; eauto with arith
    end
  | [ |- P (?k - ?j) (typ_at _ _) Val _ _ ] => 
    assert (k - j <= k) as Hleminus by omega; apply le_lt_or_eq in Hleminus;
    destruct Hleminus as [ Hle' | Hle' ]; eauto; try rewrite Hle' in *;
    apply HVAt; eauto
  | [ |- P (?k - ?j) (typ_sum _ _) Val _ (_ _) ] => 
    assert (k - j <= k) as Hleminus by omega; apply le_lt_or_eq in Hleminus;
    destruct Hleminus as [ Hle' | Hle' ]; eauto; try rewrite Hle' in *; eauto
  | [ |- P _ (typ_sum _ _) Val _ (_ _) ] => eauto
  | [ |- P _ (typ_at _ _) Val _ _ ] =>
    eauto
  | [ |- P _ (typ_mu _) Val _ (exp_roll _) ] =>
    apply HVMu; eauto; intros; apply IHWf; eauto; omega
  end.
Qed.

Global Opaque interp_induction.
