Require Import Arith.
Require Import Omega.
Require Import Bool.
Require Import List.

Require Import step_defs.
Require Import step_db_basics.
Require Import step_substitution.
Require Import step_preservation.
Require Import step_interp.
Require Import step_reduction_basics.

Hint Rewrite <- minus_n_O : steps.
Hint Resolve le_O_n.

Theorem looks_like_nat_zero : looks_like_nat exp_zero.
Proof. simpl; eauto.  Qed.
Hint Resolve looks_like_nat_zero.

Theorem looks_like_nat_val : forall a,
  looks_like_nat a -> val a.
Proof.
  induction a; intros Hlook; try inversion Hlook; eauto.
Qed.
Hint Resolve looks_like_nat_val.

Theorem looks_like_nat_typing : forall G t a,
     looks_like_nat a
  -> typing G t a typ_nat.
Proof.
  intros G t a; induction a; intros Hlook; try (inversion Hlook); eauto.
Qed.
Hint Immediate looks_like_nat_typing.

(* This one is nice just because it doesn't create a new existential
   for th or G *)
Theorem looks_like_nat_typing_inst : forall a,
     looks_like_nat a
  -> typing nil th_log a typ_nat.
Proof.
  eauto.
Qed.
Hint Immediate looks_like_nat_typing_inst.

Theorem looks_like_nat_step : forall a b,
  looks_like_nat a -> step a b -> False.
Proof.
  induction a; intros b Hlook HStep; inversion HStep; eauto.
Qed.
Hint Resolve looks_like_nat_step.

Theorem looks_like_nat_mstep : forall a b j,
     looks_like_nat a
  -> mstep j a b
  -> looks_like_nat b.
Proof.
  inversion 2; subst; eauto;
  edestruct looks_like_nat_step; eauto.
Qed.
Hint Resolve looks_like_nat_mstep.




Theorem interp_val : forall k A t a,
     interp (k,A,Val) t a 
  -> val a.
Proof.
  intros k A t a HInterp;
  apply interp_induction with (P := fun k A cv t a => cv = Val -> val a) 
     in HInterp;
  intros; try match goal with
  | [ H : Comp = Val |- _ ] => inversion H
  | [ H : Val = Comp |- _ ] => inversion H
  end; eauto.
Qed.
Hint Immediate interp_val.



Theorem interp_well_typed : forall k A cv t a,
     interp (k,A,cv) t a 
  -> typing nil t a A.
Proof.
  apply interp_induction; intros; eauto;
  match goal with
  | [ HI : interp (_,_,Val) _ ?a |- typing _ ?t _ (typ_at _ _) ] =>
    apply interp_val in HI; destruct t; eauto
  end.
Qed.
Hint Immediate interp_well_typed.

(* With interp_well_typed, we can give eauto a little more help *)
Hint Extern 1 =>
  match goal with
  | [ H : interp _ _ ?a |- context [{?n ~> ?b}?a] ] =>
    rewrite (open_closed_id _ _ _ (interp_well_typed _ _ _ _ _ H) n b) 
  | [ H : interp _ _ ?a |- context [lift_e ?a ?n ?m] ] =>
    rewrite (lift_closed_id _ _ _ (interp_well_typed _ _ _ _ _ H) n m)
  end.


(* Downward closure, and some instances that are good for auto *)
Theorem downward_closure : forall k A cv t a, 
     interp (k,A,cv) t a 
  -> forall j, j < k -> interp (j,A,cv) t a.
Proof.
  intros k A cv t a HI;
  apply interp_induction 
            with (P := fun k A cv t a => forall j, j < k 
                                               -> interp (j,A,cv) t a)
        in HI; intros; eauto; clear HI k A cv t a;
  unfold_interp; eauto; intros;
  repeat (match goal with
  | [ H : forall j _, j <= _ -> _ |- interp (?j' - ?j0, _,_) _ _ ] =>
    apply H with (j := j0); eauto; omega
  | [ H : forall j _, j < _ -> _ |- _ ] => 
    try solve [apply H; eauto; omega]; clear H
  | [ H : forall j _, j <= _ -> _ |- _ ] => 
    try solve [apply H; eauto; omega]; clear H
  end).
Qed.

Theorem downward_closure_leq : forall k j A cv t a,
     interp (k,A,cv) t a
  -> j <= k
  -> interp (j,A,cv) t a.
Proof.
  intros; match goal with
  | [ Hle : _ <= _ |- _ ] =>
    apply le_lt_or_eq in Hle; destruct Hle;
    subst; eauto using downward_closure
  end.
Qed.
Hint Immediate downward_closure downward_closure_leq.

Theorem downward_closure_minus : forall k j A cv t a,
     interp (k,A,cv) t a
  -> interp (k - j,A,cv) t a.
Proof.
  intros; assert (k - j <= k) by omega; eauto using downward_closure_leq.
Qed.
Hint Immediate downward_closure_minus.


(* V [A] t k  \subseteq  C [A] t k *)
Theorem comp_contains_values' : forall k A cv t a,
     interp (k,A,cv) t a 
  -> cv = Val
  -> interp (k,A,Comp) t a.
Proof.
  intros k A cv t a HI;
  apply interp_induction
   with (P := fun k A cv t a => cv = Val -> interp (k,A,Comp) t a)
     in HI; intros; eauto; clear HI k A cv t a;
  try match goal with
  | [ H : Comp = Val |- _ ] => inversion H
  end; unfold_interp;
  try match goal with
  | [ |- exists v, astep ?a v /\ interp _ _ v ] =>
    exists a; split; [ eauto | unfold_interp; eauto ]
  | [ |- forall _ _, _ <= _ -> mstep _ _ _ -> val _ -> interp _ _ _ ] =>
    intros; step_val_elim; unfold_interp; autorewrite with steps; eauto
  | [ |- typing _ ?t1 _ (typ_at _ ?t2) ] =>
    destruct t1; destruct t2; eauto
  end.
Qed.

Theorem comp_contains_values : forall k A t a,
     interp (k,A,Val) t a 
  -> interp (k,A,Comp) t a.
Proof.
  intros; eauto using comp_contains_values'.
Qed.
Hint Immediate comp_contains_values.


Hint Extern 2 =>
  match goal with
  | [ HI : interp ?p th_log ?a |- interp ?p ?t ?a ] =>
    destruct t
  end.

Theorem log_val_fo_contains_prog : forall k A a,
     interp (k,A,Val) th_prog a
  -> FO A
  -> interp (k,A,Val) th_log a.
Proof.
  intros k A a HI; 
  remember th_prog as t; revert Heqt;
  remember Val as cv; revert Heqcv;
  apply interp_induction
   with (P := fun k A cv t a =>    cv = Val -> t = th_prog -> FO A 
                                -> interp (k,A,cv) th_log a)
     in HI; intros; subst; eauto; clear HI k A cv t a; subst;
  fo_inversion;
  try match goal with
  | [ H : th_log = th_prog |- _ ] => inversion H
  | [ H : Comp = Val |- _ ] => inversion H
  end; unfold_interp; eauto.
Qed.
Hint Extern 2 =>
  match goal with
  | [ H : FO ?A |- interp (_,?A,Val) th_log _ ] =>
    apply log_val_fo_contains_prog
  end.


(* [A] L k  \subseteq  [A] P k *)
Theorem prog_contains_log : forall k A cv a,
     interp (k,A,cv) th_log a
  -> interp (k,A,cv) th_prog a.
Proof.
  intros k A cv a HI; remember th_log as t; revert Heqt;
  apply interp_induction
   with (P := fun k A cv t a => t = th_log -> interp (k,A,cv) th_prog a)
     in HI; intros; eauto; clear HI k A cv t a; subst;
  repeat (match goal with
  | [ H : th_prog = th_log |- _ ] => inversion H
  | [ H : th_log = th_log -> _ |- _ ] => let H' := fresh H in
    assert (H' := H eq_refl); clear H; rename H' into H
  end); unfold_interp; eauto; intros;
  match goal with
  | [ HAS : astep ?a ?v, HMS : mstep _ ?a ?v' |- _ ] =>
    assert (v = v') 
       by (inversion HAS; subst; eauto using mstep_val_deterministic_exp);
    subst; eauto
  | [ IH1 : forall _ _, _ <= _ -> interp _ _ _ -> _ -> _ = _ 
                    -> interp (_,_,Comp) _ _,
      IH2 : forall _ _, _ <= _ -> interp _ _ _ -> _ -> interp (_,_,Val) _ _
     |- _ ] =>
    apply IH1; eauto; try omega; intros; subst; 
    try typing_inversion; eauto;
    apply IH2; eauto; try omega
  end.
Qed.
Hint Immediate prog_contains_log.


(******************************************)
(******** Term environments (rhos) ********)
Notation TermEnv := (list exp).
 
(* the term interp is really just a multi-substitution *)
Fixpoint TermInterp (rho : TermEnv) (a : exp) (m : nat) : exp :=
  match rho with
  | nil       => a
  | b :: rho' => TermInterp rho' ({m ~> lift_e b m 0} a) m
  end.

Theorem TermInterp_closed_id : forall rho a m t A,
  typing nil t a A -> TermInterp rho a m = a.
Proof.
  induction rho; intros; simpl; eauto.
Qed.
Hint Extern 1 =>
  match goal with
  | [ HTyp : typing nil _ ?a _ |- context [TermInterp _ ?a _] ] =>
    rewrite (TermInterp_closed_id _ _ _ _ _ HTyp)
  | [ HInt : interp (?k,?A,?cv) ?t ?a |- context [TermInterp _ ?a _] ] =>
    rewrite (TermInterp_closed_id _ _ _ _ _ 
               (interp_well_typed _ _ _ _ _ HInt))
  end.

Inductive wfTermEnv (k : nat) : ctx -> TermEnv -> Prop :=
  | wfTENil  : wfTermEnv k nil nil
  | wfTECons : forall G rho t A a,
       interp (k,A,Val) t a
    -> wfTermEnv k G rho
    -> wfTermEnv k ((t,A) :: G) (a :: rho).
Hint Constructors wfTermEnv.

Theorem TermEnv_substitution : forall rho G2 k,
     wfTermEnv k G2 rho
  -> forall G1 t a A, typing (G1 ++ G2) t a A
                -> typing G1 t (TermInterp rho a (length G1)) A.
Proof.
  induction 1; intros; simpl; eauto;
  match goal with
  | [ HTyp : typing (?G ++ nil) _ _ _ |- typing ?G _ _ _ ] =>
    rewrite app_nil_r in HTyp; auto
  | [ IH : forall _ _ _ _, typing _ _ _ _ -> typing _ _ _ _ |- _ ] =>
    eauto using weakening_nil
  end.
Qed.
Hint Resolve TermEnv_substitution.

Theorem TermEnv_substitution_cons : forall k G rho t' A' t a A,
     wfTermEnv k G rho
  -> typing ((t',A') :: G) t a A
  -> typing ((t',A') :: nil) t (TermInterp rho a 1) A.
Proof.
  intros; match goal with
  | [ |- typing ((?t',?A') :: nil) _ _ _ ] =>
    replace ((t',A') :: nil) with (((t',A') :: nil) ++ nil) by (simpl; auto);
    eapply TermEnv_substitution; eauto
 end.
Qed.
Hint Resolve TermEnv_substitution_cons.

Theorem TermEnv_substitution_cons_2 : forall k G rho t1 A1 t2 A2 t a A,
     wfTermEnv k G rho
  -> typing ((t1,A1) :: (t2,A2) :: G) t a A
  -> typing ((t1,A1) :: (t2,A2) :: nil) t (TermInterp rho a 2) A.
Proof.
  intros; match goal with
  | [ |- typing ((?t1,?A1) :: (?t2,?A2) :: nil) _ _ _ ] =>
    replace ((t1,A1) :: (t2,A2) :: nil) 
       with (((t1,A1) :: (t2,A2) :: nil) ++ nil) by (simpl; auto);
    eapply TermEnv_substitution; eauto
 end.
Qed.
Hint Resolve TermEnv_substitution_cons_2.


Theorem TermEnv_substitution_nil : forall k G rho t a A,
     wfTermEnv k G rho
  -> typing G t a A
  -> typing nil t (TermInterp rho a 0) A.
Proof. intros; eapply TermEnv_substitution; eauto. Qed.
Hint Resolve TermEnv_substitution_nil.
Hint Extern 1 =>
  match goal with
  | [ H : typing ?G ?t (?P ?a) _, Hwf : wfTermEnv _ ?G ?rho
     |- typing nil ?t (?P (TermInterp ?rho ?a _)) _] =>
    eapply (TermEnv_substitution_nil _ _ _ _ _ _ Hwf) in H;
    autorewrite with steps in *; eauto
  | [ H : typing ?G ?t (?P _ _) ?A, Hwf : wfTermEnv _ ?G ?rho
     |- typing nil ?t (?P (TermInterp ?rho _ _) _) ?A] =>
    eapply (TermEnv_substitution_nil _ _ _ _ _ _ Hwf) in H;
    autorewrite with steps in *; eauto
  end.


Theorem wfTermEnv_downward_closure : forall k G rho,
  wfTermEnv k G rho -> forall j, j < k -> wfTermEnv j G rho.
Proof.
  induction 1; intros; eauto.
Qed.
Hint Immediate wfTermEnv_downward_closure.

Theorem use_wfTermEnv' : forall k G rho,
     wfTermEnv k G rho
  -> forall G1 t A G2, G = G1 ++ (t,A) :: G2 
       -> interp (k,A,Val) t (TermInterp rho (exp_var (length G1)) 0).
Proof.
  induction 1; intros; subst; eauto;
  match goal with
  | [ HListEq : _ = ?G1 ++ _ :: _ |- _ ] =>
    destruct G1 as [ | [? ?] G1]; simpl in *; inversion HListEq; subst
  end; autorewrite with steps; eauto.
Qed.

Theorem use_wfTermEnv : forall k G1 t A G2 rho,
     wfTermEnv k (G1 ++ (t,A) :: G2) rho
  -> interp (k,A,Val) t (TermInterp rho (exp_var (length G1)) 0).
Proof.
  intros; eauto using use_wfTermEnv'.
Qed.
Hint Resolve use_wfTermEnv.

Theorem TIOpenSwap : forall rho n v b t A,
     typing nil t v A
  ->   {n ~> v} (TermInterp rho b (S n))
     = TermInterp (v :: rho) b n.
Proof.
  induction rho; intros; simpl in *; eauto.
  try match goal with
  | [ IH : forall _ _ _ _ _, typing _ _ _ _ -> _ = _,
      HTyp : typing nil _ _ _ |- _ ] =>
    rewrite open_open_swap; eauto; erewrite IH; eauto;
    autorewrite with steps; eauto; rewrite plus_comm; eauto
  end.
Qed.

Hint Extern 2 =>
  match goal with
  | [ HTyp : typing nil _ ?a _ |- context [{_ ~> ?v}(TermInterp _ _ _)] ] =>
    rewrite (TIOpenSwap _ _ _ _ _ _ HTyp)
  | [ HN : looks_like_nat ?v
     |- context [{_ ~> ?v}(TermInterp _ _ _)] ] =>
    rewrite (TIOpenSwap _ _ _ _ _ _ (looks_like_nat_typing nil th_log _ HN))
  | [ HInt : interp _ _ ?v
     |- context [{_ ~> ?v}(TermInterp _ _ _)] ] =>
    rewrite (TIOpenSwap _ _ _ _ _ _ (interp_well_typed _ _ _ _ _ HInt))
  | [ HTyp : typing nil _ ?a _,
      HTI : context [{_ ~> ?v}(TermInterp _ _ _)] |- _ ] =>
    rewrite (TIOpenSwap _ _ _ _ _ _ HTyp) in HTI
  | [ HN : looks_like_nat ?v,
      HTI : context [{_ ~> ?v}(TermInterp _ _ _)] |- _ ] =>
    rewrite (TIOpenSwap _ _ _ _ _ _ 
                        (looks_like_nat_typing nil th_log _ HN)) in HTI
  | [ HInt : interp _ _ ?v,
      HTI : context [{_ ~> ?v}(TermInterp _ _ _)] |- _ ] =>
    rewrite (TIOpenSwap _ _ _ _ _ _
                 (interp_well_typed _ _ _ _ _ HInt)) in HTI
  end.

Theorem TI_preserves_val : forall rho v m G k,
     wfTermEnv k G rho
  -> val v
  -> val (TermInterp rho v m).
Proof.
  induction rho; intros; simpl; eauto;
  match goal with
  | [ HTI : wfTermEnv _ _ (_ :: _) |- _ ] =>
    inversion HTI; subst; eauto
  end.
Qed.
Hint Resolve TI_preserves_val.

Hint Extern 3 =>
  match goal with
  | [ IH : forall _ _, _ <= ?k -> mstep _ ?a _ -> _ 
                               -> interp (?k - _,_,_) _ _
     |- interp (?k,_,_) _ ?a ] =>
  replace k with (k - 0) by omega
  end.


Theorem comp_interp_of_val : forall m A t a,
     val a
  -> interp (m, A, Comp) t a
  -> interp (m, A, Val) t a.
Proof.
  intros; unfold_interp_hyps; try step_val_elim; eauto.
Qed.
Hint Extern 4 => 
  match goal with
  | [ H : val ?a |- interp (_,_,Val) _ ?a ] =>
    apply (comp_interp_of_val _ _ _ _ H)
  | [ |- interp (_,_,Val) _ (exp_fun ?a) ] =>
    let HVal := fresh "HVal" in
    assert (val (exp_fun a)) as HVal by eauto;
    apply (comp_interp_of_val _ _ _ _ HVal)
  end.
