Require Import Arith.
Require Import Bool.
Require Import List.

Require Import step_defs.
Require Import step_db_basics.

(* This file mostly contains a bunch of very boring congruence and
   inversion lemmas that have to do with reduction. *)

(********************)
(* Some facts about values *)
Theorem values_dont_step : forall a,
  val a -> forall b, step a b -> False.
Proof.
  induction 1; intros ? HStep;
  inversion HStep; subst; eauto.
Qed.

Theorem val_mstep : forall j a b,
  val a -> mstep j a b -> j = 0 /\ a = b.
Proof.
  inversion 2; eauto; subst;
  edestruct values_dont_step; eauto.
Qed.

Theorem val_astep : forall a b,
  val a -> astep a b -> a = b.
Proof.
  inversion 2; 
  match goal with
  | [ H : mstep _ _ _ |- _ ] =>
    apply val_mstep in H; eauto; destruct H; subst; eauto
  end.
Qed.

Ltac step_val_elim :=
  match goal with
  | [ HS : step ?a _ |- _ ] =>
    let HVal := fresh "HVal" in assert (val a) as HVal by eauto;
    destruct (values_dont_step _ HVal _ HS)
  | [ HS : mstep _ ?a _ |- _ ] =>
    let HVal := fresh "HVal" in assert (val a) as HVal by eauto;
    destruct (val_mstep _ _ _ HVal HS); subst
  | [ HS : astep ?a _ |- _ ] =>
    let HVal := fresh "HVal" in assert (val a) as HVal by eauto;
    rewrite <- (val_astep _ _ HVal HS) in *; subst
  end.

Ltac val_inversion :=
  repeat (match goal with
  | [ HV : val (exp_succ _) |- _ ] => inversion HV; subst; clear HV
  | [ HV : val (exp_inl _) |- _ ] => inversion HV; subst; clear HV
  | [ HV : val (exp_inr _) |- _ ] => inversion HV; subst; clear HV
  | [ HV : val (exp_roll _) |- _ ] => inversion HV; subst; clear HV
  | [ HV : val (exp_natcase _ _ _) |- _ ] => inversion HV
  | [ HV : val (exp_sumcase _ _ _) |- _ ] => inversion HV
  | [ HV : val (exp_app _ _) |- _ ] => inversion HV
  | [ HV : val (exp_unroll _) |- _ ] => inversion HV
  end).

(* Congruence rules for MS *)
Theorem MSTrans : forall j1 j2 a b c,
  mstep j1 a b -> mstep j2 b c -> mstep (j1 + j2) a c.
Proof.
  induction 1; intros; simpl; eauto.
Qed.

Theorem MSApp1 : forall j a a' b,
  mstep j a a' -> mstep j (exp_app a b) (exp_app a' b).
Proof.
  induction 1; intros; simpl; eauto.
Qed.

Theorem MSApp2 : forall j a b b',
  mstep j b b' -> val a -> mstep j (exp_app a b) (exp_app a b').
Proof.
  induction 1; intros; simpl; eauto.
Qed.

Theorem MSSucc : forall j a a',
  mstep j a a' -> mstep j (exp_succ a) (exp_succ a').
Proof. induction 1; intros; simpl; eauto. Qed.

Theorem MSRoll : forall j a a',
  mstep j a a' -> mstep j (exp_roll a) (exp_roll a').
Proof. induction 1; intros; simpl; eauto. Qed.

Theorem MSNatCase1 : forall j a a' b1 b2,
     mstep j a a' 
  -> mstep j (exp_natcase a b1 b2) (exp_natcase a' b1 b2).
Proof.
  induction 1; eauto.
Qed.

Theorem MSSumCase1 : forall j a a' b1 b2,
     mstep j a a' 
  -> mstep j (exp_sumcase a b1 b2) (exp_sumcase a' b1 b2).
Proof.
  induction 1; eauto.
Qed.

Theorem MSInl : forall a b n,
  mstep n a b -> mstep n (exp_inl a) (exp_inl b).
Proof. induction 1; eauto.  Qed.

Theorem MSInr : forall a b n,
  mstep n a b -> mstep n (exp_inr a) (exp_inr b).
Proof. induction 1; eauto.  Qed.

Theorem MSInlInversion : forall j a b',
     mstep j (exp_inl a) b' 
  -> exists b, exp_inl b = b' /\ mstep j a b.
Proof.
  intros j a b' HMS; remember (exp_inl a) as inla;
  revert a Heqinla; induction HMS; intros; subst; eauto.
  match goal with
  | [ IH : forall _, _ = _ -> _, Hstep : step (exp_inl _) _ |- _ ] =>
    inversion Hstep; subst; edestruct IHHMS as [? [? ?]]; eauto
  end.
Qed.

Theorem MSInrInversion : forall j a b',
     mstep j (exp_inr a) b' 
  -> exists b, exp_inr b = b' /\ mstep j a b.
Proof.
  intros j a b' HMS; remember (exp_inr a) as inra;
  revert a Heqinra; induction HMS; intros; subst; eauto.
  match goal with
  | [ IH : forall _, _ = _ -> _, Hstep : step (exp_inr _) _ |- _ ] =>
    inversion Hstep; subst; edestruct IHHMS as [? [? ?]]; eauto
  end.
Qed.

Theorem MSSuccInversion : forall j a b',
  mstep j (exp_succ a) b' -> exists b, b' = exp_succ b /\ mstep j a b.
Proof.
  intros j a b' HStep; remember (exp_succ a) as a';
  revert a Heqa'; induction HStep; eauto.
  match goal with
  | [ HStep : step _ _, H : forall _, _ -> exists _, _ |- _ ] =>
    intros; subst; inversion HStep; subst;
    edestruct IHHStep as [? [? ?]]; eauto
  end.
Qed.

Theorem MSRollInversion : forall j a b',
     mstep j (exp_roll a) b' 
  -> exists b, exp_roll b = b' /\ mstep j a b.
Proof.
  intros j a b' HMS; remember (exp_roll a) as rolla;
  revert a Heqrolla; induction HMS; intros; subst; eauto;
  match goal with
  | [ IH : forall _, _ = _ -> _, Hstep : step (exp_roll _) _ |- _ ] =>
    inversion Hstep; subst; edestruct IHHMS as [? [? ?]]; eauto
  end.
Qed.

(* Congruence rules for AS *)
Theorem ASTrans : forall a b c,
  astep a b -> astep b c -> astep a c.
Proof.
  inversion 1; inversion 1; subst; eauto using MSTrans.
Qed.   

Theorem ASApp1 : forall a a' b,
  astep a a' -> astep (exp_app a b) (exp_app a' b).
Proof.
  inversion 1; subst; eauto using MSApp1.
Qed.

Theorem ASApp2 : forall a b b',
  astep b b' -> val a -> astep (exp_app a b) (exp_app a b').
Proof.
  inversion 1; subst; eauto using MSApp2.
Qed.

Theorem ASSucc : forall a b,
  astep a b -> astep (exp_succ a) (exp_succ b).
Proof.
  inversion 1; subst; eauto using MSSucc.
Qed.

Theorem ASRoll : forall a b,
  astep a b -> astep (exp_roll a) (exp_roll b).
Proof.
  inversion 1; subst; eauto using MSRoll.
Qed.

Theorem ASStep : forall a b,
  step a b -> astep a b.
Proof. eauto. Qed.

Theorem ASNatCase1 : forall a a' b1 b2,
  astep a a' -> astep (exp_natcase a b1 b2) (exp_natcase a' b1 b2).
Proof.
  inversion 1; eauto using MSNatCase1.
Qed.

Theorem ASSumCase1 : forall a a' b1 b2,
  astep a a' -> astep (exp_sumcase a b1 b2) (exp_sumcase a' b1 b2).
Proof.
  inversion 1; eauto using MSSumCase1.
Qed.

Theorem ASInl : forall a b,
  astep a b -> astep (exp_inl a) (exp_inl b).
Proof. inversion 1; subst; eauto using MSInl.  Qed.

Theorem ASInr : forall a b,
  astep a b -> astep (exp_inr a) (exp_inr b).
Proof. inversion 1; subst; eauto using MSInr.  Qed.
Hint Immediate ASSucc ASInl ASInr ASRoll.

(* reduction is deterministic *)
Theorem step_deterministic : forall a b,
  step a b -> forall b', step a b' -> b = b'.
Proof.
  induction 1; subst; inversion 1; subst; eauto;
  try solve [step_val_elim];
  try match goal with
  | [ IHstep : forall _, step _ _ -> _ = _ |- _ ] =>
    erewrite IHstep; eauto
  end.
Qed.

Theorem mstep_val_deterministic : forall b b' a j j',
     mstep j a b 
  -> mstep j' a b'
  -> val b -> val b' -> j = j' /\ b = b'.
Proof.
  intros b b' a j j' HMS1 HMS2; revert b' j' HMS2;
  induction HMS1; subst; inversion 1; subst; intros; eauto;
  try solve [step_val_elim];
  match goal with
  | [ HStep1 : step ?a ?b, HStep2 : step ?a ?b',
      HMS1 : mstep _ ?b ?c, HMS2 : mstep _ ?b' ?c',
      IH : forall _ _, mstep _ _ _ -> _
     |- _ /\ ?c = ?c'] => 
    rewrite <- (step_deterministic _ _ HStep1 _ HStep2) in *;
    destruct (IH  _ _ HMS2); subst; eauto
  end.
Qed.

Theorem mstep_val_deterministic_exp : forall b b' a j j',
     mstep j a b 
  -> mstep j' a b'
  -> val b -> val b' -> b = b'.
Proof.
  intros b b' a j j' HMS1 HMS2 HV1 HV2;
  destruct (mstep_val_deterministic _ _ _ _ _ HMS1 HMS2 HV1 HV2);
  subst; eauto.
Qed.

(* Inversion lemmas for the reduction forms.  We need to be very
   explicit about the number of steps for the main soundness
   theorem. *)
Theorem mstep_unroll_val_inversion : forall a c j,
     mstep j (exp_unroll a) c
  -> val c
  -> exists j1,
         j = S j1
      /\ mstep j1 a (exp_roll c).
Proof.
  intros a c j HMS; remember (exp_unroll a) as eroll;
  revert a Heqeroll; induction HMS; intros; subst;
  val_inversion;
  match goal with
  | [ H  : step (exp_unroll _) _,
      IH : forall _, _ = _ -> val _ -> _
     |- exists _, S ?n = _ /\ _ ] =>
    inversion H; subst;
    match goal with
    | [ HS : step (exp_unroll (exp_roll ?b)) ?b |- _ ] =>
      exists n; eauto using MSRoll
    | [ HS : step (exp_unroll _) (exp_unroll _) |- _ ] =>
      edestruct IH as [? [? ?]]; subst; eauto
    end
  end.
Qed.

Theorem mstep_app_val_inversion : forall a b c j,
     mstep j (exp_app a b) c
  -> val c
  -> exists j1, exists j2, exists j3, exists a', exists vb,
         val vb
      /\ j = 1 + j1 + j2 + j3
      /\ mstep j1 b vb
      /\ (   (   mstep j2 a (exp_lam a') 
              /\ mstep j3 ({0 ~> vb} a') c)
          \/ (   mstep j2 a (exp_fun a') 
              /\ mstep j3 ({0 ~> vb} 
                           ({0 ~> exp_fun (lift_e a' 1 2)} a')) c)).
Proof.
  intros a b c j HMS; remember (exp_app a b) as eapp;
  revert a b Heqeapp; induction HMS; intros; subst;
  val_inversion.
  match goal with
  | [ H : step (exp_app _ _) _,
      IH : forall _ _, _ = _ -> val _ -> _
     |- exists _, exists _, exists _, exists _, exists _,
        val _ /\ S ?n = _ /\ _ ] =>
    inversion H; subst;
    match goal with
    | [ HS : step (exp_app (exp_fun _) _) ({_ ~> _} ({_ ~> _} _)) |- _ ] =>
      exists 0; exists 0; exists n; eauto 8
    | [ HS : step (exp_app (exp_lam _) _) ({_ ~> _} _) |- _ ] =>
      exists 0; exists 0; exists n; eauto 8
    | [ HS : step (exp_app ?a1 ?b) (exp_app ?a2 ?b) |- _ ] =>
      edestruct IH as [j1 [j2 [j3 [abody [vb [? [? 
                        [? [ [? ?] | [? ?] ]]]]]]]]]; 
      auto; subst; 
      exists j1; exists (S j2); exists j3; exists abody; exists vb;
      repeat (split; eauto; try omega)
    | [ HS : step (exp_app ?a ?b1) (exp_app ?a ?b2) |- _ ] =>
      edestruct IH as [j1 [j2 [j3 [abody [vb [? [? 
                        [? [ [? ?] | [? ?] ]]]]]]]]]; 
      auto; subst; 
      exists (S j1); exists j2; exists j3; exists abody; exists vb;
      repeat (split; eauto; try omega)
    end
  end.
Qed.


Theorem mstep_natcase_val_inversion : forall a b1 b2 c j,
     mstep j (exp_natcase a b1 b2) c
  -> val c
  -> exists j1, exists j2,
         j = 1 + j1 + j2
      /\ (   (   mstep j1 a exp_zero
              /\ mstep j2 b1 c)
          \/ (exists va,
                 val va
              /\ mstep j1 a (exp_succ va)
              /\ mstep j2 ({0 ~> va} b2) c)).
Proof.
  intros a b1 b2 c j HMS; remember (exp_natcase a b1 b2) as ecase;
  revert a b1 b2 Heqecase; induction HMS; intros; subst;
  val_inversion;
  match goal with
  | [ H : step (exp_natcase _ _ _) _,
      IH : forall _ _ _, _ = _ -> val _ -> _
     |- exists _, exists _, S ?n = _ /\ _ ] =>
    inversion H; subst;
    try match goal with
    | [ HS : step (exp_natcase exp_zero _ _) _ |- _ ] =>
      exists 0; exists n; eauto 6
    | [ HS : step (exp_natcase (exp_succ _) _ _) _ |- _ ] =>
      exists 0; exists n; eauto 6
    | [ HS : step (exp_natcase _ _ _) (exp_natcase _ _ _) |- _ ] =>
      edestruct IH as [j1 [j2 [? [ [? ?] | [? [? [? ?]]] ] ]]]; auto;
      subst; exists (S j1); exists j2;
      (repeat (split; eauto 6; try omega))
    end
  end.
Qed.

Theorem mstep_sumcase_val_inversion : forall a b1 b2 c j,
     mstep j (exp_sumcase a b1 b2) c
  -> val c
  -> exists j1, exists j2,
         j = 1 + j1 + j2
      /\ (exists va,
             val va
          /\ (   (mstep j1 a (exp_inl va) /\ mstep j2 ({0 ~> va}b1) c)
              \/ (mstep j1 a (exp_inr va) /\ mstep j2 ({0 ~> va}b2) c))).
Proof.
  intros a b1 b2 c j HMS; remember (exp_sumcase a b1 b2) as ecase;
  revert a b1 b2 Heqecase; induction HMS; intros; subst;
  val_inversion;
  match goal with
  | [ H : step (exp_sumcase _ _ _) _,
      IH : forall _ _ _, _ = _ -> val _ -> _
     |- exists _, exists _, S ?n = _ /\ _ ] =>
    inversion H; subst;
    try match goal with
    | [ HS : step (exp_sumcase (exp_inl _) _ _) _ |- _ ] =>
      exists 0; exists n; eauto 6
    | [ HS : step (exp_sumcase (exp_inr _) _ _) _ |- _ ] =>
      exists 0; exists n; eauto 6
    | [ HS : step (exp_sumcase _ _ _) (exp_sumcase _ _ _) |- _ ] =>
      edestruct IH as [j1 [j2 [? [? [? [ [? ?] | [? ?] ]]]]]]; auto;
      subst; exists (S j1); exists j2;
      (repeat (split; eauto 6; try omega))
    end
  end.
Qed.
