Require Import List.
Require Import Arith.
Require Import step_defs.
Require Import step_db_basics.

(* A tactic for dealing with lemmas that have to be generalized over an
   intermediate context *)
Ltac inst_equalities :=
  repeat (match goal with
  | [ H : forall _ _, _ ++ _ = _ ++ _ -> _ |- _ ] =>
    let H' := fresh H "'" in
    assert (H' := H _ _ eq_refl); clear H
  | [ H : forall _ _, _ :: _ ++ _ = _ ++ _ -> _ |- _ ] =>
    let H' := fresh H "'" in
    rewrite app_comm_cons in H;
    assert (H' := H _ _ eq_refl); clear H
  | [ H : forall _ _, _ :: _ :: _ ++ _ = _ ++ _ -> _ |- _ ] =>
    let H' := fresh H "'" in
    repeat (rewrite app_comm_cons in H);
    assert (H' := H _ _ eq_refl); clear H
  | [ H : forall _ _ _, _ :: _ ++ _ = _ ++ _ -> _ |- _ ] =>
    let H' := fresh H "'" in
    rewrite app_comm_cons in H;
    assert (H' := H _ _ _ eq_refl); clear H
  | [ H : forall _ _ _, _ :: _ :: _ ++ _ = _ ++ _ -> _ |- _ ] =>
    let H' := fresh H "'" in
    repeat (rewrite app_comm_cons in H);
    assert (H' := H _ _ _ eq_refl); clear H
  | [ H : forall _, _ ++ _ :: _ ++ _ :: _ = _ ++ _ -> _ |- _ ] =>
    let H' := fresh H "'" in
    rewrite app_comm_cons in H; rewrite app_assoc in H;
    assert (H' := H _ eq_refl); clear H
  | [ H : forall _ _ _, _ ++ _ :: _ ++ _ :: _ = _ ++ _ -> _ |- _ ] =>
    let H' := fresh H "'" in
    rewrite app_comm_cons in H; rewrite app_assoc in H;
    assert (H' := H _ _ _ eq_refl); clear H
  | [ H : forall _ _ _, _ ++ _ = _ ++ _ -> _ |- _ ] =>
    let H' := fresh H "'" in
    assert (H' := H _ _ _ eq_refl); clear H
  end).

(* These two theorems invert some equalities on lists, given the types of
   information we get out of the clarify_compares operation *)
Theorem app_eq_lengths_eq : forall {A} (G1 G2 G3 G4 : list A),
     G1 ++ G2 = G3 ++ G4
  -> length G1 = length G3
  -> G1 = G3 /\ G2 = G4.
Proof.
  induction G1; intros G2 G3 G4 Heq Hlen;
  destruct G3; simpl in *; inversion Hlen; auto;
  inversion Heq; subst; eauto.
  match goal with
  | [ IH : forall _ _ _, ?G1 ++ _ = _ -> length ?G1 = _ -> _ /\ _,
      Heq : ?G1 ++ _ = _,
      Hlen : length ?G1 = _ |- _ ] =>
      destruct (IH _ _ _ Heq Hlen); subst; eauto
  end.
Qed.

Theorem app_eq_lengths_lt : forall {A} (G1 G2 G3 G4 : list A),
     G1 ++ G2 = G3 ++ G4
  -> length G1 < length G3
  -> exists G5, G3 = G1 ++ G5 /\ G2 = G5 ++ G4.
Proof.
  induction G1; intros G2 G3 G4 Heq Hlen;
  destruct G3; simpl in *; subst; eauto;
  try solve [inversion Hlen];
  inversion Heq; subst;
  match goal with
  | [ IH : forall _ _ _, _ ++ _ = _ ++ _ -> _ -> exists _ , _,
      Heq : _ ++ _ = _ ++ _ |- _ ] =>
    destruct (IH _ _ _ Heq) as [G5 [Heq1 Heq2]]
  end; subst; eauto.
Qed.

(* This tactic just tries to apply the previous two theorems. *)
Ltac invert_list_equalities :=
  (* Begin by making the > case look like the < case. *)
  try match goal with
  | [ H1 : length ?G2 > length ?G1, H2 : ?G2 ++ _ = ?G1 ++ _ |- _ ] =>
    unfold gt in H1; apply eq_sym in H2
  end;
  match goal with
  | [ Heq : ?G1 ++ _ = ?G2 ++ _, Hlen : length ?G1 = length ?G2 |- _ ] =>
    destruct (app_eq_lengths_eq _ _ _ _ Heq Hlen); subst; clear Hlen Heq
  | [ Heq : ?G1 ++ _ = ?G2 ++ _, Hlen : length ?G1 < length ?G2 |- _ ] =>
    let G5 := fresh "G5" in
    let HEq1 := fresh "HEq1" in let HEq2 := fresh "HEq2" in
    destruct (app_eq_lengths_lt _ _ _ _ Heq Hlen) as [G5 [Heq1 Heq2]]; subst
  end.

(* this one looks for the situation where we know:

     length G1 < length (G1 ++ G5)

   Then G5 must be a cons, so we expose it *)
Ltac find_cons := 
  match goal with
  | [ Hlen : length ?G1 < length (?G1 ++ ?G5) |- _ ] =>
    destruct G5; 
      [  subst; rewrite app_length in Hlen; rewrite plus_comm in Hlen; 
         simpl in Hlen; destruct (lt_irrefl _ Hlen) 
       | subst; clear Hlen ];
    match goal with
    | [ Heq2 : _ :: _ = (_ :: _) ++ _ |- _ ] =>
      inversion Heq2; subst; clear Heq2
    end
  | [ HConsEq : _ :: _ = _ :: _ |- _ ] =>
    inversion HConsEq; subst; clear HConsEq
  end.

  
(* Weakening *)
Theorem weakening' : forall G t a A,
     typing G t a A 
  -> forall G1 G3, G = G1 ++ G3
  -> forall G2,
      let n := length G2 in
      let m := length G1 in
        typing (G1 ++ G2 ++ G3) t (lift_e a n m) A.
Proof.
  induction 1;
  intros; subst; eauto;
  repeat (match goal with | [n := length _ : nat |- _] => subst n end);
  try solve [inst_equalities; simpl in *; autorewrite with steps in *; eauto];
    simpl in *.

  (* The var case *)
  clarify_compares; subst; eauto; invert_list_equalities;
  [ 
    (* Eq *)
    inst_equalities;
    autorewrite with steps; rewrite app_assoc; simpl; eauto;
    apply TVar; autorewrite with steps; eauto
  |   
    (* Lt *)
    find_cons; autorewrite with steps; simpl;
    rewrite <- app_assoc; rewrite <- app_comm_cons; eauto
  | 
    (* Gt *) 
    repeat (rewrite app_assoc); apply TVar;
    autorewrite with steps; omega
  ].
Qed.

Theorem weakening : forall G1 G2 G3 t a A,
     typing (G1 ++ G3) t a A 
  -> typing (G1 ++ G2 ++ G3) t (lift_e a (length G2) (length G1)) A.
Proof.
  intros; eauto using weakening'.
Qed.

Theorem weakening_pre_app : forall G1 G2 t a A,
     typing G2 t a A
  -> typing (G1 ++ G2) t (lift_e a (length G1) 0) A.
Proof.
  intros;
  apply weakening with (G1 := nil) (G2 := G1) in H; eauto.
Qed.
Hint Resolve weakening_pre_app.

Theorem weakening_cons : forall G t' B t a A,
     typing G t a A
  -> typing ((t',B) :: G) t (lift_e a 1 0) A.
Proof.
  intros G t' B t a A HTyp;
  apply weakening_pre_app with (G1 := (t',B) :: nil) in HTyp;
  simpl in *; eauto.
Qed.
Hint Resolve weakening_cons.

Theorem weakening_nil : forall G t a A,
  typing nil t a A -> typing G t a A.
Proof.
  intros G t a A HTyp;
  replace G with (G ++ nil) by (rewrite app_nil_r; auto);
  replace a with (lift_e a (length G) 0) 
              by (erewrite lift_closed_id; eauto);
  eauto.
Qed.
Hint Immediate weakening_nil.


Section Substitution.

  Variable v : exp.

  Theorem substitution' : forall G t t' b B,
       typing G t b B
    -> forall G1 A G2, G = G1 ++ ((t',A) :: G2)
        -> val v
        -> typing G2 t' v A
        -> typing (G1 ++ G2) t
                  ({length G1 ~> lift_e v (length G1) 0} b) B.
  Proof.
    induction 1;
    intros; subst; simpl in *; eauto;
    try solve [inst_equalities; try rewrite lift_lift_swap_0;
               simpl in *; autorewrite with steps in *;
               eauto].

    (* Var case *)
    clarify_compares; subst; simpl in *; invert_list_equalities;
    find_cons; eauto;
    [ (* lt *)
      rewrite <- app_assoc; simpl; eauto
    | (* gt *)
      rewrite app_assoc; apply TVar; autorewrite with steps;
      simpl; omega
    ].
  Qed.

  Theorem substitution : forall G1 t' A G2 t b B,
       typing (G1 ++ (t',A) :: G2) t b B
    -> val v
    -> typing G2 t' v A
    -> typing (G1 ++ G2) t ({length G1 ~> lift_e v (length G1) 0} b) B.
  Proof.
    intros; eauto using substitution'.
  Qed.

  Theorem substitution_front : forall t' A G t b B,
       typing ((t',A) :: G) t b B
    -> val v
    -> typing G t' v A
    -> typing G t ({0 ~> v}b) B.
  Proof.
    intros;
    match goal with 
    | [ |- typing ?G _ ({0 ~> ?v} ?b) _] =>
      replace G with (nil ++ G); eauto;
      replace ({0 ~> v}) 
         with (   {length (@nil (th * exp)) 
               ~> lift_e v (length (@nil (th * exp))) 0})
         by (autorewrite with steps; eauto)
    end;
    eapply substitution; eauto.
  Qed.

End Substitution.
Hint Resolve substitution substitution_front.
